#include <SoftwareSerial.h>
#include <MemoryFree.h>
#include <limits.h>

/*

Programme pour une station météo qui capte:
  - la vitesse du vent,
  - l'orientation du vent, et 
  - la température en degré Celsius.
Ces données sont transmises sur Internet via un module cellulaire.

Projet réalisé par les frères Guylain et Paul-André Plante.

Avril 2016   

*/

/********  Déclarations des constantes et des prototypes des procédures relatives au LCD  ********/
// Les liaisons électriques entre le LCD et l'Arduino Uno Rev 3 ne sont pas les mêmes
// que celles du projet 11 nommé Crystal Ball décrit aux pages 114 et suivantes
// du manuel inclus dans le Arduino Starter Kit.
const unsigned int LCD_pin_D7 = 6;       //  sur Arduino Digital pin  6    (port PD6 de l'ATmega328P)
const unsigned int LCD_pin_D6 = 7;       //  sur Arduino Digital pin  7    (port PD7 de l'ATmega328P)
const unsigned int LCD_pin_D5 = 8;       //  sur Arduino Digital pin  8    (port PB0 de l'ATmega328P)
const unsigned int LCD_pin_D4 = 9;       //  sur Arduino Digital pin  9    (port PB1 de l'ATmega328P)

const unsigned int LCD_pin_RS = 12;      //  sur Arduino Digital pin 12    (port PB4 de l'ATmega328P)
//   RS = Register Select        IR (Instruction Register aussi appelé command register) = 0
//                               DR (Data Register) = 1
const unsigned int LCD_pin_Enable = 11;  //  sur Arduino Digital pin 11    (port PB3 de l'ATmega328P)
// Entrée Read/Write du LCD:  dans plusieurs  projets disponibles sur Internet, la connexion Read/Write du LCD
// est mise à la masse, laissant ainsi la connexion à Write de manière permanente. C'est ce qui sera fait
// dans le présent logiciel, et par conséquent il n'y aura aucune référence relative à cette entrée du LCD.
//
//  Voici le brochage d'un LCD "standard":
//                          E
//                          n
//      V   V               a                                   L   L
//      s   c       R   R   b                                   E   E
//      s   c   V   /   /   l   D   D   D   D   D   D   D   D   D   D
//     (-) (+)  o   S   W   e   0   1   2   3   4   5   6   7   +   -
//  -----------------------------------------------------------------------------------------------
//  |   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15  16                             |
//  |                                                                                             |
//  |                                                                                             |
//

void LCD_Initialisation();
void LCD_SetAddress_WriteData(byte, byte, byte);
void LCD_Initialiser6Pins_Write(byte);
void LCD_BitsPoidsFort_Write(byte);
void LCD_BitsPoidsFaible_Write(byte);
void LCD_Enable();

void LCD_ClearDisplay();
void LCD_ReturnHome();

void LCD_AfficherNombre(byte, signed int, byte);
void LCD_AfficherChaine(byte, byte []);
void LCD_AfficherChaine(byte, const char *); //Guylain - "Function overloading" pour fonction du meme nom
void LCD_AfficherChaine(byte, const __FlashStringHelper *); //Guylain - "Function overloading" pour fonction du meme nom

void LCD_CaracteresSurMesure();

/********  Déclarations des constantes et des prototypes des procédures relatives aux senseurs  **/
//  Senseur pour la lecture de la direction du vent via une girouette (Wind vane).
//  Pour ce senseur, on utilise une entrée analogique.
const unsigned int Senseur_pin_Girouette = A0;       //  sur Arduino Analog  pin A0    (port PC0 de l'ATmega328P)
//                                                       aussi appelée Digital pin 14
//  A0 est une constante déjà dans le système, et correspond à 14.

void Meteo_LireVentDirection(byte);

//  Senseur pour la lecture de la vitesse du vent avec l'anémomètre.
//  Pour ce senseur, on utilise une entrée digitale (qui dans ce cas est aussi une entrée analogique).
const unsigned int Senseur_pin_Anemometre = A1;      //  sur Arduino Digital pin 15     (port PC1 de l'ATmega328P)
//                                                       aussi appelée Analog pin A1

void Meteo_LireVentVitesse(byte);

//  Senseur pour la lecture de la température
//  Pour ce senseur, on utilise une entrée digitale (qui dans ce cas est aussi une entrée analogique).
const unsigned int Senseur_pin_Temperature = A3;     //  sur Arduino Digital pin 17     (port PC3 de l'ATmega328P)
//                                                       aussi appelée Analog pin A3

void    Temperature_Principal_01(byte);
boolean Temperature_Principal_02();
boolean Temperature_Reset();
void    Temperature_EnvoyerCode(byte);
boolean Temperature_LireDonnees(int);
boolean Temperature_CRC8(int);
void    Temperature_CalculsFinaux();

/********  Déclarations des constantes et des prototypes des procédures relatives au Fona 3G  ****/

// ************************************************
// Guylain Plante le 10 avril 2016
// Les routines qui parlent au FONA 3G utilisent beaucoup de code emprumpté
// à la librairie FONA (https://github.com/adafruit/Adafruit_FONA) 
// développée par Adafruit Industries (http://www.adafruit.com).
//
// Puisque la librairie prend trop de place mémoire (à cause de la mémoire limité du Arduino Uno de seulement 2048 octets), 
// j'ai du uniquement prendre les sections de la librairie dont j'avait besoin 
// pour le projet et les adapter pour nos besoins.
// Nous utilisons la carte FONA 3G (American version) de Adafruit Industries.
// Le chip sur cette carte est un SIMCOM_SIM5320A révision SIM5320A_V1.5.
// ************************************************

#define FONA_DEFAULT_TIMEOUT_MS  500

#define FONA_RX  2   //  sur Arduino Digital pin  2    (port PD2 de l'ATmega328P)
#define FONA_TX  3   //  sur Arduino Digital pin  3    (port PD3 de l'ATmega328P)
//#define FONA_TX  10  //  sur l'Arduino Mega 2560 On doit plutôt utiliser la pin 10

#define FONA_RST 4   //  sur Arduino Digital pin  4    (port PD4 de l'ATmega328P)
#define FONA_KEY 5   //  sur Arduino Digital pin  5    (port PD5 de l'ATmega328P)

typedef const __FlashStringHelper * FONAFlashStringPtr;
char    fona3g_replybuffer[255];

boolean fona3g_sendDataToWebServer(char const *dataToSend); 
boolean fona3g_init(Stream &port);
void    fona3g_power_down();
boolean fona3g_getAndProcessPageBody(uint16_t timeout);
boolean fona3g_waitForDeferredResponse(const __FlashStringHelper * expectedResponse, uint16_t timeout);
boolean fona3g_waitForDeferredResponse(char const *expectedResponse, uint16_t timeout);
uint8_t fona3g_sendOnly(char const *atCommand);

boolean fona3g_sendCheckReply(char *send, char *reply, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
boolean fona3g_sendCheckReply(FONAFlashStringPtr send, FONAFlashStringPtr reply, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
boolean fona3g_sendCheckReply(char* send, FONAFlashStringPtr reply, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
boolean fona3g_sendCheckReply(FONAFlashStringPtr prefix, char *suffix, FONAFlashStringPtr reply, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
boolean fona3g_sendCheckReply(FONAFlashStringPtr prefix, int32_t suffix, FONAFlashStringPtr reply, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
boolean fona3g_sendCheckReply(FONAFlashStringPtr prefix, int32_t suffix, int32_t suffix2, FONAFlashStringPtr reply, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
void    fona3g_flushInput();
uint8_t fona3g_readline(uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS, boolean multiline = false);
uint8_t fona3g_getReply(char *send, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
uint8_t fona3g_getReply(FONAFlashStringPtr send, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
uint8_t fona3g_getReply(FONAFlashStringPtr prefix, char *suffix, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
uint8_t fona3g_getReply(FONAFlashStringPtr prefix, int32_t suffix, uint16_t timeout = FONA_DEFAULT_TIMEOUT_MS);
uint8_t fona3g_getReply(FONAFlashStringPtr prefix, int32_t suffix1, int32_t suffix2, uint16_t timeout); // Don't set default value or else function call is ambiguous.

SoftwareSerial fonaSS = SoftwareSerial(FONA_TX, FONA_RX);
SoftwareSerial *fonaSerial = &fonaSS;
Stream *mySerial_Fona;

/********  Déclarations de la constante et du prototype de la procédure relative au clavier  *****/
//  La lecture du clavier se fait via une entrée analogique.
const unsigned int Clavier_pin =  A2;                 //  sur Arduino Analog  pin A2    (port PC2 de l'ATmega328P)
//                                                        aussi appelée Digital pin 16
//  A2 est une constante déjà dans le système, et correspond à 16

void Clavier_Lire(int);


/********  Déclarations des prototypes de procédures diverses  ***********************************/
void Convertir_Integer_ASCII(unsigned int);

#define METEO_DEBUG
//#define METEO_DEBUG_DO_NOT_SEND_DATA_TO_SERVER //Decommenter ce define pour na PAS envoyer les donnees sur le serveur. Pratique pour faire des tests et economiser de la bande passante.
#define DebugStream              Serial
#define DEBUG_PRINT(...)         DebugStream.print(__VA_ARGS__)
#define DEBUG_PRINTLN(...)       DebugStream.println(__VA_ARGS__)

#define INFO_TO_DISPLAY_INTERVAL                 3000 
#define INFO_TO_DISPLAY_LAST_TYPE_NUMBER           12

#define INFO_TO_DISPLAY_TYPE_DATE                   1 //Date et heure
#define INFO_TO_DISPLAY_TYPE_WIND                   2 //Finalement, j'affiche aussi la temperature sur la meme ligne.
#define INFO_TO_DISPLAY_TYPE_NBR_LECTURES           3 //Nombre de lectures faites vs total a faire avant envoi au serveur (1 lecture par minute)
#define INFO_TO_DISPLAY_TYPE_BATTERY                4 //Voltage de la batterie //J'affiche aussi la memoire de disponible sur cette ligne.
#define INFO_TO_DISPLAY_TYPE_LAST_COMM_STATUS       5 //Statut, derniere communication avec serveur web (peu imorte si la reponse retournee par la page web etait erreur (probablement a cause que les donnees envoyees n'ont pu etre enregistrees dans la base de donnees))
#define INFO_TO_DISPLAY_TYPE_LAST_RESPONSE_STATUS   6 //Dernier reponse de la page web 
#define INFO_TO_DISPLAY_TYPE_INFO_1                 7 //Projet realise
#define INFO_TO_DISPLAY_TYPE_INFO_2                 8 //par les freres 
#define INFO_TO_DISPLAY_TYPE_INFO_3                 9 //Guylain et
#define INFO_TO_DISPLAY_TYPE_INFO_4                10 //Paul-Andre
#define INFO_TO_DISPLAY_TYPE_INFO_5                11 //Plante
#define INFO_TO_DISPLAY_TYPE_EMPTY                 12 

#define MAX_LCD_CHARACTERS                         16 //Le LCD a un maximum de 16 colonnes.
#define METEO_VALUE_UNSET                         255 
#define METEO_VALUE_OUT_OF_BOUNDS                 254 
#define METEO_CMD_STATUS_UNDEFINED                 -1
#define METEO_CMD_STATUS_ERROR                      0
#define METEO_CMD_STATUS_SUCCESS                    1

#define SENSOR_VALUES_RECORD_PADDING_CHARACTER    "x"
#define SENSOR_VALUES_RECORD_LENGTH                10 // Vitesse vent = 3 caracteres ("xx0" a "999"), Direction vent = 1 caractere ("0" a "7"), Temperature = 3 caracteres ("-99" a "x99"), Vent vitesse maximum atteinte = 3 caracteres ("xx0" a "999") (x = SENSOR_VALUES_RECORD_PADDING_CHARACTER)

#define SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES                           15 // Sert pour tableau sensorValuesArray
#define SENSOR_VALUES_ARRAY_NBR_ENTRIES_TO_KEEP_BEFORE_SENDING_TO_SERVER   5 // Sert pour tableau sensorValuesArray. Ne peut etre plus grand que SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES
#define SENSOR_VALUES_ARRAY_WAIT_DELAY_IN_SECONDS                        120 // Normalement 120 pour qu'on garde une entree par deux minutes

#define SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES                   40 // Sert pour tableau sensorAverageValuesArray

#define SENSOR_VALUES_DIMENSION_VENT_VITESSE          0  // Sert pour tableaux sensorValuesArray et sensorAverageValuesArray
#define SENSOR_VALUES_DIMENSION_VENT_DIRECTION        1  // Sert pour tableaux sensorValuesArray et sensorAverageValuesArray
#define SENSOR_VALUES_DIMENSION_TEMPERATURE           2  // Sert pour tableau sensorValuesArray
#define SENSOR_VALUES_DIMENSION_VENT_VITESSE_MAXIMUM  3  // Sert pour tableau sensorValuesArray

#define DIRECTION_VENT_NORD       0
#define DIRECTION_VENT_NORD_EST   1
#define DIRECTION_VENT_EST        2
#define DIRECTION_VENT_SUD_EST    3
#define DIRECTION_VENT_SUD        4
#define DIRECTION_VENT_SUD_OUEST  5
#define DIRECTION_VENT_OUEST      6
#define DIRECTION_VENT_NORD_OUEST 7

#define LEAPOCH (946684800LL + 86400*(31+29))    //sert pour fonction Meteo_secs_to_tm
#define DAYS_PER_400Y (365UL *400 + 97)          //sert pour fonction Meteo_secs_to_tm
#define DAYS_PER_100Y (365UL *100 + 24)          //sert pour fonction Meteo_secs_to_tm
#define DAYS_PER_4Y   (365UL *4   + 1)           //sert pour fonction Meteo_secs_to_tm

struct tm { //sert pour fonction Meteo_secs_to_tm
  int tm_sec;   //seconds after the minute 0-59
  int tm_min;   //minutes after the hour  0-59
  int tm_hour;  //hours since midnight  0-23
  int tm_mday;  //day of the month  1-31
  int tm_mon;   //months since January  0-11
  int tm_year;  //years since 1900
  int tm_wday;  //days since Sunday  0-6
  int tm_yday;  //days since January 1  0-365
  //int tm_isdst; //Daylight Saving Time flag //Ne sert pas dans la fonction Meteo_secs_to_tm utilisee avec cette structure ci-bas.
};

struct tm my_tm;
struct tm *my_tm_ptr = &my_tm ; // a pointer to my_tm

// GP: Vu que c'est mon premier projet avec Arduino, ca me tente de faire quelques tests avec PROGMEM mais je pourrait aussi tout simplement faire des #define, ce serait plus simple mais bon :-)
// Donc, pour sauver de la mémoire, je place quelques constantes dans la mémoire FLASH au lieu de la mémoire SRAM, à l'aide du keyword PROGMEM.
// Cependant, ceci nécessite une manipulation spéciale. Premièrement, les variables PROGMEM doivent être déclarées globalement OU si déclarées à 
// l'intérieur d'une fonction, doivent être déclarées à l'aide du keyword static. 
// De plus, j'utilise des fonctions différentes pour manipuler ces constantes (chaines) comme par exemple la fonction strcat_P() au lieu de la fonction strcat();
// Plus d'informations à https://www.arduino.cc/en/Reference/PROGMEM

const PROGMEM char mySettings_cell_apn[]                        = "rogers-core-appl1.apn";  // Access Point Name (APN). Dans le FONA 3G, j'utilise une carte SIM de Petro-Canada (http://mobility.petro-canada.ca) et ceux-ci semblent passer via le reseau de Rogers (http://www.rogers.com)
const PROGMEM char mySettings_upload_web_server_host_name[]     = "www.votre-site-web.com"; // Changer pour nom de votre serveur web
const PROGMEM char mySettings_upload_web_server_port[]          = "443";                    // Normalement, SSL est sur le port 443 mais vous pourriez utiliser n'importe quel port si vous configurez votre serveur pour répondre sur le port que vous indiqué ici. Pratique si pas exemple vous désirez utiliser un "Self-Signed Certificate" puis pour ne pas déranger l'accès à votre site web par vos visiteurs, vous décidez de configurer les communications SSL sur un autre port, seulement pour l'accès par la station météo.
const PROGMEM char mySettings_upload_web_server_page_name[]     = "page1.php";              // page web sur votre site qui va s'occuper de traiter les données transmises.
const PROGMEM char mySettings_upload_web_server_page_password[] = "123456";                 // Mettez le même mot de passe ici et dans votre page web.

int8_t    myResponse_status            = METEO_CMD_STATUS_UNDEFINED;
int8_t    myResponse_last_comm_status  = METEO_CMD_STATUS_UNDEFINED;

//Va arreter de fonctioner en 2038... a verifier plus tard... :-)
uint32_t myResponse_server_date_time         = 0;   //Unix timestamp -> Number of seconds since the Unix Epoch (January 1 1970 00:00:00 GMT)
uint32_t myResponse_server_date_time_millis  = 0;   //On garde millis() dans cette variable au meme moment ou on affecte la variable myResponse_server_date_time
float    myResponse_battery_voltage          = 0.0;

uint8_t sensorValuesArray[SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES][4];
uint8_t sensorAverageValuesArray[SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES][2]; //init, add, getAverage

unsigned long lastTime_AverageCalculated = 0;
unsigned long lastTime_Meteo_updateLCD_fired = 0;

int  VentDirection_int = METEO_VALUE_UNSET;
int  VentVitesse_int = 0;
int  Temperature_int = 0;

void    Meteo_sensorValuesArray_reset();
int     Meteo_sensorValuesArray_getNbrEntry();
void    Meteo_sensorValuesArray_addTo(int ventVitesse, int ventDirection, int temperature, int ventVitesseMaximum);

void    Meteo_sensorAverageValuesArray_reset();
int     Meteo_sensorAverageValuesArray_getNbrEntry();
void    Meteo_sensorAverageValuesArray_addTo(int ventVitesse, int ventDirection);
void    Meteo_sensorAverageValuesArray_to_sensorValuesArray();

char *  Meteo_getString_dataToSend();
char *  Meteo_VitesseVent_int_to_string_plusPadding(int i);
char *  Meteo_DirectionVent_int_to_string(int i);
char *  Meteo_Temperature_int_to_string_plusPadding(int i);
void    Meteo_updateLCD();
void    Meteo_updateLCD(int lineNumber, const char * stringToDisplay);
char *  Meteo_getInfoToDisplayOnLCD(int info_to_display);
char *  Meteo_getVentDirectionIntToString(int ventDirection);
int     Meteo_secs_to_tm(long long t, struct tm *tm); //http://git.musl-libc.org/cgit/musl/tree/src/time/__secs_to_tm.c?h=v0.9.15

/********  Déclarations d'autres variables globales  **************************************************/
byte VentDirection[3];
byte VentVitesse[4];
byte Temperature[10];       //   variable globale de 10 octets;  elle est initialisée à  "+999.9999"
                            //   1 octet  pour le signe
                            //   3 octets pour le nombre entier
                            //   1 octet  pour le point décimal
                            //   4 octets pour les décimales
                            //   1 octet  pour le null
byte ClavierCaractereLu[2];

byte Chaine001_ASCII[5];   // variable globale qui contient les valeurs ASCII d'un integer converti;
// ces valeurs sont destinées à l'affichage du LCD.
// Le 5ième caractère contient le NULL.
// Cette variable globale peut être utilisée pour autre chose si désiré.

byte donnees_temp[9];       // variable globale générale;  sert surtout pour le sensor de température

/*************************************************************************************************/
/********  Procédure d'initialisation  (obligatoire pour l'Arduino)  *****************************/
void setup()
  {
  Serial.begin(115200);    
  while (!Serial);
  Serial.print(__func__);
  Serial.println(F("() debut")); 

  // Note au sujet de la Digital pin 13 (port PB5 de l'Atmega328P) pour la LED clignotante.
  // Plusieurs croquis fournis avec l'Arduino ont assigné cette pin à la LED clignotante sur le circuit lui-même.
  // Pour le moment, je préfère ne pas l'utiliser, mais elle reste disponible bien sûr.
  // En cas de doute ou de bogues difficiles à cerner, on peut au moins tester
  // que l'Arduino fonctionne normalement en faisant simplement clignoter cette LED.
  pinMode(13, OUTPUT);
  // Bout de code emprunté à Blink, juste pour voir si l'Arduino peut rouler normalement, en cas de problème.
  digitalWrite(13, HIGH);   // turn the LED on (HIGH is the voltage level)
  delay(1000);            
  digitalWrite(13, LOW);    // turn the LED off by making the voltage LOW
  delay(500);             

  // Initialisation du LCD
  LCD_Initialisation();
  LCD_ClearDisplay();
  LCD_CaracteresSurMesure();
  LCD_AfficherChaine(0x00,F("Station M{t{o   "));
  LCD_AfficherChaine(0x40,F("Bonjour!        "));
  Meteo_sensorValuesArray_reset();
  Meteo_sensorAverageValuesArray_reset();
  delay(INFO_TO_DISPLAY_INTERVAL);
  
  Clavier_Lire(30); 
  // Si le clavier a retourné 'P' pour Production, on change la valeur pour '0'
  if (ClavierCaractereLu[0] == 'P')  {  ClavierCaractereLu[0] = '0';   }
  Serial.print("Clavier :  ");   Serial.println(char(ClavierCaractereLu[0]));

  switch (ClavierCaractereLu[0])
    {
    case '1' :  Meteo_LireVentDirection(' ');                     break;    // va boucler à l'infini
    case '2' :  Meteo_LireVentVitesse(' ');                       break;    // va boucler à l'infini    
    case '3' :  Temperature_Principal_01(' ');                    break;    // va boucler à l'infini    
    case '4' :                                                    break;    
    case '5' :                                                    break;    
    case '6' :                                                    break;    
    case '7' :                                                    break;    
    case '8' :                                                    break;    
    case '9' :                                                    break;    
    case '0' :                                                    break;       
    }

  lastTime_AverageCalculated = millis();
  Serial.print(__func__);
  Serial.println(F("() fin")); 

  }

/*************************************************************************************************/
/********  Boucle principale  (obligatoire pour l'Arduino)  **************************************/
void loop()
  {
  unsigned long currentTime = millis();  
  char * p_dataToSend = NULL;

  //if (lastTime_AverageCalculated > 0) {
  // Meteo_updateLCD();
  //}
  
  // On lit les sensors sans arret et on garde les donnees dans le tableau sensorAverageValuesArray et au 2 minutes, 
  // on fait la moyenne puis on garde le resultat dans le tableau sensorValuesArray. Ce sont les donnees de ce dernier tableau qu'on envoi su serveur.
  if ( (currentTime >= (lastTime_AverageCalculated + (SENSOR_VALUES_ARRAY_WAIT_DELAY_IN_SECONDS * 1000UL))) ) { //Si je ne met pas UL pour "Unsigned Long" à droite de la constante 1000 et bien je recoit un Warning a la compilation puis je n'aime pas ca meme si ce ne serait pas grave. (warning: integer overflow in expression [-Woverflow])
   //Apres 2 minutes, on fait la moyenne des valeurs lues et on transfert cela dans le tableau sensorValuesArray.
   Meteo_sensorAverageValuesArray_to_sensorValuesArray();   

   #ifdef METEO_DEBUG
     for (int i = 0; i < SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {
      if (sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE] != METEO_VALUE_UNSET) {
       for (int x = 0; x <= 3; x++) {
        #ifdef METEO_DEBUG   
         DEBUG_PRINT(F("i="));
         DEBUG_PRINT(i);
         DEBUG_PRINT(F(" x="));
         DEBUG_PRINT(x);
         DEBUG_PRINT(F(" => "));
         DEBUG_PRINTLN(sensorValuesArray[i][x]);
        #endif
       }
      }
     }
    #endif   

   // Si on a 5 entrees (donc apres 10 minutes mais en fait le nombre d'entree a garder est determine par SENSOR_VALUES_ARRAY_NBR_ENTRIES_TO_KEEP_BEFORE_SENDING_TO_SERVER)  
   // on envoi les donnees sur le serveur puis ensuite on reset le tableau pour recommencer.
   //
   // Si jamais pour une raison ou une autre, on ne peut envoyer les donnees au serveur, et bien les lectures vont 
   // quand-meme continuer mais nous allons toujours seulement garder dans le tableau les 15 derniers resultats (ou plutot le nombre dans la constante SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES).
   //
   // Aussi, au démarrage de l'Arduino on va chercher la date et heure sur le serveur donc dans un tel cas, on attend pas 10 minutes mais seulement 2 minutes.
   // Je me sert de la date et l'heure du serveur web pour l'afficher sur le LCD mais surtout comme "id de transaction" si on peut dire pour 
   // que le server sache si on a deja envoye ces donnees. Consultez le fichier exemple page1.php pour plus de détails à propos du numéro de transaction.
   //
   if ((Meteo_sensorValuesArray_getNbrEntry() >= SENSOR_VALUES_ARRAY_NBR_ENTRIES_TO_KEEP_BEFORE_SENDING_TO_SERVER) || 
       (myResponse_server_date_time == 0) ) { //Si on a pas encore la date du serveur, on force un envoi des donnees au serveur pour aller la chercher.       
     p_dataToSend = Meteo_getString_dataToSend();
     if ( fona3g_sendDataToWebServer(p_dataToSend) ) {
       //L'envoi des donnees au serveur a reussie dont on vide le tableau
       Meteo_sensorValuesArray_reset();
     } else {
       // Sinon et bien un nouvel essai sera fait dans 2 minutes tout simplement donc on ne vide pas le tableau
     }
     free(p_dataToSend);
   }

   lastTime_AverageCalculated = millis(); //Je le met apres l'envoi au serveur pour que chaque groupe de lecture dure 2 minutes chacun.
  }
  
  //if (true) {

    // On va lire les sensors.
    #ifdef METEO_DEBUG   
     DEBUG_PRINTLN(F("Meteo_LireVentDirection()..."));
    #endif
    Meteo_LireVentDirection('P');
    //delay(INFO_TO_DISPLAY_INTERVAL);

    #ifdef METEO_DEBUG   
     DEBUG_PRINTLN(F("Meteo_LireVentVitesse()..."));
    #endif
    Meteo_LireVentVitesse('P');  
    //delay(INFO_TO_DISPLAY_INTERVAL);

    #ifdef METEO_DEBUG   
     DEBUG_PRINTLN(F("Temperature_Principal_01()..."));
    #endif
    Temperature_Principal_01('P');
    //delay(INFO_TO_DISPLAY_INTERVAL);
    
    // On garde les resultats lus dans le tableau sensorAverageValuesArray (sauf la temperature qu'on a pas besoin de faire de moyenne).
    //Meteo_sensorAverageValuesArray_addTo(random(0,99),random(0,7)); //Si je veut faire des tests
    Meteo_sensorAverageValuesArray_addTo(VentVitesse_int,VentDirection_int);

    #ifdef METEO_DEBUG
     for (int z = 0; z < SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; z++) {
      if (sensorAverageValuesArray[z][SENSOR_VALUES_DIMENSION_VENT_VITESSE] != METEO_VALUE_UNSET) {
       for (int x = 0; x <= 1; x++) {
        #ifdef METEO_DEBUG   
         DEBUG_PRINT(F("z="));
         DEBUG_PRINT(z);
         DEBUG_PRINT(F(" x="));
         DEBUG_PRINT(x);
         DEBUG_PRINT(F(" => "));
         DEBUG_PRINTLN(sensorAverageValuesArray[z][x]);
        #endif
       }
      }
     }     
    #endif          
        
  //}
    
  }

/*************************************************************************************************/
/********  Procédures propres au LCD  ************************************************************/

void LCD_Initialisation()
  {
  // Cette procédure regroupe les trois instructions nécessaires pour initialiser le LCD,
  // c'est-à-dire, dans l'ordre d'exécution:
  //   1- l'instruction Function set
  //   2- l'instruction Display on/off control
  //   3- l'instruction Entry mode set

  delay(100);     // recommandé après le démarrage du LCD (cold start).

  // Mais au préalable il faut initialiser sur l'Arduino lui-même les 2 liens de contrôle du LCD, c'est-à-dire:
  //   - RS (Register Select), et
  //   - Enable.
  // On peut, sans danger électronique, les initialiser de manière permanente en OUTPUT
  pinMode(LCD_pin_RS, OUTPUT);
  pinMode(LCD_pin_Enable, OUTPUT);
  // On place ces deux pins à LOW.
  digitalWrite(LCD_pin_RS, LOW);
  digitalWrite(LCD_pin_Enable, LOW);

  /*
  1ière étape de l'initialisation du LCD:  l'instruction Function set

  Cette instruction paramétrise:
    - la connexion logique sur 4 ou 8 bits (ou connexion électrique sur 4 ou 8 fils),
    - le nombre de lignes de caractères sur le LCD, et
    - le choix de la matrice des caractères.

  Configuration des bits D7 à D0:
    D7 = 0
    D6 = 0
    D5 = 1  bit-clé permettant de sélectionner une instruction de type Function set
    D4 = 0  DL  Data Length          8 bits:  DL = 1            Les échanges peuvent se faire sur une connexion
                                     4 bits:  DL = 0            électrique à 8 fils (8-bit), ou à 4 fils (4-bit)

    D3 = 1  N   Number of display lines    2 lignes:  N = 1           1 ligne:  N = 0
    D2 = 0  F   character Font             5 x 10 pixels:  F = 1      5 x 8 pixels:  F = 0
    D1 = 0, mais sans importance
    D0 = 0, mais sans importance
    On obtient donc 0010 1000, ou 28 en hexadécimal.

  Dans le cadre du présent projet, le LCD sera relié par une connexion à 4 fils (mode 4-bit)
  au niveau des données.  Le seul avantage est d'utiliser seulement 4 fils, au lieu de 8 fils.
  Toutes les données continuent d'être traitées sur 8 bits, tant sur l'Arduino que dans le LCD.

  Du côté du LCD, seules les connexions des 4 bits de poids fort sont reliées à l'Arduino.
  Les connexions des 4 bits de poids faible sont "laissées en l'air" ou "flottantes",
  c'est-à-dire sans aucune connexion.

  Du côté de l'Arduino, comme le transfert des données vers le LCD se fera sur 4 bits, il faudra donc y aller
  en conséquence, et le transfert des données devra se faire en deux temps:
    - d'abord envoyer les 4 bits de poids fort, et puis
    - ensuite envoyer les 4 bits de poids faible.

  Cependant, il est important de savoir que lors d'un démarrage (cold start), le LCD est programmé en mode 8-bit.
  Exceptionnellement, pour une seule fois au tout début, il faudra envoyer la première donnée d'initialisation
  comme si le bus de données était de 8 bits, donc en un seul envoi;  comme les connexions pour
  les 4 bits de poids faible sur le LCD ne sont pas reliés à l'Arduino, on accepte d'avance
  que ces 4 bits de poids faible soient sans importance dans cette situation précise.
  Juste pour mieux comprendre, on voit ci-haut que les bits importants pour la présente instruction Function set
  sont les bits D5 et D4, donc dans le groupe de bits de poids fort, et comme les bits de poids fort
  ont une connexion avec le LCD, on est sûr qu'ils seront pris en compte.
  La donnée sera interprétée par le LCD, et le fera passer en mode 4-bit.

  Dans le précédent paragraphe, le mot "donnée" a été utilisé, mais en réalité ce sera une donnée
  qui est en fait une instruction (ou command) à exécuter, plutôt qu'une véritable donnée qui s'afficherait sur le LCD.

  On envoie donc la donnée, deux fois de suite, de la façon suivante:
    1- envoi de la donnée 0010 1000 comme si on était sur un bus de 8 bits;
       concrètement on place les 4 bits de poids fort sur les 4 sorties prévus sur l'Arduino, et on actionne Enable,
       en ayant au préalable programmé correctement les lignes RS et RW;
       les 4 bits de poids faible sont comme oubliés; le LCD passe immédiatement en mode 4-bit.
    2- on envoie à nouveau la même donnée, mais cette fois-ci tous ses 8 bits;
       comme le LCD est dorénavant en mode 4-bit, il faudra envoyer la donnée en deux temps consécutifs,
       tel qu'expliqué ci-avant;  pour les 4 bits de poids fort, l'effet sera redondant, sachant
       que le bit D4 sert à programmer le mode 8-bit ou 4-bit;  comme cette fois-ci les 4 bits de poids faible
       seront pris en compte, les bits D3 et D2 pourront maintenant avoir leur effet.

  Tout ce qui vient d'être dit est valable seulement lorsque le LCD sort de l'usine..., ou après un redémarrage (cold start),
  à condition que le programme déjà dans l'Arduino fonctionne correctement.

  Cependant, ce n'est toujours le cas...  En effet, en mode développement, lors d'un plantage par exemple,
  on ne sait plus trop si le LCD est demeuré en mode 4-bit ou 8-bit, ou si lors du redémarrage (cold start),
  il a été correctement initialisé par le programme déjà dans l'Arduino, programme qui peut s'avérer possiblement bogué.
  De plus, si l'échange est en 4-bit, on ne sait pas si on est rendu au 1er ou au 2ième envoi de bits.
  Il existe donc beaucoup d'incertitude en pareilles situations.
  La solution est d'expédier 3 fois les bits de poids fort d'une donnée bien précise, pour forcer le LCD
  à passer en mode 8-bit comme lors d'un redémarrage (cold start);  les bits de poid fort sont comme
  les bits D7 à D4 fixés ci-haut, mais cette fois-ci le bit D4 sera à 1 (DL = 1 ou mode 8-bit).
  Comme on envoie seulement les bits de poids fort, ceux de poids faible n'ont aucune importance.
  Également, en envoyant la même donnée 3 fois, et même plus si on le désirait, mais ce serait inutile,
  on peut resynchroniser le bus de donnée;  en effet, en mode 8-bit, l'envoi d'une donnée nécessite un seul échange avec le LCD,
  tandis qu'en mode 4-bit il en faut deux, ce qui exige donc un ordonnancement des deux envois (bits de poids fort avant
  ceux de poids faible).  En revenant au mode 8-bit, temporairement faut-il le préciser, on peut être assuré que
  la synchronisation sera faite convenablement lorsqu'on passera au mode 4-bit.

  Les bits de poids fort à envoyer seront donc 0011;  les bits de poids faible sont sans importance car ils ne
  seront jamais envoyés;  on les met comme prévu ci-haut, soit 1000.
  Ainsi, il existe deux possibilités:
    - si le LCD était en mode 4-bit:
            - si le mode 4-bit est bien synchronisé, donc attendant des bits de poids fort en premier:
                    - 1er   envoi:  - envoi des bits de poids fort;
                    - 2ième envoi:  - envoi des bits de poids faible, mais l'effet est sans importance dans ce cas précis.
                    Le LCD est maintenant en mode 8-bit.
                    - 3ième envoi:  - les bits de poids fort et de poids faible arrivent en même temps;
                                    - les bits de poids fort forcent à nouveau le LCD à passer en mode 8-bit (redondance prévisible);
                                    - les bits de poids faible sont sans importance dans ce cas précis;
            - si le LCD était désynchronisé, donc attendant des bits de poids faible, la resynchronisation se ferait alors au
              2ième envoi, et le 3ième envoi servirait pour les bits de poids faible.
            Peu importe, dans les deux cas on réussit à passer en mode 8-bit, ce qui est le seul et unique but recherché.
    - si le LCD était déjà en mode 8-bit, pour les 3 envois l'effet sera identique:
                                    - les bits de poids fort et de poids faible arrivent en même temps;
                                    - les bits de poids fort forcent le LCD à passer en mode 8-bit (redondance prévisible);
                                    - les bits de poids faible sont sans importance dans ce cas précis.
  */

  // On force donc le LCD à passer en mode 8-bit, peu importe son état.
  // Le 2ième paramètre de la procédure LCD_SetAddress_WriteData permet d'envoyer
  // seulement les bits de poids fort, ou tous les 8 bits:
  //   - 4 signifie qu'on envoie seulement les 4 bits de poids fort;
  //   - 8 signifie qu'on envoie tous les 8 bits.
  // Dans le cas très particulier de l'initialisation du LCD, il faut envoyer seulement les 4 bits de poids fort.
  LCD_SetAddress_WriteData('I', 4, B00111000);    //  On force une 1ière fois...
  delay(6);
  LCD_SetAddress_WriteData('I', 4, B00111000);    //  On force une 2ième fois...
  delay(6);
  LCD_SetAddress_WriteData('I', 4, B00111000);    //  On force une 3ième fois..., question de couvrir tous les cas possibles

  // Le LCD est maintenant en mode 8-bit, comme lors d'un redémarrage (cold start).
  // On envoie maintenant l'instruction pour faire passer le LCD à 4-bit, en mettant le bit D4 à 0.
  // Encore une fois les bits de poids faible sont comme sans importance.
  LCD_SetAddress_WriteData('I', 4, B00101000);    //
  delay(6);

  /*
  On peut maintenant être assuré que l'ordonnancement exigé pour l'envoi des données
  sur une connexion avec 4 fils (4-bit) sera correct:
    - les 4 bits de poids fort dans un 1er temps, puis ensuite
    - les 4 bits de poids faible dans un 2ième temps.
  Le LCD étant dorénavant en mode 4-bit, on peut maintenant lui envoyer tous les 8 bits
  de l'instruction Function set;  ce sera un effet redondant pour des bits D5 et D4,
  mais pour les bits D3 et D2 ils pourront maintenant être pris en compte.
  */
  LCD_SetAddress_WriteData('I', 8, B00101000);
  delay(6);

  /*
  2ième étape de l'initialisation du LCD:  l'instruction Display on/off control

  Cette instruction permmet de mettre à on/off 3 éléments biens distincts:
    - l'afficheur;
    - le curseur;
    - l'affichage de tous les pixels d'une matrice (en fait, un rectangle vertical rempli), au-dessus du curseur;
      remarque:  ça fait "chargé" visuellement, si on utilise les deux, c'est-à-dire un curseur, plus ce rectangle plein.

  Configuration des bits D7 à D0:
    D7 = 0
    D6 = 0
    D5 = 0
    D4 = 0

    D3 = 1  bit-clé pour sélectionner l'instruction Display on/off control
    D2 = 1  D  Display on/off
    D1 = 1  C  Cursor on/off
    D0 = 0  B  Blinking of cursor position character  (on/off)

  On obtient donc 0000 1110, ou 0E en hexadécimal.
  */

  LCD_SetAddress_WriteData('I', 8, B00001110);

  /*
  3ième étape de l'initialisation du LCD:  l'instruction Entry mode set

  Cette instruction permet:
    - de sélectionner si le curseur (ou en plus le Blinking si activé, voir les lignes précédentes)
      ira vers la droite (Increment), ou vers la gauche (Decrement).
    - d'activer le display Shift:  le curseur ne bouge plus, et le "display" ou les caractères décalent
      vers la droite ou vers la gauche.

  Configuration des bits D7 à D0:
    D7 = 0
    D6 = 0
    D5 = 0
    D4 = 0

    D3 = 0
    D2 = 1  bit-clé pour sélectionner l'instruction Entry mode set
    D1 = 1  I/D  (Increment/Decrement)         I/D = 1:  Increment      I/D = 0:  Decrement
          Spécifie la direction du curseur (et/ou du Blinking s'il est activé):
            - increment, déplacement vers la droite;
            - decrement, déplacement vers la gauche.
    D0 = 1  Spécifie le display Shift         S = 1:  il y a display shift        S = 0:  il n'y a pas de display shift
          Remarque:  si le display Shift n'est pas activé:
                       - le curseur bouge
                       - le "display" ou l'ensemble des caractères ne bougent pas;
                     si le display Shift est activé:
                       - le curseur ne bouge pas;
                       - le display ou l'ensemble des caractères bougent.

  Explications plus claires:
  Le mieux est de faire les 4 tests, soit les 4 combinaisons des bits D2 et D1:
    - on se repositionne à chaque fois à l'adresse 5 (AC ou Address Counter passe à 5), mais
      sur l'affichage c'est la 6ième case (normal, la première case est à l'adresse logique 0);
    - on écrit 1, 2, et 3;
    - dans ces 4 exemples le curseur est à on (visible).
  1 - pas de Shift (bit D0 à 0), et I/D à 0 (Decrement)    01234567890123456
                                                                 _1
                                                                _21
                                                               _321

  2 - pas de Shift (bit D0 à 0), et I/D à 1 (Increment)    01234567890123456
                                                                  1_
                                                                  12_
                                                                  123_

  3 - Shift activé (bit D0 à 1),  et I/D à 0 (Decrement)   01234567890123456
                                                                  _1
                                                                  _21
                                                                  _321

  4 - Shift activé (bit D0 à 1),  et I/D à 1 (Increment)   01234567890123456
                                                                 1_
                                                                12_
                                                               123_

  Remarque:  si on place le curseur à off (et le Blinking s'il y a lieu), on obtient exactement le même affichage,
             moins les curseurs (montrés ici par des soulignés);  dans ce cas, l'exemple 4 peut être assimilé à
             l'écran d'une calculatrice.

  Pour un mode d'affichage "normal", comme on écrirait une ligne dans la vie courante:
    - désactiver le shift, et
    - mettre I/D à Incrément (progression vers la droite);
    Pour avoir un style "calculatrice", on peut prendre l'exemple 4, mais en enlevant le curseur.

  Remarque importante:  l'instruction Clear Display annule le I/D en le remettant systématiquement à Increment

  On obtient donc 0000 0110, ou 06 en hexadécimal.
  */
  LCD_SetAddress_WriteData('I', 8, B00000110);
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_SetAddress_WriteData(byte ParamID_DR, byte NombreDeBitsAEcrire, byte ParamValeur)
  {
  /*
  Cette procédure est assimilée aux instructions suivantes:
    - Write data to CG or DDRAM
    - Set CGRAM address
    - Set DDRAM address


  Cette procédure permet d'écrire dans l'un ou l'autre des deux registres accessibles du LCD, soit:
    - IR (Instruction Register aussi appelé command register), ou
    - DR (Data Register).

  Lorsque le DR (Data Register) est sélectionné, cette procédure correspond exactement
  à l'instruction Write data to CG or DDRAM.
  CGRAM:  Character Generator RAM;  les 8 espaces où l'on peut entreposer les dessins (patterns)
                                    de 8 caractères dessinés sur mesure
  DDRAM:  Data Display RAM;         les 80 cases mémoire où sont entreposés les caractères à afficher;
                                    minimalement il en faut 32 pour un écran de 16 x 2.

  Lorsque le IR (Instruction Register aussi appelé command register) est sélectionné, et que le bit 6 de la donnée
  est activé, cette procédure correspond exactement à l'instruction Set CGRAM address.

  Lorsque le IR (Instruction Register aussi appelé command register) est sélectionné, et que le bit 7 de la donnée
  est activé, cette procédure correspond exactement à l'instruction Set DDRAM address.

  Le 2ième paramètre de cette procédure permet d'envoyer seulement les bits de poids fort, ou tous les 8 bits:
    - 4 signifie qu'on envoie seulement les 4 bits de poids fort;
    - 8 signifie qu'on envoie tous les 8 bits.
  */

  // On initialise à chaque fois les 6 pins nécessaires.
  LCD_Initialiser6Pins_Write(ParamID_DR);

  // On envoie les 4 bits de poids fort, suivi d'un Enable.
  LCD_BitsPoidsFort_Write(ParamValeur);  LCD_Enable();

  if (NombreDeBitsAEcrire == 8)
    {
    // Au cas où on intercalerait un bout de code (déjà dans la procédure LCD_Enable, mais en commentaire)
    // pour tester le Busy Flag, on va répéter la réinitialisation des 6 pins nécessaires.
    LCD_Initialiser6Pins_Write(ParamID_DR);

    // On envoie les 4 bits de poids faible, si demandé, ce qui est habituellement le cas, suivi d'une Enable.
    // Juste pour rappel, la seule occasion où on ne les envoie pas est lors du cas très particulier de l'initialisation du LCD.
    LCD_BitsPoidsFaible_Write(ParamValeur);  LCD_Enable();
    }

  // Par sécurité, on remet immédiatement en INPUT les 4 pins de données de l'Arduino,
  // et même mieux, en INPUT_PULLUP, pour éviter que les entrées restent flottantes.
  pinMode(LCD_pin_D7, INPUT_PULLUP);  pinMode(LCD_pin_D6, INPUT_PULLUP);  pinMode(LCD_pin_D5, INPUT_PULLUP);  pinMode(LCD_pin_D4, INPUT_PULLUP);
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_Initialiser6Pins_Write(byte ParamID_DR)
  {
  // Cette procédure permet d'initialiser les 5 pins nécessaires pour envoyer des données au LCD
  // Ici il n'y a pas de paramètre pour la pin Read/Write;  elle est forcée à Write, ayant été
  // mise à la masse (ground) lors de l'assemblage des pièces.

  // On sélectionne le registre IR ou DR
  if (ParamID_DR == 'I')
    {
    digitalWrite(LCD_pin_RS, LOW);   // IR (instruction register aussi appelé command register) est sélectionné
    }
  else
   {
   digitalWrite(LCD_pin_RS, HIGH);  // DR (data register) est sélectionné
   }

  // On programme en OUTPUT les 4 pins de données
  pinMode(LCD_pin_D7, OUTPUT);  pinMode(LCD_pin_D6, OUTPUT);  pinMode(LCD_pin_D5, OUTPUT);  pinMode(LCD_pin_D4, OUTPUT);
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_BitsPoidsFort_Write(byte ParamValeur)
  {
  /*
  // Cette section en commentaire permettait de faire afficher dans le moniteur série exactement ce qui se passait
  int i;

  Serial.println("LCD_BitsPoidsFort_Write");
  Serial.print("ParamValeur BIN et HEX = ");
  // C'est la seule façon d'obtenir un affichage convenable (il manquait toujours tous les zéros à gauche).
  for (i = 7;  i >= 0;  i--)
    {
    Serial.print((ParamValeur >> i) & B00000001, HEX);
    if (i == 4) { Serial.print(" "); }
    }
  Serial.print("    ");
  // C'est la seule façon d'obtenir un affichage convenable (il manquait toujours tous le zéro à gauche).
  if (ParamValeur <= B00001111) { Serial.print("0"); }
  Serial.print(ParamValeur, HEX);  Serial.println();
  Serial.print("Ce qui sera ecrit:       ");
  for (i = 7;  i >= 4;  i--)
    {
    Serial.print(((ParamValeur >> i) & B00000001) & B00000001, HEX);
    }
  Serial.println();
  */

  // On envoie les 4 bits de poids fort
  digitalWrite(LCD_pin_D7, (ParamValeur >> 7) & B00000001);
  digitalWrite(LCD_pin_D6, (ParamValeur >> 6) & B00000001);
  digitalWrite(LCD_pin_D5, (ParamValeur >> 5) & B00000001);
  digitalWrite(LCD_pin_D4, (ParamValeur >> 4) & B00000001);
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_BitsPoidsFaible_Write(byte ParamValeur)
  {
  /*
  // Cette section en commentaire permettait de faire afficher dans le moniteur série exactement ce qui se passait
  int i;

  Serial.println("LCD_BitsPoidsFaible_Write");
  Serial.print("ParamValeur BIN et HEX = ");
  // C'est la seule façon d'obtenir un affichage convenable (il manquait toujours tous les zéros à gauche).
  for (i = 7;  i >= 0;  i--)
    {
    Serial.print((ParamValeur >> i) & B00000001, HEX);
    if (i == 4) { Serial.print(" "); }
    }
  Serial.print("    ");
  // C'est la seule façon d'obtenir un affichage convenable (il manquait toujours tous le zéro à gauche).
  if (ParamValeur <= B00001111) { Serial.print("0"); }
  Serial.print(ParamValeur, HEX);  Serial.println();
  Serial.print("Ce qui sera ecrit:            ");
  for (i = 3;  i >= 0;  i--)
    {
    Serial.print(((ParamValeur >> i) & B00000001) & B00000001, HEX);
    }
  Serial.println();
  */

  // On envoie les 4 bits de poids faible
  digitalWrite(LCD_pin_D7, (ParamValeur >> 3) & B00000001);
  digitalWrite(LCD_pin_D6, (ParamValeur >> 2) & B00000001);
  digitalWrite(LCD_pin_D5, (ParamValeur >> 1) & B00000001);
  digitalWrite(LCD_pin_D4, (ParamValeur >> 0) & B00000001);
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_Enable()
  {
  /*
  // Voir les explications ci-après.
  unsigned long microsecondes;
  unsigned long microsecondesApres;
  unsigned long microsecondesDiff;
  unsigned int Compteur = 0;
  boolean Succes = true;
  */

  // Pour l'envoi d'un enable, pas vraiment de commentaire
  // Les tests montrent qu'on peut descendre delayMicroseconds jusqu'à 1 microseconde sans problème.
  delayMicroseconds(5);
  digitalWrite(LCD_pin_Enable, LOW); delayMicroseconds(5);
  digitalWrite(LCD_pin_Enable, HIGH); delayMicroseconds(5);
  digitalWrite(LCD_pin_Enable, LOW); delayMicroseconds(5);

  /*
  // Section pour tester le Busy Flag.
  //
  // Cette section n'a pas donné les résultats escomptés...
  //
  // C'est à peine si j'ai pu obtenir au moins 1 boucle dans le while() dans les deux cas supposément
  // les plus voraces en temps dans le LCD, c'est-à-dire les intructions ClearDisplay et ReturnHome.
  // Le LCD est vraisemblablement plus véloce que les anciennes versions, ou les manuels retrouvés
  // au hasard sur Internet ne sont pas vraiment mis à jour.  Peu importe.
  // Mon LCD a été acheté en décembre 2015, et il était dans un Arduino Starter Kit.
  // Peut-être qu'il y aurait des résultats plus significatifs en programmant directement
  // en langage "assembler", mais je ne pousserai pas plus loin l'affaire.
  //
  // Pour le test, enlever delay(2) dans la procédure LCD_ClearDisplay;
  // Il faut aussi arrêter tout le processus immédiatement après l'appel
  // de la procédure LCD_ClearDisplay() dans setup(), sinon ça va générer trop de lignes
  // dans le moniteur série.
  // Il faut aussi enlever les commentaires des 5 variables ci-avant.
  //
  microsecondesAvant = micros();
  // Ne pas changer la séquence ci-après;  du point de vue de l'électronique c'est plus prudent.
  // On programme en INPUT les pins des données
  pinMode(LCD_pin_D7, INPUT);  pinMode(LCD_pin_D6, INPUT);  pinMode(LCD_pin_D5, INPUT);  pinMode(LCD_pin_D4, INPUT);
  digitalWrite(LCD_pin_RS, LOW);    // IR (instruction register  aussi appelé command register) de la pin Register Select est sélectionné.
  digitalWrite(LCD_pin_RW, HIGH);   // Read de la pin Read/Write est sélectionné.

  // On lance la vérification du Busy Flag
  while (Compteur < 1000)
    {
    digitalWrite(LCD_pin_Enable, HIGH);   delayMicroseconds(1);
    if (digitalRead(LCD_pin_D7) == HIGH)  // le LCD en encore en mode travail
      {
      digitalWrite(LCD_pin_Enable, LOW);  delayMicroseconds(1);
      Compteur++;
      if (Compteur == 1000)
        {
        // Ce n'est pas normal autant de boucles;  arrêter et examiner les résultats.
        Succes = false;
        break;
        }
      }
    else
      {
      digitalWrite(LCD_pin_Enable, LOW);  delayMicroseconds(1);
      break;        // ce qui provoque la fin de la procédure, avec un résultat normal
      }
    }
  microsecondesApres = micros();
  microsecondesDiff = microsecondesApres - microsecondesAvant;
  Serial.print("microsecondesAvant   :  ");  Serial.println(microsecondesAvant);
  Serial.print("microsecondesApres   :  ");  Serial.println(microsecondesApres);
  Serial.print("microsecondesDiff    :  ");  Serial.println(microsecondesDiff);
  Serial.print("Nombre de boucles--- :  ");   Serial.println(Compteur);
  // On remet Read/Write à Write;  les entrées de données du LCD redeviennent en input.
  digitalWrite(LCD_pin_RW, LOW);        // Write est sélectionné
  // Par sécurité, on remet immédiatement en INPUT les 4 pins de données de l'Arduino
  // et même mieux, en INPUT_PULLUP, pour éviter que les entrées restent flottantes.
  pinMode(LCD_pin_D7, INPUT_PULLUP);  pinMode(LCD_pin_D6, INPUT_PULLUP);  pinMode(LCD_pin_D5, INPUT_PULLUP);  pinMode(LCD_pin_D4, INPUT_PULLUP);
  */
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_ClearDisplay()
  {
  // Cette procédure est identique dans son effet à LCD_ReturnHome, sauf pour un effet:
  //   - les 80 cases de mémoire dans la DDRAM sont remplacées par des espaces blancs (20 hex),
  //     et donc l'écran est complètement effacé.
  //
  // Les autres effets sont:
  //   - le curseur est retourné en haut à gauche (il peut y avoir un curseur ou un "blanc clignotant")
  //     selon la configuration préalable;
  //   - le compteur d'adresse (AC ou Address Counter) passe à 0;
  //   - ID (Increment/Decrement) est forcé à 1, soit Increment mode;
  //
  // Pour l'instruction Clear display, la valeur à passer est 0000 0001, ou 1 en hexadécimal.
  LCD_SetAddress_WriteData('I', 8, B00000001);
  delay(2);
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_ReturnHome()
  {
  // Cette procédure est identique dans son effet à LCD_ClearDisplay, sauf pour un effet:
  //  -  les caractères demeurent tel quel dans la DDRAM, et demeurent donc affichés.
  //
  // Les autres effets sont:
  //   - le curseur est retourné en haut à gauche (il peut y avoir un curseur ou un "blanc clignotant")
  //     selon la configuration préalable;
  //   - le compteur d'adresse (AC ou Address  Counter) passe à 0;
  //   - ID (Increment/Decrement) est forcé à 1, soit Increment mode;
  //
  // Pour l'instruction Return home, la valeur à passer est 0000 0010, ou 2 en hexadécimal;
  // (mais pourrait aussi être 3, vu que le bit 0 est sans importance).
  LCD_SetAddress_WriteData('I', 8, B00000010);
  delay(2);
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_AfficherNombre(byte ParamAdresse, signed int ParamNombre, byte ParamBlancsAGauche)
  {
  // Cette procédure affiche un nombre sur le LCD.
  // Seule la variable globale de type chaîne byte Chaine001_ASCII est utilisée.

  // Le paramètre ParamAdresse permet de passer l'adresse où débute l'affichage du nombre;
  // sur un LCD 2 lignes par 16 caractères, les adresses sont: 1ière ligne:  0x00 à 0x0F
  //                                                           2ième ligne:  0x40 à 0x4F
  // Le paramètreParamNombre indique le monbre de chiffres à afficher;  cela peut servir lorsqu'il
  // est prévisible que le nombre sera petit.  Les seules valeurs admissibles sont 4, 3, 2, 1,
  // sachant que les milliers vont disparaître en premier, puis ensuite les centaines,
  // et les dizaines, mais jamais les unités.

  // Le paramètre ParamBlancsAGauche permet de remplacer les zéros à gauche par des blancs;
  // seule la valeur 'b' est valable, toute autre valeur laisse les zéros affichés.

  boolean EncoreDesZerosAGauche = true;

  int i;

  // On passe l'adresse au LCD
  ParamAdresse = ParamAdresse | B10000000;   // On met à 1 le bit 7, tous les autres demeurent inchangés.
  LCD_SetAddress_WriteData('I', 8, ParamAdresse);

  // On affiche sur le LCD les valeurs qui ont été placées dans la variable globale de type chaîne byte Chaine001_ASCII
  for (i = 4 - ParamNombre;  i < 4;  i++)
    {
    // On doit toujours laisser le chiffre des unités intact.  
    if (i < 3)  
      {
      // Si on doit effacer les zéros à gauche
      if (EncoreDesZerosAGauche == true && Chaine001_ASCII[i] == 48 && ParamBlancsAGauche == 'b')
        {  Chaine001_ASCII[i] = ' ';   }
      else
        {  EncoreDesZerosAGauche = false; }
      }  
    LCD_SetAddress_WriteData('D', 8, Chaine001_ASCII[i]);
    }
  }

/*-----------------------------------------------------------------------------------------------*/
void LCD_AfficherChaine(byte ParamAdresse, byte ParamValeur[])
  {
  // Cette procédure affiche une chaîne sur le LCD.
  //
  // Le paramètre ParamAdresse permet de passer l'adresse où débute l'affichage du nombre;
  // sur un LCD 2 lignes par 16 caractères, les adresses sont: 1ière ligne:  0x00 à 0x0F
  //                                                           2ième ligne:  0x40 à 0x4F
  // Le paramètre ParamValeur contient la chaîne passée dans un tableau de type byte, avec NULL (\0) à la fin.
  // Il n'y a pas de longueur maximale, en principe, mais comme le LCD peut afficher jusqu'à 16 caractères par ligne,
  // on devrait s'en tenir à 16, plus un caractère \0, ce qui fait 17.

  int i;

  // On passe l'adresse au LCD
  ParamAdresse = ParamAdresse | B10000000;   // On met à 1 le bit 7, tous les autres demeurent inchangés.
  LCD_SetAddress_WriteData('I', 8, ParamAdresse);

  // On affiche sur le LCD les valeurs qui ont été placées dans la variable tableau locale
  i = 0;  while (ParamValeur[i] != '\0')
    {  LCD_SetAddress_WriteData('D', 8, ParamValeur[i]);   i++;   }
  }


//Guylain - "Function overloading" pour fonction du meme nom
void LCD_AfficherChaine(byte ParamAdresse, const __FlashStringHelper * stringToDisplay ) { 
  char * myStringOfChar = (char *) calloc(MAX_LCD_CHARACTERS + 1, sizeof(char)); 
  strncpy_P(myStringOfChar,(const char PROGMEM *)stringToDisplay,MAX_LCD_CHARACTERS);

  #ifdef METEO_DEBUG     
    //DEBUG_PRINT(F("__FlashStringHelper * stringToDisplay="));
    //DEBUG_PRINTLN(stringToDisplay);
    //DEBUG_PRINT(F("myStringOfChar="));
    //DEBUG_PRINTLN(myStringOfChar);
  #endif
  
  LCD_AfficherChaine(ParamAdresse,myStringOfChar);
  free(myStringOfChar);
  
  return;
}

//Guylain - "Function overloading" pour fonction du meme nom
void LCD_AfficherChaine(byte ParamAdresse, const char * stringToDisplay ) { 
  
  byte myStringOfBytes[MAX_LCD_CHARACTERS + 1];
  int x = 0;
  const char * ci;

  //DEBUG_PRINTLN(strlen(stringToDisplay));
  
  //Transfert char to bytes  
  for (ci=stringToDisplay; *ci; ci++) {    
   if (ci == '\0') {
    break;
   }

   if (x >= (MAX_LCD_CHARACTERS)) {
    break;
   }
   
   myStringOfBytes[x] = ci[0];

   if (myStringOfBytes[x] == '{') { //Patch GP
    myStringOfBytes[x] = 0x01;     // image du 'é' créée dans le LCD
   }

   if (myStringOfBytes[x] == '}') { //Patch GP
    myStringOfBytes[x] = 0x02;     // image du 'è' créée dans le LCD
   }
   
   //DEBUG_PRINTLN(ci[0]);
   x++;
   //Serial.println(ci[0]);      
  }
  myStringOfBytes[x] = '\0';

  #ifdef METEO_DEBUG     
    //DEBUG_PRINT(F("stringToDisplay="));
    //DEBUG_PRINTLN(stringToDisplay);    
  #endif
        
  LCD_AfficherChaine(ParamAdresse,myStringOfBytes);

  return;
}

/*-----------------------------------------------------------------------------------------------*/
void LCD_CaracteresSurMesure()
  {
  /*
  La CGRAM (Character Generator RAM) est un autre espace mémoire du LCD, normalement prévu
  pour y placer des dessins ou patterns de nouveaux caractères librement dessinés à sa convenance,
  ou "sur mesure".  En anglais on dirait "customized", "user-defined", ou "homemade".
  Lorsqu'on utilise la matrice 5 x 8, on peut dessiner jusqu'à 8 nouveaux caractères;  sachant qu'il faut
  saisir pour chaque caractère 8 rangées de pixels, on en arrive à un total de 64 cases de mémoire de 8 bits.
  On peut aussi utiliser ces espaces comme espace mémoire additonnel pour l'Arduino.

  Les adresses sont sur 6 bits:  - 3 bits de poids fort pour l'adresse de chacun des 8 caractères, ce qui fait 8 possibilités;
                                 - 3 bits comme "sous-adresse" pour chacune des rangées de pixels, ce qui fait 8 possibilités;
  Configuration des bits D7 à D0 pour une utilisation comme paramètres de l'instruction Set CGRAM address
    D7 = 0
    D6 = 1  bit-clé pour sélectionner l'instruction Set GCRAM address
    D5 =    bit 2 de l'adresse dans la CGRAM
    D4 =    bit 1 de l'adresse dans la CGRAM
    D3 =    bit 0 de l'adresse dans la CGRAM
    D2 =    bit 2 du numéro de la rangée de pixels
    D1 =    bit 1 du numéro de la rangée de pixels
    D0 =    bit 0 du numéro de la rangée de pixels

  Ainsi, l'adressage se fait en 2 paquets de 3 bits chacun:
    - les bits D5 à D3 indique les adresses 00 à 07 de la CGRAM;
    - les bits D2 à D0 indique la rangée de pixels;  un dessin contient 8 rangées de pixels;
      la rangée 8 est normalement laissée à blanc, elle représente l'emplacement du curseur;

  Résumé de toutes les 64 (8 x 8) adresses possibles:
    pour dessiner un   1er nouveau caractère, qui sera à l'adresse 000 (sur 3 bits):
        pour la  0ième case de mémoire:     0 1 000 000 ce qui donne donc 01000000
        pour la  1ière case de mémoire:     0 1 000 001 ce qui donne donc 01000001
        pour la  2ième case de mémoire:     0 1 000 010 ce qui donne donc 01000010
        (...)
        pour la  7ième case de mémoire:     0 1 000 111 ce qui donne donc 01000111

    pour dessiner un 2ième nouveau caractère, qui sera à l'adresse 001 (sur 3 bits):
        pour la  8ième case de mémoire:     0 1 001 000 ce qui donne donc 01001000
        (...)
        pour la 15ième case de mémoire:     0 1 001 111 ce qui donne donc 01001111

    pour dessiner un 3ième nouveau caractère, qui sera à l'adresse 010 (sur 3 bits):
        pour la 16ième case de mémoire:     0 1 010 000 ce qui donne donc 01010000
        (...)
        pour la 23ième case de mémoire:     0 1 010 111 ce qui donne donc 01010111
    pour dessiner un 4ième nouveau caractère, qui sera à l'adresse 111 (sur 3 bits):
        (...)
    pour dessiner un 5ième nouveau caractère, qui sera à l'adresse 100 (sur 3 bits):
        (...)
    pour dessiner un 6ième nouveau caractère, qui sera à l'adresse 101 (sur 3 bits):
        (...)
    pour dessiner un 7ième nouveau caractère, qui sera à l'adresse 100 (sur 3 bits):
        (...)
    pour dessiner un 8ième nouveau caractère, qui sera à l'adresse 111 (sur 3 bits):
        (...)
    et ainsi de suite jusqu'à la dernière case diponible, soit la 63ième:
        pour la 63ième case de mémoire:    0 1 111 111 ce qui donne donc 01111111

  Tel que dit précédemment, le maximum de nouveaux caractères est de 8 lorsque la grille de 5 x 8 est utilisée.

  Pour une utilisation de ces nouveaux caractères sur l'écran, les valeurs à utiliser vont de 00 hex à 07 hex,
  tout comme on utiliserait par exemple 41 hexadécimal pour faire afficher le caractère A majuscule.

  Tel que dit dans la procédure LCD_SetAddress_WriteData:
    Lorsque le IR (Instruction Register aussi appelé command register) est sélectionné, et que le bit 6 de la donnée
    est activé, cette procédure correspond exactement à l'instruction Set CGRAM address.
  */

  // Dessin d'un nouveau caractères:  on va dessiner un é minuscule.
  // L'adresse sera 0x01 hexadécimal dans la grille;  il faudra donc l'appeler par 0x01, comme 
  // on appellerait A majuscule par 0x41, tel que retrouvé dans la grille ASCII
  // Remarque:  tel que dit ci-avant, on a le choix du numéro à assigner à ce nouveau caractère,
  // c'est-à-dire de 0 à 7.   Cependant, pour le présent logiciel, il faut éviter le 0 car
  // plusieurs chaînes se terminent par le caractère NULL ('\0'), et la programmation
  // est généralement construite en tenant compte de ce caractère de terminaison.
  
  LCD_SetAddress_WriteData('I', 8, B01001000);  LCD_SetAddress_WriteData('D', 8, B00000010);
  LCD_SetAddress_WriteData('I', 8, B01001001);  LCD_SetAddress_WriteData('D', 8, B00000100);
  LCD_SetAddress_WriteData('I', 8, B01001010);  LCD_SetAddress_WriteData('D', 8, B00001110);
  LCD_SetAddress_WriteData('I', 8, B01001011);  LCD_SetAddress_WriteData('D', 8, B00010001);
  LCD_SetAddress_WriteData('I', 8, B01001100);  LCD_SetAddress_WriteData('D', 8, B00011111);
  LCD_SetAddress_WriteData('I', 8, B01001101);  LCD_SetAddress_WriteData('D', 8, B00010000);
  LCD_SetAddress_WriteData('I', 8, B01001110);  LCD_SetAddress_WriteData('D', 8, B00001110);
  LCD_SetAddress_WriteData('I', 8, B01001111);  LCD_SetAddress_WriteData('D', 8, B00000000);
  // Si on s'amuse à enlever les 0, ça donne vraiment un é minuscule:                   1
  //                                                                                   1
  //                                                                                  111 
  //                                                                                 1   1
  //                                                                                 11111
  //                                                                                 1    
  //                                                                                  111 
  //                                                                              

  // Dessin d'un nouveau caractères:  on va dessiner un è minuscule.
  // L'adresse sera 0x02 hexadécimal dans la grille;  il faudra donc l'appeler par 0x02. 
  LCD_SetAddress_WriteData('I', 8, B01010000);  LCD_SetAddress_WriteData('D', 8, B00001000);
  LCD_SetAddress_WriteData('I', 8, B01010001);  LCD_SetAddress_WriteData('D', 8, B00000100);
  LCD_SetAddress_WriteData('I', 8, B01010010);  LCD_SetAddress_WriteData('D', 8, B00001110);
  LCD_SetAddress_WriteData('I', 8, B01010011);  LCD_SetAddress_WriteData('D', 8, B00010001);
  LCD_SetAddress_WriteData('I', 8, B01010100);  LCD_SetAddress_WriteData('D', 8, B00011111);
  LCD_SetAddress_WriteData('I', 8, B01010101);  LCD_SetAddress_WriteData('D', 8, B00010000);
  LCD_SetAddress_WriteData('I', 8, B01010110);  LCD_SetAddress_WriteData('D', 8, B00001110);
  LCD_SetAddress_WriteData('I', 8, B01010111);  LCD_SetAddress_WriteData('D', 8, B00000000);
  // Si on s'amuse à enlever les 0, ça donne vraiment un è minuscule:                 1  
  //                                                                                   1
  //                                                                                  111 
  //                                                                                 1   1
  //                                                                                 11111
  //                                                                                 1    
  //                                                                                  111 
  //                                                                              
  
  }

/*-----------------------------------------------------------------------------------------------*/
void Convertir_Integer_ASCII(unsigned int ParamInt)
  {
  // Cette procédure convertit un unsigned integer dans ses valeurs lisibles en ASCII, pour des nombres allant de 0 à 9999.
  // La valeur retournée est un tableau de 5 éléments de type byte, qui sont les codes ASCII,
  // obligatoirement dans la variable globale Chaine001_ASCII, qui a été déclarée comme ceci:  byte Chaine001_ASCII[5]
  // Les 4 premiers bytes sont pour la valeur retournée, et le 5ième est pour le null (\0).
  // On ne peut donc dépasser 9999.
  //
  //  Note:  pour le moment, il n'a pas été possible de retourner un tableau, en tout cas ça ne semble pas évident.
  //
  // Exemple:  435 sera retourné dans le tableau comme 48, 52, 51, 53, qui sont ses valeurs lisibles en ASCII.
  // Les zéros à gauche sont toujours inscrits, comme ici 0435.
  //
  // La procédure convertit au rythme d'environ 400 conversions par seconde, mais ça peut varier pour de petits nombres.

  unsigned int t[4] = { 48, 48, 48, 48};  // 48 (décimal) est le caractère 0 (zéro) dans la table ASCII
  unsigned int i;

  for (i = 0; i < ParamInt; i++)
    {
    switch (t[0])
      {
      case 57 :  t[0] = 58;   break;    // si = 9, deviendra :
      case 56 :  t[0] = 57;   break;    // si = 8, deviendra 9
      case 55 :  t[0] = 56;   break;    // si = 7, deviendra 8
      case 54 :  t[0] = 55;   break;    // si = 6, deviendra 7
      case 53 :  t[0] = 54;   break;    // si = 5, deviendra 6
      case 52 :  t[0] = 53;   break;    // si = 4, deviendra 5
      case 51 :  t[0] = 52;   break;    // si = 3, deviendra 4
      case 50 :  t[0] = 51;   break;    // si = 2, deviendra 3
      case 49 :  t[0] = 50;   break;    // si = 1, deviendra 2
      case 48 :  t[0] = 49;   break;    // si = 0, deviendra 1
      }
    if (t[0] == 58)           // si l'unité     est devenu   : , soit 9 + 1, il faut le ramener à zéro, et incrémenter la dizaine
      {
      t[0] = 48;              // l'unité     redevient à 0
      t[1] = t[1] + 1;        // la dizaine  est incrémentée de 1
      if (t[1] == 58)         // si la dizaine  est devenue  : , soit 9 + 1, il faut la ramener à zéro, et incrémenter la centaine
        {
        t[1] = 48;            // la dizaine  redevient à 0
        t[2] = t[2] + 1;      // la centaine est incrémentée de 1
        if (t[2] == 58)       // si la centaine est devenue  : , soit 9 + 1, il faut la ramener à zéro, et incrémenter le millier
          {
          t[2] = 48;          // la centaine redevient à 0
          t[3] = t[3] + 1;    // le millier  est incrémenté  de 1
          }
        }
      }
    }

  // On recopie les valeurs dans la variable globale de type chaine Chaine001_ASCII, et on fait un casting en byte.
  // On en profite pour remettre les chiffres dans le bon ordre.
  Chaine001_ASCII[0] = byte(t[3]);
  Chaine001_ASCII[1] = byte(t[2]);
  Chaine001_ASCII[2] = byte(t[1]);
  Chaine001_ASCII[3] = byte(t[0]);
  Chaine001_ASCII[4] = byte('\0');

  // Serial.print(ParamInt);  Serial.print("   ");
  // i = 0;  while (Chaine001_ASCII[i] != byte('\0'))  {  Serial.print(char(Chaine001_ASCII[i]));  Serial.print("  ");  i++;  }
  // Serial.println();
  }

/*-----------------------------------------------------------------------------------------------*/
void Clavier_Lire(int ParamAttenteSecondes)
  {
  /*  
  Cette procédure permet de lire un caractère sur un clavier 3 x 4 de style téléphone, via 
  un jeu de résistances qui octroie un lien avec le microcontrôleur via un seul fil de données.
  
  Après un délai (selon le paramètre ParamAttenteSecondes) sans avoir reçu de réponse de l'utilisateur,
  la procédure retourne 'P', signifiant 'Production'.
  
  Cependant, les valeurs '*' et '#' saisies par l'utilisateur forcent la procédure à demeurer
  en mode attente (idle) de manière permanente, jusqu'à ce qu'une des touches de '0' à '9' ait été pressée.
  
  Les valeurs  '0' à '9' saisies par l'utilisateur n'ont pas de signification particulière pour
  la présente procédure.

  Si la valeur passée en paramètre dépasse 99, elle est ramenée à 99.
  Il est permis de passer en paramètre la valeur 0.
  
  Pour pallier aux effets de rebonds mécaniques lorsque le contact mécanique d'une touche est fait,
  on va lire l'entrée analogique Clavier_pin 1000 fois consécutives, donc durant environ 1/10 de seconde
  (entre 112 et 113 millisecondes);  les 1000 valeurs lues doivent  être identiques à + ou - 2,
  compte tenu des effets du bit le moins significatif (least significant bit) dont la valeur
  est souvent aléatoire en pareilles circonstances.

  La boucle principale met un peu moins de 1/4 de seconde à s'exécuter.
  */  

  unsigned long u_long_ParamAttenteSecondes;

  unsigned long millisInitial;
  int millisReste;

  int ValeurLueInitiale;
  int ValeurLue;

  boolean AttentePermanente;

  int NbLectures;

  //byte Chaine_Clavier[] = "Clavier:  ";
  //byte Chaine_AsterisqueDiese[] = "(*#)";

  LCD_ClearDisplay();

  // Ce casting est absolument obligatoire, compte tenu que plus loin dans cette procédure on aura à faire
  // la multiplication u_long_ParamAttenteSecondes * 1000;  s'il n'y avait pas de casting, et qu'en plus
  // le produit dépassait 32767, donc plus grand que 32 secondes, le résultat de cette multiplication
  // donnerait un résultat curieux. 
  u_long_ParamAttenteSecondes = (unsigned long)ParamAttenteSecondes;

  // Si le paramètre est plus grand que 99, on le ramène à 99.
  if (u_long_ParamAttenteSecondes > 99)  {  u_long_ParamAttenteSecondes = 99;  }

  // On sélectionne le voltage de référence à 5 volts.
  analogReference(DEFAULT);
  // On fait quelques lectures, car l'aide en ligne dit que les premières lectures après
  // un changement de voltage de référence peuvent être moins précises.
  // Dans ce logiciel on s'en tient toujours à la valeur par défaut, c'est-à-dire 5 volts.
  for (NbLectures = 0;  NbLectures < 9;  NbLectures++)
    {  ValeurLue = analogRead(Clavier_pin);  delay(1);  }

  millisInitial = millis();
  AttentePermanente = false;
  
  while (true)
    {
    // Ce segment de code fait afficher sur le LCD:
    //   - si l'attente n'est pas encore devenue permanente:  le nombre de secondes qui restent avant d'atteindre zéro.
    //   - si l'attente est devenue permanente             :  les signes astérisque et dièse "(*#)";
    if (AttentePermanente == false) 
      {  
      LCD_AfficherChaine(0x40, F("Clavier:  "));
      Chaine001_ASCII[0] = '('; Chaine001_ASCII[1] = '\0';
      LCD_AfficherChaine(0x4C, Chaine001_ASCII);

      millisReste = u_long_ParamAttenteSecondes - (  (millis() - millisInitial) / 1000  );
      Convertir_Integer_ASCII(millisReste);
      LCD_AfficherNombre(0x4D, 2, 'b');

      Chaine001_ASCII[0] = ')'; Chaine001_ASCII[1] = '\0';
      LCD_AfficherChaine(0x4F, Chaine001_ASCII);
      }
    else
      {
      LCD_AfficherChaine(0x4C, F("(*#)"));
      }

    // Ce segment de code exécute une boucle à l'infini, à moins qu'il y ait
           // 1000 lectures consécutives donnant un résultat presque identique (+- 2).
    // Si aucune touche n'est pressée, la valeur lue est 0, parfois 1,
    // et les 1000 boucles sont valables dans ce cas également.
    while (true)     
      {
      ValeurLueInitiale = analogRead(Clavier_pin);
      NbLectures = 0;
      while(NbLectures < 1000)
        {
        ValeurLue = analogRead(Clavier_pin);
        // Comme le bit le moins significatif (least significant bit) est très difficile à maintenir
        // uniforme au fil des lectures, on accorde une marge de + ou - 2, ce qui est plus que suffisant.
        // Si la valeur lue est en dehors de cette marge, en principe ça ne devrait pas se produire,
        // alors le résultat ne pourra pas être valable, et on fait un break.
        if (ValeurLue > ValeurLueInitiale + 2 || ValeurLue < ValeurLueInitiale - 2) {  break;  }
        NbLectures++;
        }
      if (NbLectures == 1000) {  break;  }
      }

    // Ce segment de code détermine laquelle des touches a été pressée, si cela s'est produit.
    // Si aucune touche n'a été pressée, la valeur retournée par analogRead est de 0, parfois 1,
    // d'où la limite fixée empiriquement à 5.
    if (ValeurLue >=    0 && ValeurLue <=    5) {  ClavierCaractereLu[0] = ' ';  }  // Valeur lue en test:    0 -   1
    if (ValeurLue >=    6 && ValeurLue <=  174) {  ClavierCaractereLu[0] = '1';  }  // Valeur lue en test:  153 - 154 
    if (ValeurLue >=  175 && ValeurLue <=  223) {  ClavierCaractereLu[0] = '4';  }  // Valeur lue en test:  197
    if (ValeurLue >=  224 && ValeurLue <=  280) {  ClavierCaractereLu[0] = '2';  }  // Valeur lue en test:  249 - 250
    if (ValeurLue >=  281 && ValeurLue <=  325) {  ClavierCaractereLu[0] = '5';  }  // Valeur lue en test:  311 - 310
    if (ValeurLue >=  326 && ValeurLue <=  376) {  ClavierCaractereLu[0] = '3';  }  // Valeur lue en test:  340 - 341
    if (ValeurLue >=  377 && ValeurLue <=  479) {  ClavierCaractereLu[0] = '6';  }  // Valeur lue en test:  412 - 413
    if (ValeurLue >=  480 && ValeurLue <=  617) {  ClavierCaractereLu[0] = '7';  }  // Valeur lue en test:  544 - 545
    if (ValeurLue >=  618 && ValeurLue <=  735) {  ClavierCaractereLu[0] = '8';  }  // Valeur lue en test:  690 - 691
    if (ValeurLue >=  736 && ValeurLue <=  819) {  ClavierCaractereLu[0] = '9';  }  // Valeur lue en test:  780 - 781
    if (ValeurLue >=  820 && ValeurLue <=  892) {  ClavierCaractereLu[0] = '*';  }  // Valeur lue en test:  857 - 858
    if (ValeurLue >=  893 && ValeurLue <=  942) {  ClavierCaractereLu[0] = '0';  }  // Valeur lue en test:  926 - 927
    if (ValeurLue >=  943 && ValeurLue <= 1023) {  ClavierCaractereLu[0] = '#';  }  // Valeur lue en test:  957 - 958 - 959
    ClavierCaractereLu[1] = '\0';

    // Ce segment de code détermine que:
    //   - si on a pressé une des touches    * ou #, le processus d'attente (idle) va dorénavant
    //     se poursuivre indéfiniment jusqu'à ce qu'une touche de '0' à '9' soit pressée;
    //   - si on a pressé une des touches de 0  à 9, on quitte immédiatement (break).
    if (ClavierCaractereLu[0] == '*'  || ClavierCaractereLu[0] == '#')
      {   AttentePermanente = true;  }
    if (ClavierCaractereLu[0] >= '0'  && ClavierCaractereLu[0] <= '9')
      {  break;  }

    // Ce segment de code détermine que:
    //   - si l'attente N'est PAS devenue permanente, et que l'on vient de dépasser le délai passé en paramètre,
    //     alors on arrête tout, et la procédure retournera le caractère 'P' signifiant Production.
    if ( AttentePermanente == false &&  (millis() - millisInitial) > (u_long_ParamAttenteSecondes * 1000) )
      {
      // On fait afficher la seconde numéro zéro (ça paraît plus élégant).
      Convertir_Integer_ASCII(0);
      LCD_AfficherNombre(0x4D, 2, 'b');
      break;  
      }
    }
 
  // Conclusion de la procédure
  // Si aucune touche n'a été pressée, on retourne 'P' pour passer automatiquement en mode Production.
  // Autrement, ce sera un des caractères de '0' à '9'.
  if (ClavierCaractereLu[0] == ' ')
    {  ClavierCaractereLu[0] = 'P';  }

  // On affiche le caractère retourné, qui ne peut être que 'P', et '0' à '9'.
  LCD_AfficherChaine(0x4A, ClavierCaractereLu);
  // Serial.print("ClavierCaractereLu     :  ");   Serial.println(char(ClavierCaractereLu[0]));

  delay(1500);
  }
  
/*-----------------------------------------------------------------------------------------------*/
void Meteo_LireVentDirection(byte ParamProduction)
  {
  // Cette procédure permet de lire la direction du vent avec une girouette (Wind Vane).
  // Le module utilisé ici permet de lire jusqu'à 16 directions, mais mécaniquement 8 de ces directions
  // représentent chacune un angle d'environ 5 degrés, ce qui devient difficile à gérer et à interpréter.
  // Ces 8 directions ont donc été associées avec sa ou ses voisines, ce qui donne un angle beaucoup plus large;
  // cette façon de faire correspond également à la réalité sur le terrain où 8 directions sont suffisantes.

  int ValeurLue;

  int i;
  int j;
                          
  //byte Chaine_Girouet[] = "Girouette";
  //byte ParentheseGauche[] = "(";
  //byte ParentheseDroite[] = ")";
  byte VD[11];

  // Cette boucle tourne à l'infini.
  // Si ParamProduction = 'P' (pour Production), le while effectue une seule boucle.
  while(true)
    {
    LCD_ClearDisplay();
    LCD_AfficherChaine(0x00, F("Girouette"));
    
    // On sélectionne le voltage de référence à 5 volts.
    analogReference(DEFAULT);
    // On fait quelques lectures, car l'aide en ligne dit que les premières lectures après
    // un changement de voltage de référence peuvent être moins précises.
    // Dans ce logiciel on s'en tient toujours à la valeur par défaut, c'est-à-dire 5 volts.
    for (i = 0;  i < 9;  i++)
      {  ValeurLue = analogRead(Clavier_pin);  delay(1);  }

    /*
    // Cette section en commentaire a servi à établir les valeurs.
    //
    // Cette boucle en commentaire boucle 40 fois, et permet de prendre 20 mesures consécutives à chaque fois, puis
    // de les placer dans un tableau, plus un 2ième tableau du nombre de fois où une valeur est retrouvée.
    // l'expérience montre que:
    //   - parfois la valeur lue est retrourvée 20 fois dans le tableau, ce qui est donc le plus stable;
    //   - dans les autres cas, la valeur lue est retrouvée entre 15 à 18 fois, et les valeurs adjacentes
    //     sont toujours de +1 et de -1, comme par exemple:
    //       185     1 fois
    //       186    17 fois
    //       187     2 fois,  pour un total de 20 fois (ou 20 lectures).
    //  Le plus important est que les valeurs adjacentes ne dépassent jamais +1 et -1,
    //  ce qui est donc très rassurant, compte tenue des marges de manoeuvre.
    //
    // Ce test va durer environ 40 secondes, et il faut faire tourner la girouette à la main, très lentement,
    // afin d'obtenir l'ensemble des valeurs lues, soit 16.

    int j;
    int k;
    int TableauValeurLue[20];
    int TableauPresence[20];

    for (k = 0;  k < 40;  k++)
      {
      // On prend la mesure 20 fois
      for (i = 0;  i < 20;  i++) { TableauValeurLue[i] = 0;  TableauPresence[i] = 0;  }
      for (i = 0;  i < 20;  i++)
        {
        ValeurLue = analogRead(Senseur_pin_Girouette);
        delay(30);

        // Si la valeur lue est déjà présente dans le tableau TableauValeurLue,
        // on augmente simplement de +1 l'indice correspondant dans le tableau TableauPresence.
        // Si la valeur lue n'est pas déjà présente dans le tableau TableauValeurLue,
        // on l'ajoute, et on met à 1 l'indice correspondant dans le tableau TableauPresence.
        for (j = 0;  j < 20;  j++)
          {
          if (TableauValeurLue[j] == 0)
            {
            TableauValeurLue[j] = ValeurLue;
            TableauPresence[j] = 1;
            break;
            }
          else
            {
           if (TableauValeurLue[j] == ValeurLue)
             {
              TableauPresence[j]++;
              break;
              }
            }
          }
        }
      // On affiche le tableau des valeurs , et le nombre de fois retourvé pour chacune d'elles.
      Serial.print("-------------- boucle "); Serial.print(k);  Serial.println("  de 40");
      for (j = 0;  j < 20;  j++)
        {
        if (TableauValeurLue[j] == 0) { break; }
        Serial.print(TableauValeurLue[j]);  Serial.print("    "); Serial.println(TableauPresence[j]);
        }
      }
    */

    // On fait la lecture 10 fois, avec un délai de 10 millisecondes, même si on ne traite pas
    // les différences possibles entre les lectures;  les tests montrent que tout est très stable.
    for (i = 0;  i < 10;  i++)   {  ValeurLue = analogRead(Senseur_pin_Girouette);  delay(10); }

    VentDirection_int = METEO_VALUE_OUT_OF_BOUNDS;
    // Selon des relevés faits lors des tests, les valeurs suivantes seront utilisées:    
    if (ValeurLue >=    0 && ValeurLue <=  109) {  VentDirection[0] = 'S';   VentDirection[1] = 'O';  VentDirection_int = DIRECTION_VENT_SUD_OUEST;  }
    if (ValeurLue >=  110 && ValeurLue <=  214) {  VentDirection[0] = 'O';   VentDirection[1] = ' ';  VentDirection_int = DIRECTION_VENT_OUEST;      }
    if (ValeurLue >=  215 && ValeurLue <=  347) {  VentDirection[0] = 'N';   VentDirection[1] = 'O';  VentDirection_int = DIRECTION_VENT_NORD_OUEST; }
    if (ValeurLue >=  348 && ValeurLue <=  531) {  VentDirection[0] = 'S';   VentDirection[1] = ' ';  VentDirection_int = DIRECTION_VENT_SUD;        }
    if (ValeurLue >=  532 && ValeurLue <=  666) {  VentDirection[0] = 'N';   VentDirection[1] = ' ';  VentDirection_int = DIRECTION_VENT_NORD;       }
    if (ValeurLue >=  667 && ValeurLue <=  806) {  VentDirection[0] = 'S';   VentDirection[1] = 'E';  VentDirection_int = DIRECTION_VENT_SUD_EST;    }
    if (ValeurLue >=  807 && ValeurLue <=  916) {  VentDirection[0] = 'E';   VentDirection[1] = ' ';  VentDirection_int = DIRECTION_VENT_EST;        }
    if (ValeurLue >=  917 && ValeurLue <= 1024) {  VentDirection[0] = 'N';   VentDirection[1] = 'E';  VentDirection_int = DIRECTION_VENT_NORD_EST;   }
    VentDirection[2] = '\0';

    LCD_AfficherChaine(0x0A, F("("));
    Convertir_Integer_ASCII(ValeurLue);
    LCD_AfficherNombre(0x0B, 4, 'b');
    LCD_AfficherChaine(0x0F, F(")"));
    // LCD_AfficherChaine(0x43, VentDirection);

    // On affiche en clair la direction du vent sur le LCD
    i = 0;
    for (j = 0;  j < 2;  j++)
      {
      if (VentDirection[j] == 'S') {  VD[i] = 'S'; i++;  VD[i] = 'u'; i++;  VD[i] = 'd';  i++;  }
      if (VentDirection[j] == 'O') {  VD[i] = 'O'; i++;  VD[i] = 'u'; i++;  VD[i] = 'e';  i++;  VD[i] = 's'; i++;  VD[i] = 't';  i++;  }
      if (VentDirection[j] == 'N') {  VD[i] = 'N'; i++;  VD[i] = 'o'; i++;  VD[i] = 'r';  i++;  VD[i] = 'd'; i++;  }
      if (VentDirection[j] == 'E') {  VD[i] = 'E'; i++;  VD[i] = 's'; i++;  VD[i] = 't';  i++;  }

      if (VentDirection[1] == ' ')  {  break;  }    // cas où il n'y a qu'un point cardinal (donc pas composé de 2 points cardinaux)      
      if (j == 1)                   {  break;  }

      VD[i] = '-';  i++;  
      }
    VD[i] = '\0';
    LCD_AfficherChaine(0x43, VD);
    
    if (ParamProduction == 'P') {  break;  }
    if (ParamProduction == ' ') {  delay(3000);  }
    
    }

  // Serial.println("Girouette:");
  // Serial.print("ValeurLue:  ");   Serial.println(ValeurLue);
  // i = 0;  while (VentDirection[i] != '\0')  {  Serial.print(char(VentDirection[i]));  i++;  }
  // Serial.println();

  }

/*-----------------------------------------------------------------------------------------------*/
void Meteo_LireVentVitesse(byte ParamProduction)
  {
  // Cette procédure permet de lire le senseur de la vitesse du vent, ou anémomètre (anemometer).
  // Le fabricant dit qu'à chaque tour un contact est établi comme un aimant qui passe
  // devant un interrupteur: "a contact as a magnet moves past a switch".
  // Il précise qu'un vent de 1.492 mph ou 2.4 km/h produit 1 fermeture de l'interrupteur par seconde.
  // Toutefois les tests réels montrent qu'il semble y avoir deux micro-switches dans le dispositif,
  // ce qui fait qu'à chaque tour il y a 2  On et 2  Off qui sont lus sur la pin,
  // et donc 4 changements de signal.
  // Les tests montrent qu'à chaque tour:
  //   - Off revient deux fois, et à chaque fois il couvre environ 50 degrés;      50 degrés x 2 = 100 degrés
  //   - On  revient deux fois, et à chaque fois il couvre environ 130 degrés;    130 degrés x 2 = 260 degrés
  //   - ce qui donne 360 degrés en tout.                                                  Total = 360 degrés
  // Les tests montrent également qu'il n'y a pas de rebond mécanique lorsque le contact s'établit
  // ou s'enlève, ce qui permet d'alléger le présent code.
  //
  // Ici il a été prévu que la lecture se fasse durant 3 secondes, ou 3000 millisecondes,
  // et on compte le nombre de changements de On à Off, et de Off à On.
  // La question est de savoir si on doit tenir compte de ce que dit le fabricant,
  // ou de la réalité:  il faudra faire des tests sur le terrain pour savoir s'il
  // faut diviser le résultat pas 2, vu qu'il semble y avoir 2 micro-switches au lieu de 1 comme annoncé.
  //
  // Pour des calculs plus précis 1.492 mph = 2.401 km/h.

  unsigned int EtatPrecedent;
  unsigned int EtatActuel;
  unsigned int NombreChangements;

  unsigned long millisInitial;
  // unsigned long millisArret;
  unsigned long NombreBouclesWhile;

  //int VentVitesse_int; //Je l'ai declaree globalement. GP
  float VentVitesse_float;

  int i;

  //byte Chaine_Off[] = "Off";
  //byte Chaine_On[] = "On ";
  //byte Chaine_Anemom[] = "AnXmomXtre";  Chaine_Anemom[2] = 0x01;     // image du 'é' créée dans le LCD
  //                                      Chaine_Anemom[6] = 0x02;     // image du 'è' créée dans le LCD
  //byte Chaine_km_h[] = "km/h";

  // Cette boucle tourne à l'infini.
  // Si ParamProduction = 'P' (pour Production), le while effectue une seule boucle.
  while(true)
    {
    LCD_ClearDisplay();

    // On initialise la pin Senseur_pin_Anemometre (voir constante au début).
    // On va se servir de l'option INPUT_PULLUP pour maintenir l'entrée digitale à HIGH en l'absence
    // de contact, ce qui évite de surcroît d'avoir à utiliser une résistance externe
    // pour faire exactement le même travail.
    pinMode(Senseur_pin_Anemometre, INPUT_PULLUP);

    // On affiche "Anémomètre" sur le LCD
    LCD_AfficherChaine(0x00, F("An{mom}tre"));

    // On note la valeur de millis() au début, et on fait quelques initialisations
    millisInitial = millis();
    NombreChangements = 0;
    NombreBouclesWhile = 0;
   
    // La commande digitalRead retourne HIGH ou LOW.  Électroniquement, lorsqu'une micro-switch est à On,
    // dans le présent cas, la pin est forcée au ground, ou à 0 volt, ce qui évidememnt donne un LOW en lecture,
    // ce qui peut être mélangeant car instinctivement on assimile On à HIGH.  Lorsque le contact s'enlève, la
    // pin remonte à 5 volts, du à l'effet INPUT_PULLUP (voir explications ci-haut).
    // Comme ici on calcule les changements d'état ou de signal, finalement ça n'a pas d'importance.
  
    // On détermine l'état initial de la pin
    if (digitalRead(Senseur_pin_Anemometre) == HIGH) {  EtatPrecedent = 1;  } else  {  EtatPrecedent = 0;  }

    // On boucle durant 3 secondes, ou 3000 millisecondes,
    // ce qui représente environ 1531 boucles du while ci-après;  on place une sécurité à 3000
    while (NombreBouclesWhile < 3000)
      {
      if (digitalRead(Senseur_pin_Anemometre) == HIGH) {   EtatActuel    = 1;  } else  {  EtatActuel    = 0;  }
      switch (EtatActuel)
        {
        case 1:
          if (EtatPrecedent == 1)  {  }
          else
            {
            EtatPrecedent = 1;
            NombreChangements++;
            }
          break;
        case 0:
          if (EtatPrecedent == 1)
            {
           EtatPrecedent = 0;
            NombreChangements++;
            }
          else  {  }
          break;
        }

      // Une décision a été prise;  les 2 appels de procédure ci-après grugent beaucoup de millisecondes.  
      // Sans ces 2 lignes, on effectue environ 420000 boucles durant 3 secondes, ce qui est 
      // plus que confortable pour ne pas rater les changements d'état des micro-switches.
      // Avec ces 2 appels de procédures, le nombre de boucles descend à 1531 durant 3 secondes,
      // ce qui fait 510 boucles à la seconde (1531 / 3);  considérant qu'il y a 4 changements pas tour,
      // à partir de 127 tours secondes (510 / 4), on peut risquer d'échapper des changements d'état.
      // Le fabricant dit que 1 fermeture d'interrupteur (le texte est très ambigu) donne
      // une vitesse de vent de 2.401 km/h, ce qui signifie qu'à partir de 304,9 km/h (127 x 2.401),
      // on risque d'avoir une mesure imprécise.  Comme le cas est plustôt extrême, j'ai décidé
      // de conserver les 2 appels de procédure ci-après. S'il y a plutôt deux micro-switches dans le dispositif,
      // alors soit que la mesure devient imprécise à partir de 152.4 km/h, ou à partir de 609.8 km/h,
      // ce qui est extrême là encore.
      //
      // C'est juste que quant'à avoir un afficheur LCD, autant en profiter pour afficher
      // en temps réel ce qui se passe.
      Convertir_Integer_ASCII(NombreChangements);
      // Exceptionnellement, on va se limiter à 3 chiffres affichés, ce qui est amplement suffisant
      LCD_AfficherNombre(0x40, 3, 'b');
      if (EtatActuel == 1) {  LCD_AfficherChaine(0x44, F("Off"));  }
      if (EtatActuel == 0) {  LCD_AfficherChaine(0x44, F("On "));  }
  
      /*
      // Le segment de code ci-après est très difficile à tester, compte tenu que la situation se produit
      // à tous les 50 jours (voir aide en ligne de millis();  pour le moment, il est mis de côté
      // et une sécurité a été mise au niveau du While avec NombreBouclesWhile.  L'erreur possible
      // sera que la vitesse peut être augmentée erronnément de quelques pourcents dans cette situation précise,
      // qui peut se reproduire une fois tous les 50 jours.

      if (millis() < millisInitial)
        {
        // On vient de passer le cap des 49.9 jours étant donné que millis vient tout juste de repasser par zéro,
        // et avait donc atteint son maximum de unsigned long, soit 2 exposant 32 ou 4,294,967,296.
        // On note la différence d'avec ce maximum, et on repart de zéro,
        // tout en donnant une nouvelle valeur plus réduite à millisArret.
        millisArret = 3000 - (4294967296 - millisInitial);
        millisInitial = 0;
        delay(2);  // comme le microcontrôleur effectue environ 120531 boucles à la seconde, il est
                   // préférable de le zéro soit passé avec certitude
        }
      */
    
      if (  (millis() - millisInitial)  > 3000)  {  break;  }
      NombreBouclesWhile++;
      //  Serial.print("NombreBouclesWhile  :  ");   Serial.println(NombreBouclesWhile);
      }
    /*
    if (NombreBouclesWhile >= 3000)
      {
      // Serial.println("----  ERREUR  ----  ERREUR  ----");
      // Serial.println("Nombre de boucles excessif;  procédure interrompue.");
      }
    else
      {
      // Serial.println("--------------------------------");
      // Serial.print("millisInitial     :  ");   Serial.println(millisInitial);
      // Serial.print("millisMaintenant  :  ");   Serial.println(millisMaintenant);
      // Serial.print("NombreBouclesWhile:  ");   Serial.println(NombreBouclesWhile);
      // Serial.print("NombreChangements :  ");   Serial.println(NombreChangements);
      }
    */
    // Le texte du fabricant spécifie:
    //   "The cup-type anemometer measures wind speed by closing a contact as a magnet moves past a switch.
    //    A wind speed of 1.492 MPH (2.4 km/h) causes the switch to close once per second."
    // J'ai détecté deux switches, tel que décrit ci-haut.
    // Le fabricant ne dit pas qu'il n'y a qu'une seule switch...
    // Le fabricant dit qu'un vent de 2.401 km/h provoque la fermeture de la switch une fois chaque seconde.
    // Logiquement, s'il y a 2 switches comme les tests l'ont montré, on aura donc une vitesse de 4.802 km/h
    //   pour chaque 2 fermetures de switch à chaque seconde.
    // Ici on utilise les changements d'état, et il y en a 4 par tour, on peut donc dire qu'à chaque
    //   groupe de 4 changements par seconde, on a une vitesse de 4.802 km/h.
    // Comme on détecte sur une période de 3 secondes, il faudra 12 changements par seconde
    //   pour avoir cette même vitesse de 4.802 km/h.
    // Il faut tout calculer sur cette base de 3 secondes.
    // Donc si 12 changements donnent 4.802 km/h, x changements donneront v vitesse...................
    //              À compléter dans le champ...................
  
    // Seul le nombre ci-après 4.801 est ambigu............  et peut-être aussi 12.....
    // Le nombre de changements est absolu quant'à lui, car il donne exactement ce qui a été relevé sur la pin digitale.
    VentVitesse_float = (NombreChangements * 4.801) / 12;

    // L'aide en ligne de "division" dit de caster pour convertir:
    //   "Use the cast operator e.g. (int)myFloat to convert one variable type to another on the fly".
    VentVitesse_int = int(VentVitesse_float);

    // On dépose le résultat dans la variable globale de type chaine byte VentVitesse.
    // On recopie seulement les 4 derniers caractères de la chaîne Chaine001_ASCII
    // vers les 4 caractères de la chaîne VentVitesse (vu qu'on a besoin de seulement 3 chiffres, plus le null).
    Convertir_Integer_ASCII(VentVitesse_int);
    for (i = 0;  i < 4;  i++)  {  VentVitesse[i] = Chaine001_ASCII[i + 1];  }

    // On fait afficher la vitesse du vent sur le LCD
    // Exceptionnellement, on va se limiter à 3 chiffres affichés, ce qui est amplement suffisant
    LCD_AfficherNombre(0x48, 3, 'b');
    LCD_AfficherChaine(0x4C, F("Km/h"));
  
    if (ParamProduction == 'P') {  break;  }
    if (ParamProduction == ' ') {  delay(3000);  }

    }

  // Serial.print("Nombre changements:  ");   Serial.println(NombreChangements);
  // Serial.print("VentVitesse_float :  ");   Serial.println(VentVitesse_float);
  // Serial.print("VentVitesse_int   :  ");   Serial.println(VentVitesse_int);
  ///Serial.print("VentVitesse_chaine:  ");
  // i = 0;  while (VentVitesse[i] != '\0')  {  Serial.print(char(VentVitesse_chaine[i]));  i++;  }
  // Serial.println();  Serial.println();
    
  }

/*************************************************************************************************/
/********  Procédures propres au senseur de températuee DS18B20  *********************************/

void Temperature_Principal_01(byte ParamProduction)
  {
  int essai_Temperature;

  // Cette boucle tourne à l'infini.
  // Si ParamProduction = 'P' (pour Production), le while effectue une seule boucle.
  while(true)
    {
    // On fait jusqu'à 3 essais, au cas où il y aurait une erreur de transmission de la part du DS18B20
    for (essai_Temperature = 0;  essai_Temperature < 3;  essai_Temperature++)
      {  if (Temperature_Principal_02() == true)   {   break;   }  }
  
    if (ParamProduction == 'P') {  break;  }
    if (ParamProduction == ' ') {  delay(3000);  }
  
    }    
  }
  
/*-----------------------------------------------------------------------------------------------*/
boolean Temperature_Principal_02()
  {
  /*
  Cette procédure concerne la sonde thermique digitale étanche à l'eau, ou senseur DS18B20.
  (Waterproof Digital Thermal Probe or Sensor DS18B20).  Elle est digitale parce qu'elle utilise
  la puce appelée "thermomètre digital programmable DS18B20", ou simplement DS18B20.
  La puce contient de la ROM, de l'EEPROM, et de la SRAM.

  Dans le cadre de la présente procédure, la configuration par défaut sera utilisée, et donc
  aucune programmation sera exercée sur la puce.

  La température retournée par la puce est une valeur integer signée de 12 bits, soit 4096 possibilités.
  Dans ce cas, le pas est de 0.0625 degré Celsius (1/16 de degré).
  Comme il s'agit d'une valeur signée, il y a 2048 possibilités en + ou en -, et donc
  les températures pourront s'étendre, théoriquement, de -128 degrés à +128 degrés;
  mais en pratique:
    - le senseur retourne des valeurs allant de -55 degrés à + 125 degrés;
    - la précision est de  +- 0.5 degré dans la zone qui va de -10 degrés à +85 degrés;
    - en dehors de cette zone le senseur perd un peu en précision.

  Le senseur peut retourner deux paquets de données:
    - un numéro d'identifiant unique sur 64 bits (8 octets), gravé en usine;
    - une série de données variables sur 72 bits (9 octets), comprenant entre autres la température.

  Le senseur n'a qu'une ligne de données (d'où le nom de la librairie qui aurait pu
  être utilisée, soit OneWire); ci-après on parlera tout simplement de "la ligne".

  Comme les données sont véhiculées en série, dans des "paquets" de 64 bits ou de 72 bits,
  le senseur ajoute lui-même un CRC (Cyclic Redundanty Check), qui est un chiffre auto-vérificateur
  calculé suivant une formule utilisant abondamment les opérateurs de type bitwise.
  L'hôte, dans ce cas-ci l'Arduino, peut refaire exactement les mêmes calculs, et ainsi démontrer
  que la transmission de données a été effectuée convenablement (error free), ou non.
  Dans la présente utilisation, le CRC sera systématiquement utilisé et recalculé,
  pour être sûr qu'à chaque fois les transmissions ont été correctement effectuées (error free).

  Toute transmission se fait de façon très individuelle, bit par bit, même si dans les faits
  on procède toujours par paquet 8 bits, c'est-à-dire un octet, que ce soit en lecture
  ou en écriture.  La transmission d'un bit est toujours initiée par le microcontrôleur
  en mettant la ligne à LOW, pour un temps déterminé selon la circonstance.
  Le fabricant utilise l'expression "time slot", que l'on peut traduire par "créneau de temps",
  pour nommer ces transmissions, et un "time slot" peut être caractérisé comme ceci:
    - «1 "time slot" carries 1 bit», c'est-à-dire «1 "time slot" transporte 1 bit»;
    - un "write time slot" transporte 1 bit du microcontrôleur au DS18B20:
        - il y a le "write 1" qui transporte un 1;
        - il y a le "write 0" qui transporte un 0;
        - ce sont essentiellement les "timings" qui font la différence, et
          le DS18B20 est alors capable d'échantillonner (sample) pour déterminer un 1 ou un 0;
        - chaque "write time slot" doit durer au minimum 60 μs, avec un temps de
          de rétablissement (recovery time) d'un minimum de 1 μs entre chaque slot;
        - chaque "write time slot" est initié par le microcontrôleur qui met
          la ligne de données à LOW pour au moins 3 μs;
        - un "write 1" est écrit:
            - en "relâchant" immédiatement la ligne;
            - on attend un temps additionnel de 57 μs;
        - un "write 0" est écrit:
            - en maintenant la ligne à LOW pour un temps additionnel de 57 μs;
            - en "relâchant" ensuite la ligne;
        - on ajoute un "recovery time" de 1 μs, ce qui fait un total de 61 μs.
        - du côté du DS18B20, on sait qu'il échantillonne (samples) la ligne durant une fenêtre qui dure
          de 15 μs à 60 μs après que le microcontrôleur ait initié un "write time slot":
            -  si la ligne est à HIGH durant l'échantillonnage, 1 est écrit;
            -  si la ligne est à LOW  durant l'échantillonnage, 0 est écrit;

      Remarque sur le mot "relâchant":  dans le contexte, le microcontrôleur remet sa pin en INPUT ou
                                        en INPUT_PULLUP, ce qui a pour effet de laisser la ligne "flottante"
                                        car ni lui ni le DS18B20 effectue une action sur la ligne;  mais
                                        grâce à la résistance pull-up, la ligne remonte à HIGH.

    - un "read time slot" transporte 1 bit du DS18B20 au microcontrôleur:
        - il n'y a pas de "read 1" ou de "read 0", car du côté du DS18B20 on ne sait comment
          les choses se déroulent;  c'est plutôt au microcontroleur de déterminer s'il s'agit d'un 1 ou d'un 0;
        - ce sont essentiellement les "timings" qui font la différence.
        - chaque "read time slot" doit durer au minimum 60 μs, avec un temps de
          de rétablissement (recovery time) d'un minimum de 1 μs entre chaque slot;
        - chaque "read time slot" est initié par le microcontrôleur qui met
          la ligne de données à LOW pour au moins 3 μs;
        - ensuite la ligne est "relâchée" par le microcontrôleur, afin que le
          DS18B20 puis commencer à émettre ses bits;
        - la donnée est valide pour un maximum de 15 μs après le front descendant
          qui a initié la "read time slot";  par conséquent, le microcontrôleur
          doit récupérer la donnée à l'intérieur de ce délai;
        - les tests montrent que la donnée 0 ne reste pas plus de 16 μs, parfois 20 μs, sur la ligne,
          cette gradation étant due au fait que la ligne de code suivante a été utilisée
            while (digitalRead(Senseur_pin_Temperature) == LOW);
          pour calculer le temps, et qu'elle requiert manifestement 4 μs;
          ceci confirme le 15 μs mentionné au paragrapge précédent;
        - du côté du DS18B20:
            - il écrit un 1 en laissant la ligne à HIGH;
            - il écrit un 0 en mettant  la ligne à LOW;
            - si ce n'est pas déjà fait, il relâche la ligne à la fin du "read time slot";

  Valeur retournée:  la variable globale Temperature.

  En cas d'erreur:  la procédure retourne la variable globale Temperature, mais avec un contenu
                    entre "+999.9999" et "+999.9980", selon la liste suivante:
                      - "+999.9999" :  la procédure actuelle a été interrompue dès le départ, et
                                       aucune des sous-procédures n'a été appelée;
                      - "+999.9998" :  le DS18B20 est débranché, absent ou brisé;
                      - "+999.9997" :  le numéro unique du DS18B20 n'a pas pu être lu convenablement;
                      - "+999.9996" :  les données de la SRAM n'ont pas pu être lues convenablement;
                      - "+999.9995" :  Erreur a la demande de reset (1) au DS18B20
                      - "+999.9994" :  Erreur pour lire les données de la ROM
                      - "+999.9993" :  Erreur a la demande de reset (2) au DS18B20
                      - "+999.9992" :  Erreur - délai MILLIsecondes hors normes
                      - "+999.9991" :  Erreur a la demande de reset (3) au DS18B20
                      - "+999.9990" :  Erreur pour lire les données de la SRAM
                      - "+999.9989" :  message à venir.
                      - "+999.9988" :  message à venir.
                      - "+999.9987" :  message à venir.
                      - ...
  */

  unsigned long Debut_millis;
  unsigned long Fin_millis;
  unsigned long Diff_millis;

  int i;

  //byte Etiquette_Temperature[] = "TempXrature";  Etiquette_Temperature[4] = 0x01;     // image du 'é' créée dans le LCD
  
  // On initialise la variable globale Temperature à une valeur équivalant à une erreur, soit "+999.9999".
  Temperature[0] = '+';
  Temperature[1] = '9';  Temperature[2] = '9';  Temperature[3] = '9';
  Temperature[4] = '.';
  Temperature[5] = '9';  Temperature[6] = '9';  Temperature[7] = '9';  Temperature[8] = '9';
  Temperature[9] = '\0';

  LCD_ClearDisplay();
  LCD_AfficherChaine(0x00, F("Temp{rature"));

  //-----------------  1ière partie  -  vérification de la présence d'un DS18B20
  //-----------------                -  vérification de la transmission
   
  // On doit initialiser le DS18B20 en lui demandant un reset            
  // Serial.print("Demande (1) de reset au DS18B20...  ");
  if (Temperature_Reset() == false) 
    {
    // Serial.println("Erreur a la demande de reset (1) au DS18B20");
    Temperature[7] = '9';  Temperature[8] = '5';  
    LCD_AfficherChaine(0x42, Temperature);
    return false;
    }
  
  // On lit le numéro unique gravé en usine, le "64-bit lasered ROM code"
  //
  // Chaque DS18B20 contient un numéro unique gravé en usine, et mémorisé dans une ROM de 8 octets,
  // pour un total de 64 bits;  dans la fiche technique on parle de "64-bit lasered ROM code".
  // Ces 64 bits, ou 8 octets, se divisent en 3 parties, et c'est-à-dire,
  // en partant des bits les moins significatifs:
  //   -  8 bits pour le numéro de modèle;  dans le présent cas, il s'agit de 28 hexa;
  //   - 48 bits pour le numéro de série, qui sert principalement à identifier la pièce
  //     dans le cas où il y en aurait plusieurs placées en parallèle sur la ligne de données;
  //   -  8 bits pour le CRC pour vérifier les 56 bits précédents (8 + 48 = 56).
  //
  // Dans le présent cas, il n'est absolument pas utile de connaître ces informations, car le DS18B20
  // est seul sur la ligne de données;  mais comme le fabricant a déjà gravé le CRC dans la ROM,
  // on va donc se servir de toute cette information pour vérifier qu'une transmission
  // peut s'effectuer convenablement (error free), en comparant le CRC transmis, avec celui
  // calculé après la transmission par la procédure Temperature_CRC8.

  // On passe la commande Read ROM, qui a pour numéro 0x33
  // Exceptionnellement, il n'est pas requis d'utiliser au préalable Skip ROM (voir plus loin)
  // pour cette commande précise.
  Temperature_EnvoyerCode(0x33);

  // On effectue la lecture des 64 bits, ou 8 octets, et on calcule immédiatement le CRC
  if (Temperature_LireDonnees(8) == false)
    {
    // Serial.println("Erreur pour lire les données de la ROM");
    Temperature[7] = '9';  Temperature[8] = '4';  
    LCD_AfficherChaine(0x42, Temperature);
    return false;
    } 
    
  // On détruit ces données immédiatement au cas où ultérieurement des lectures erronnées surviendraient,
  // et que d'anciennes données seraient réutilisées.
  for (i = 0;  i < 9;  i++)
    {  donnees_temp[i] =  i * i;  }  //  N'importe quoi comme valeur!  Et le CRC est erronné (vérifié).
  
  // Remarque sur les mots "commande" et "fonction" qui sont utilisé ci-après.
  //   - le mot "commande" permet de connaître ou de cibler une ou plusieurs unités de type DS18B20, ou autres,
  //     présentes sur la ligne, via leur numéro unique gravé en ROM;
  //   - le mot "fonction" permet de faire exécuter un tâche au DS18B20, mais 
  //     seulement après qu'une unité ait été bien identifiée via son numéro unique;
  //     on peut cependant s'adresser à toutes les unités sans distinction
  //     via Skip ROM (expliqué un peu plus loin).
  // Également, le manuel du fabricant précise qu'il faut toujours répéter les 3 étapes suivantes
  // pour que les choses se passent bien:
  //   1- initialisation (reset);
  //   2- commande ROM (qui peut être suivi du numéro unique, mais non utilisé ici); 
  //   3- fonction (tâche à exécuter).

  //-----------------  2ième partie  -  demande de conversion de la température en binaire
  //-----------------                
  // Pour que le DS18B20 puisse faire une lecture de son sensor, en effectuer ensuite la 
  // conversion en binaire, et finalement déposer la valeur dans sa SRAM,
  // il doit recevoir la fonction Convert T (0x44).
  // Pour que cette fonction soit bien prise en compte, il faut absolument qu'au préalable
  // le DS18B20 ait reçu une confirmation que la commande s'adresse bien à lui, individuellement,
  // et donc il faut envoyer au préalable la commande Match ROM (0x55),
  // suivie de son numéro unique de 64 bits.
  // Toutefois, on peut éviter cette lourdeur en utilisant une autre commande, Skip ROM (0xCC),
  // qui ordonne à toute unité qui pourrait être présente sur la ligne d'effectuer la tâche
  // dont le code suivra tout de suite après;  c'est ce qui sera fait ici.
  
  // On doit initialiser à nouveau le DS18B20 en lui demandant un reset
  if (Temperature_Reset() == false)
    {
    // Serial.println("Erreur a la demande de reset (2) au DS18B20");
    Temperature[7] = '9';  Temperature[8] = '3';  
    LCD_AfficherChaine(0x42, Temperature);
    return false;
    }

  // On passe la commande Skip ROM, qui a pour numéro 0xCC
  Temperature_EnvoyerCode(0xCC);

  // On lance la commande Convert T pour que le DS18B20 lise la température, et
  // la convertisse en 12 bits;  le temps d'attente doit être de 750 ms (millisecondes) au minimum.
  // On passe la commande Convert T qui a pour numéro 0x44
  Temperature_EnvoyerCode(0x44);

  // On doit attendre au minimum 750 ms (MILLIseconde, exceptionnellement).
  // Cependant on peut vérifier la progression de la conversion en observant la ligne:
  //   - si elle est à 0, la conversion est en progression  
  //   - si elle est à 1, la conversion est terminée
  // Les tests montrent que le délai est en réalité de 658 à 660 ms (milliseconde, exceptionnellement).
  Debut_millis = millis();
  while ( (millis() - Debut_millis) < 3000)
    {
    // noInterrupts();
    pinMode(Senseur_pin_Temperature, OUTPUT);             // On initialise la pin en mode OUTPUT
    digitalWrite(Senseur_pin_Temperature, LOW);           // On place la pin à LOW
    pinMode(Senseur_pin_Temperature, INPUT_PULLUP);       // On relâche la pin;  la ligne remonte à HIGH
    delayMicroseconds(9);                                 // On attend  9 μs (pour respecter "read time slot")
    if (digitalRead(Senseur_pin_Temperature) == LOW)
      {   }
    else
      {  break;  }
    delayMicroseconds(55);                                // On attend 55 μs (pour respecter "read time slot")
    // interrupts();
    } 
  Fin_millis = millis();
  Diff_millis = Fin_millis - Debut_millis;
  // Serial.print("delai MILLIsecondes =  ");  Serial.println(Diff_millis);
  if (Diff_millis < 600) //640
    {
    // Serial.println("Erreur - délai MILLIsecondes hors normes");
    Temperature[7] = '9';  Temperature[8] = '2';  
    LCD_AfficherChaine(0x42, Temperature);
    return false;
    }

  //-----------------  3ième partie  -  récupération des données de température
  // -----------------                

  // On récupère les données de la SRAM (Scratchpad) pour obtenir les données de la température
  // qui vient d'être lue par le sensor et convertie en binaire sur 12 bits.
  // Mais en pratique, on n'a pas le choix de tout récupérer le "Scratchpad",
  // qui est un ensemble de 9 octets formant toute la SRAM.

  // On doit initialiser à nouveau le DS18B20 en lui demandant un reset
  if (Temperature_Reset() == false)
    {
    Serial.println(F("Erreur a la demande de reset (3) au DS18B20"));
    Temperature[7] = '9';  Temperature[8] = '1';  
    LCD_AfficherChaine(0x42, Temperature);
    return false;
    }

  // On passe la commande Skip ROM, qui a pour numéro 0xCC
  Temperature_EnvoyerCode(0xCC);

  // On envoie le code la fonction Read Scratchpad, soit 0xBE
  Temperature_EnvoyerCode(0xBE);

  if (Temperature_LireDonnees(9) == false)
    {
    // Serial.println("Erreur pour lire les données de la SRAM");
    Temperature[7] = '9';  Temperature[8] = '0';  
    LCD_AfficherChaine(0x42, Temperature);
    return false;
    } 

  Temperature_CalculsFinaux();
  LCD_AfficherChaine(0x43, Temperature);

  return true;

  }

/*-----------------------------------------------------------------------------------------------*/
boolean Temperature_CRC8(int ParamNombreOctets)
  {
  /*
  Cette section concerne le CRC (Cyclic Redundanty Check), qui est un chiffre auto-vérificateur
  qui permet de démontrer qu'une transmission de données a été effectuée convenablement (error free).
  Le CRC se retrouve toujours dans le dernier octet envoyé.

  Pour les fins du présent logiciel, la vérification du CRC est effectuée systématiquement
  à chaque lecture.  Les deux situations possibles sont les suivantes:
    1- vérifier que numéro unique gravé en ROM a été bien transmis (8 octets);
    2- vérifier que les données de la SRAM ont été bien transmises (9 octets).

  Pour ce qui est de l'algorithme, même s'il est simple, il est difficile à expliquer, même
  à partir des textes du fabricant.  Alors exceptionnellement, il n'y aura pas d'explication ici.
  Toutefois, on peut cependant dire que deux approches étaient prévues par le fabricant:
    1 - un code "lent" mais plus petit (5 ou 6 lignes);
    2 - un code plus rapide mais demandant plus de mémoire, en fait une table de 256 octets
        permettant de comparer plus rapidement des valeurs;
  Ici on a choisi le code "lent" mais plus petit, car la lenteur est très relative, les tests
  montrent que toute la vérification requiert seulement de 76 à 80 μs pour 8 octets,
  et de 88 à 96 μs pour 9 octets

  On examine tous les octets (8 ou 9), incluant le dernier qui contient le CRC, et
  dans ce cas, le résultat final doit être 0.
  Par contre, on aurait aussi pu examiner seulement les octets du premier jusqu'à l'avant-dernier,
  ce qui aurait donné un résultat correspondant au contenu du dernier octet.
  */

  int boucle06;
  int boucle07;

  byte byte_examine;
  byte bitwise_temp;
  byte crc_temp;

  //unsigned long Debut_microsecondes_crc;
  //unsigned long Fin_microsecondes_crc;

  //Debut_microsecondes_crc = micros();
  crc_temp = 0;
  for (boucle06 = 0;  boucle06 < ParamNombreOctets;  boucle06++)
    {
    byte_examine = donnees_temp[boucle06];

    for (boucle07 = 0;  boucle07 < 8;  boucle07++)
      {
      bitwise_temp = (crc_temp ^ byte_examine) & 0x01;
      crc_temp >>= 1;
      if (bitwise_temp) {  crc_temp ^= 0x8C;   }
      byte_examine >>= 1;
      }

    // Si on désire voir le résultat pour les premiers octets jusqu'à l'avant-dernier, pour bien
    // montrer qu'il est bien égal à celui proposé par le fabricant dans le dernier octet.
    // if (ParamNombre - 2)
    //   {
    //   Serial.print("crc provisoire =  ");  Serial.println(crc_temp, HEX);
    //   Serial.print("dernier octet  =  ");  Serial.println(ParamValeur[ParamNombreOctets - 1], HEX);
    //   }
    }
  //Fin_microsecondes_crc = micros();
  // Serial.print("reponse crc_temp        =  ");  Serial.println(crc_temp, HEX);
  // Serial.print("Debut_microsecondes_crc =  ");  Serial.println(Debut_microsecondes_crc);
  // Serial.print("Fin_microsecondes_crc   =  ");  Serial.println(Fin_microsecondes_crc);
  // Serial.print("delai microsecondes     =  ");  Serial.println(Fin_microsecondes_crc - Debut_microsecondes_crc);

  // Si le crc n'est pas égal à 0, il y a une erreur.
  // Ici, il n'y a pas de message d'erreur;  il sera plutôt déterminé par la procédure appelante.
  if  (crc_temp == 0 )
    {  return true;   }
  else
    {  return false;  }
  }

/*-----------------------------------------------------------------------------------------------*/
void Temperature_EnvoyerCode(byte ParamCode)
  {
  
  int boucle05;

  // unsigned long Debut_microsecondes;
  // unsigned long Fin_microsecondes;

  // Serial.print("Commande a envoyer:  ");   Serial.println(ParamCode, HEX);

  // Debut_microsecondes = micros();
  for (boucle05 = 0;  boucle05 < 8;  boucle05++)
    {
    //  Le test suivant a été fait:        xyz = (  0x01 << (byte)boucle03  )        , et
    //  xyz est égal à 1, 2, 4, 8, 16, 32, 64, 128, en fonction de boucle03 qui va de 0 à 7 inclus.
    //  La boucle débute bien par le LSB.
    if (  ParamCode & (  0x01 << (byte)boucle05)  )
      {
      // On écrit un 1
      pinMode(Senseur_pin_Temperature, OUTPUT);         // On initialise la pin en mode OUTPUT
      digitalWrite(Senseur_pin_Temperature, LOW);       // On place la pin à LOW
      // delayMicroseconds(1);                          // l'exécution de la prochaine ligne est déjà suffisante
                                                        // pour créer un délai

      pinMode(Senseur_pin_Temperature, INPUT_PULLUP);   // On relâche la pin;  la ligne remonte  à HIGH
      delayMicroseconds(60);                            // On attend un temps additionnel de 60 μs

      delayMicroseconds(3);                             // On attend pour le "recovery time" qui doit être
                                                        // d'un minimum de 1 μs.
      }
    else
      {
      // On écrit un 0
      pinMode(Senseur_pin_Temperature, OUTPUT);         // On initialise la pin en mode OUTPUT
      digitalWrite(Senseur_pin_Temperature, LOW);       // On place la pin à LOW

      delayMicroseconds(60);                            // On attend 60 μs
      pinMode(Senseur_pin_Temperature, INPUT_PULLUP);   // On relâche la pin;  la ligne remonte à HIGH

      delayMicroseconds(3);                             // On attend pour le "recovery time" qui doit être
                                                        // d'un minimum de 1 μs.
      }
    }
  // Fin_microsecondes = micros();
  // Serial.print("delai microsecondes =  ");  Serial.println(Fin_microsecondes - Debut_microsecondes);
  }

/*-----------------------------------------------------------------------------------------------*/
boolean Temperature_Reset()
  {
  /*
  Toute série d'échanges avec le DS18B20 commence à chaque fois par une séquence d'initialisation
  qui comprend  deux étapes:
    1 - le microcontrôleur envoie une pulsion sur la ligne pour forcer un reset du DS18B20;
    2 - le DS18B20 répond en envoyant à son tour une pulsion sur la ligne pour signaler sa présence.

  Concrètement, les choses se passent ainsi selon le manuel:
    1 – le microcontrôleur envoie au DS18B20 une pulsion de reset (reset pulse);
          - le microcontrôleur met à LOW la ligne, pour un minimum de 480 μs;
          - immédiatement après, il "relâche" la ligne (en remettant sa pin en INPUT_PULLUP);
          - la résistance pull-up présente sur la ligne "attire" la ligne à HIGH;
    2 – le DS18B20, ayant reconnu une une pulsion de reset (reset pulse),
        il répond par une pulsion de présence (presence pulse);
          - le DS18B20 détecte que la ligne a été à LOW au moins 480 μs, et
            qu'ensuite elle est remontée à HIGH;
          - il attend de 15 μs à 60 μs;
          - il transmet une pulsion de présence (presence pulse) en mettant
            la ligne de donnée à LOW pour une période de  60 μs à 240 μs;
  */

  // unsigned long Debut_microsecondes;
  // unsigned long Fin1_microsecondes;
  // unsigned long Fin2_microsecondes;

  boolean ValeurRetour;

  // Serial.println("On entre dans Temperature_Reset");

  pinMode(Senseur_pin_Temperature, OUTPUT);         // On initialise la pin en mode OUTPUT
  digitalWrite(Senseur_pin_Temperature, LOW);       // On place la pin à LOW
  delayMicroseconds(500);                           // On attend 500 μs  (minimum recommandé:  480 μs)
  pinMode(Senseur_pin_Temperature, INPUT_PULLUP);   // On remet la pin en mode INPUT_PULLUP;

  // Les tests montrent que la pin reste à HIGH 28 μs parfois 24 μs lorsqu'une résistance externe est placée;
  // Les tests montrent que la pin reste à HIGH 32 μs parfois 28 μs sans résistance externe;
  // pour la partie LOW, c'est 144 μs, parfois 148 μs lorsqu'une résistance externe est placée;   parfois 132 μs...
  // pour la partie LOW, c'est 148 μs, parfois 152 μs sans résistance externe;
  // Sil n'y a pas noInterrupts() et interrupts(), le LOW est à 4 μs, ce qui est invraisemblable.
  // On remarquera que ce sont des multiples de 4, vraisemblablement parce que
  // while (digitalRead(Senseur_pin_Temperature) == xxxx) requiert 4 μs de temps d'exécution.
  
  /*
  noInterrupts();
  Debut_microsecondes = micros();
  while (digitalRead(Senseur_pin_Temperature) == HIGH);
  Fin1_microsecondes = micros();
  while (digitalRead(Senseur_pin_Temperature) == LOW);
  interrupts();
  Fin2_microsecondes = micros();
  Serial.print("delai microsecondes etait HIGH =  ");  Serial.println(Fin1_microsecondes - Debut_microsecondes);
  Serial.print("delai microsecondes etait LOW  =  ");  Serial.println(Fin2_microsecondes - Debut_microsecondes);
  delayMicroseconds(150);                        // On attend 150 μs.
  return true;
  */
  
  delayMicroseconds(32);
  delayMicroseconds(61);

  // Maintenant on est prêt à vérifier (sample) l'état de la ligne.
  // Si la ligne est à LOW,  c'est que le DS18B20 est bien présent et à réagi à la demande de reset.
  // Si la ligne est à HIGH, c'est que le DS18B20 est débranché, absent ou brisé.
  noInterrupts();
  if (digitalRead(Senseur_pin_Temperature) == LOW)
    {  ValeurRetour = true;  }
  else
    {  ValeurRetour = false; }
  interrupts();

  delayMicroseconds(150);                        // On attend 150 μs.

  return ValeurRetour;
  }

/*-----------------------------------------------------------------------------------------------*/
boolean Temperature_LireDonnees(int ParamNombreOctetsALire)
  {
  int boucle03;
  int boucle04;

  int integer_temp_1;
  int integer_temp_2;

  // unsigned long Debut_microsecondes;
  // unsigned long Fin_microsecondes;

  // Debut_microsecondes = micros();
  for (boucle03 = 0;  boucle03 < ParamNombreOctetsALire;  boucle03++)
    {
    integer_temp_1 = 0;
    for (boucle04 = 0;  boucle04 < 8;  boucle04++)
      {
      // Pour ce qui est du delayMicroseconds() placé juste après pinMode(Senseur_pin_Temperature, INPUT_PULLUP):
      // les tests (35 boucles) montrent qu'avec une résistance externe sur la ligne,
      // c'est impeccable pour à 1 à 16 μs;  à 17 μs et plus , rien ne va plus, tout est en erreur;
      // sans résistance externe, il y a parfois une erreur, parfois pas du tout, ou parfois bien plus;
      // on va choisir 9 parce que le manuel indique d'aller plus vers le maximum recommandé qui est 15;
      // 8 est au milieu.  c'est difficile à dire;  peut-être 10 aussi.
      // On va choisir d'utiliser la résistance externe, tel que recommandé, même si
      // le microcontrôleur de l'Arduino a sa propre résistance interne sur ses entrées,
      // d'où l'utilisation du mot INPUT_PULLUP.
      noInterrupts();
      pinMode(Senseur_pin_Temperature, OUTPUT);             // On initialise la pin en mode OUTPUT
      digitalWrite(Senseur_pin_Temperature, LOW);           // On place la pin à LOW
      pinMode(Senseur_pin_Temperature, INPUT_PULLUP);       // On relâche la pin;  la ligne remonte à HIGH
      delayMicroseconds(9);                                 // On attend 9 μs
      if (digitalRead(Senseur_pin_Temperature) == HIGH)
        {  integer_temp_2 = 1;  }
      else
        {  integer_temp_2 = 0;  }
      integer_temp_1 = integer_temp_1 + (  integer_temp_2 << boucle04  );
      interrupts();

      delayMicroseconds(45);                                // On attend 45 μs (voir commentaires pour plus d'explications)
      }
    donnees_temp[boucle03] = (byte)integer_temp_1;
    }

  // Fin_microsecondes = micros();
  // Serial.print("delai microsecondes =  ");  Serial.println(Fin_microsecondes - Debut_microsecondes);

  //for (boucle03 = 0;  boucle03 < ParamNombreOctetsALire;  boucle03++)
  //  {  Serial.print("byte ");  Serial.print(boucle03);  Serial.print(":  ");  Serial.println(donnees_temp[boucle03], HEX);  }

  // On calcule le CRC immédiatement
  if (Temperature_CRC8(ParamNombreOctetsALire) == true)
    {  return true;   }
  else
    {  return false;  }

  }

/*-----------------------------------------------------------------------------------------------*/
void Temperature_CalculsFinaux()
  {
  
  // byte resolution;

  unsigned int temperature_temp;
  unsigned int temperature_calcul;
  
  unsigned int signe_int;
  byte signe_byte;

  unsigned int entier = 0;
  unsigned int decimal;
  unsigned int reste;
  
  unsigned int i;
  unsigned int j;

  //float    temperatureFloat; //GP

  // float celsius_12;  float celsius_11;  float celsius_10;  float celsius_09;  float celsius_08;

  // Description des 9 bytes du Scratchpad
  //   byte 0 :    Temperature LSB  
  //   byte 1 :    Temperature MSB
  //   byte 2 :    Registre pour alarme (TH)
  //   byte 3 :    Registre pour alarme (TL)
  //   byte 4 :    Configuration register
  //   byte 5 :    Réservé   toujours à FF hexa ? 
  //   byte 6 :    Réservé                              serait le COUNT_REMAIN
  //   byte 7 :    Réservé   toujours à 10 hexa ?       serait le COUNT_PER_C
  //   byte 8 :    CRC

  // On vérifie que la conversion a bel et bien été faite sur 12 bits en examinant le
  // byte 4, Configuration register, où les bits 5 et 6 représente la résolution.
  // bit 7    bit 6    bit5    bit 4    bit 3    bit 2    bit 1    bit 0
  //  -        R1       R0       -        -        -        -        -
  //
  // 00 =  9 bits      01 = 10 bits      02 = 11 bits      03 = 12 bits      
  /*
  resolution = donnees_temp[4] & 0x60;    //  on extrait les 2 bits (bit 6 et bit 5)
  if (resolution == 0x60)
    {  Serial.println("La resolution est sur 12 bits");  }
  else
    {  Serial.println("ERREUR:  la resolution N'est PAS sur 12 bits");  return false;  }
  */

  // On transfère les 2 bytes de la température transmise dans une variable 16 bits
  temperature_temp = (donnees_temp[1] << 8) | donnees_temp[0];
  // Serial.print("Temperature lue:  ");  Serial.print(temperature_temp, HEX);  Serial.print("(hex)    ");
  // Serial.print(temperature_temp, DEC);  Serial.print("(dec)    ");

  // On vérifie s'il s'agit d'une valeur positive ou négative.
  // Pour ce test, on décale le dernier bit à gauche (MSB) jusqu'à la position 0;
  signe_int = temperature_temp  >> 15;
  // if (signe_int == 1)  {  Serial.println("nombre negatif");  } else { Serial.println("nombre positif");  } 
  if (signe_int == 1)
    {
    // le nombre est négatif  
    temperature_calcul = ~temperature_temp;         //  on inverse d'abord tous les bits, et ensuite 
    temperature_calcul = temperature_calcul + 1;    //  on additionne +1
    signe_byte = '-';
    } 
  else  
    {
    // le nombre est positif
    temperature_calcul = temperature_temp;
    signe_byte = '+';
    } 
  
  // Serial.print("Final signe et nombre :  ");  Serial.print(char(signe_byte));   Serial.print("    ");
  // Serial.print(temperature_calcul, HEX);      Serial.print("(hex)    ");
  // Serial.print(temperature_calcul, DEC);      Serial.println("(dec)");

  /*
  // On calcule la température  (mais ces calculs ne servent plus finalement).
  // La résolution sur 12 bits fait que 1 bit = 0.0625 ou 1/16
  celsius_12 = (float)temperature_calcul * 0.0625;                // on aurait pu aussi diviser par 16
  Serial.print("12 bits:  ");    Serial.print("temperature celcius:  ");  Serial.println(celsius_12, 4);  
  
  celsius_11 = (float)(temperature_calcul >> 1) * 0.125;          // on aurait pu aussi diviser par  8
  Serial.print("11 bits:  ");    Serial.print("temperature celcius:  ");  Serial.println(celsius_11, 4);  
  
  celsius_10 = (float)(temperature_calcul >> 2) * 0.25;           // on aurait pu aussi diviser par  4
  Serial.print("10 bits:  ");    Serial.print("temperature celcius:  ");  Serial.println(celsius_10, 4);  
  
  celsius_09 = (float)(temperature_calcul >> 3) * 0.5;            // on aurait pu aussi diviser par  2
  Serial.print(" 9 bits:  ");    Serial.print("temperature celcius:  ");  Serial.println(celsius_09, 4);  

  celsius_08 = (float)(temperature_calcul >> 4) * 1;              // on aurait pu aussi diviser par  1
  Serial.print(" 8 bits:  ");    Serial.print("temperature celcius:  ");  Serial.println(celsius_08, 4);  
  */

  // Extraction de la partie entière et de la partie décimale
  // Le plus simple est de faire une boucle... 
  j = 0;
  for (i = 0;  i <= temperature_calcul;  i++)   // le maximum positif retourné par le DS18B20 est 07D0 (hex) ou 2000 (déc)
    {           
    reste = i % 16;
    if (reste == 0)
      {
      entier = j;
      j++;
      } 
    decimal = (i - (entier * 16)) * 625;
    }
   //Serial.print("Final  signe, entier, nombre:  ");  Serial.print(char(signe_byte));   Serial.print("    ");
   //Serial.print(entier);   Serial.print("    ");  Serial.println(decimal);


  // On recopie le tout dans la variable globale Temperature[10]
  Temperature[0] = signe_byte;

  Convertir_Integer_ASCII(entier);
  Temperature[1] = Chaine001_ASCII[1];  Temperature[2] = Chaine001_ASCII[2];  Temperature[3] = Chaine001_ASCII[3];
  // On enlève les zéros au début de l'entier
  if (Temperature[1] == '0')
    {
    Temperature[1] = ' ';
    if (Temperature[2] == '0') { Temperature[2] = ' '; }
    }

  Temperature[4] = '.';

  Convertir_Integer_ASCII(decimal);
  Temperature[5] = Chaine001_ASCII[0];  Temperature[6] = Chaine001_ASCII[1];
  Temperature[7] = Chaine001_ASCII[2];  Temperature[8] = Chaine001_ASCII[3];
  
  Temperature[9] = '\0';
  
  // i = 0;  while (Temperature[i] != byte('\0'))  {  Serial.print(char(Temperature[i]));  Serial.print("");  i++;  }
  // Serial.println();

  // On détruit ces données immédiatement au cas où ultérieurement des lectures erronnées surviendraient,
  // et que d'anciennes données seraient réutilisées.
  for (i = 0;  i < 9;  i++)
    {  donnees_temp[i] =  i * i;  }  //  N'importe quoi comme valeur!  Et le CRC est erronné (vérifié).

  //Patch GP
  Temperature_int = entier;
  
  if (signe_byte == '-') {
    Temperature_int = Temperature_int * -1;
  }
   
  if ( (Temperature[5] == '5') ||
       (Temperature[5] == '6') ||
       (Temperature[5] == '7') ||
       (Temperature[5] == '8') ||
       (Temperature[5] == '9') ) {

     if (Temperature_int >= 0) {
       Temperature_int = Temperature_int + 1;  
     } else {
       Temperature_int = Temperature_int - 1;  
     }
  }
     
  //Fin Patch GP
  
  }

//*******************************************************************************************
//*******************************************************************************************
//*******************************************************************************************

//Fonction qui vide le tableau sensorValuesArray 
void Meteo_sensorValuesArray_reset() { 
  for (int i = 0; i < SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {   
   sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE]         = METEO_VALUE_UNSET;
   sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION]       = METEO_VALUE_UNSET;
   sensorValuesArray[i][SENSOR_VALUES_DIMENSION_TEMPERATURE]          = METEO_VALUE_UNSET;    
   sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE_MAXIMUM] = METEO_VALUE_UNSET;    
  }
  
  return;
}

//Fonction qui retourne le nombre d'entree du tableau sensorValuesArray qui contient des donnees.
int Meteo_sensorValuesArray_getNbrEntry() {
  int nbrEntry = 0;

  //On calcule le nombre d'entree du tableau 
  for (int i = 0; i < SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {   
    if (sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE] != METEO_VALUE_UNSET) {
     nbrEntry++;
    } else {
     break;
    }    
  }     
    
  return nbrEntry;
}

//Fonction qui ajoute des donnes au tableau sensorValuesArray 
void Meteo_sensorValuesArray_addTo(int ventVitesse, int ventDirection, int temperature, int ventVitesseMaximum) {
  int nbrEntry = Meteo_sensorValuesArray_getNbrEntry();
  int myValue;

  // On ajoute à la suite du tableau
  // Si le tableau est plein et bien on est pret a perdre l'entree la plus ancienne (donc la premiere entree qui est l'indice 0).
  // Donc decale toutes les entrees de -1 (ex.: l'entree 1 va devenir l'entree 0, l'entree 2, va devenir l'entree 1, etc.
  // puis ensuite on ajoute l'entree a la dernierre position (donc SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES - 1)

  if (nbrEntry >= SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES) {
   //On decale les entrees
   for (int i = 1; i < SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {   
    sensorValuesArray[i - 1][SENSOR_VALUES_DIMENSION_VENT_VITESSE]         = sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE];
    sensorValuesArray[i - 1][SENSOR_VALUES_DIMENSION_VENT_DIRECTION]       = sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION];
    sensorValuesArray[i - 1][SENSOR_VALUES_DIMENSION_TEMPERATURE]          = sensorValuesArray[i][SENSOR_VALUES_DIMENSION_TEMPERATURE];
    sensorValuesArray[i - 1][SENSOR_VALUES_DIMENSION_VENT_VITESSE_MAXIMUM] = sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE_MAXIMUM];
   }  

   nbrEntry = (SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES - 1);
  }

  //Pour sauver de la memoire, ce tableau est un tableau de uint8_t donc pour pouvoir
  //garder des valeurs de temperature negatives, on ajoute +100 systematiquement a la temperature.
  //Donc voici les valeurs acceptables:
  // ventVitesse:   0 a 250 (donc ne change pas). 
  // ventDirection: 0 a 7 (donc ne change pas). 
  // temperature:   -100 a +150 (Puisqu'on ajoute systematiquement +100, les valeurs dans le tableau seront de 0 a 250)
  // A titre informatif, les valeures de 251 a 255 sont reservees. Voir par exemple METEO_VALUE_UNSET et METEO_VALUE_OUT_OF_BOUNDS

  myValue = ventVitesse;
  if ((myValue < 0) || (myValue > 250)) {
    myValue = METEO_VALUE_OUT_OF_BOUNDS;
  }
  sensorValuesArray[nbrEntry][SENSOR_VALUES_DIMENSION_VENT_VITESSE]   = myValue;
  
  myValue = ventDirection;
  if ((myValue < 0) || (myValue > 7)) {
    myValue = METEO_VALUE_OUT_OF_BOUNDS;
  }
  sensorValuesArray[nbrEntry][SENSOR_VALUES_DIMENSION_VENT_DIRECTION] = myValue;

  myValue = temperature + 100;
  if ((myValue < 0) || (myValue > 250)) {
    myValue = METEO_VALUE_OUT_OF_BOUNDS;
  }
  sensorValuesArray[nbrEntry][SENSOR_VALUES_DIMENSION_TEMPERATURE]    = myValue;    

  myValue = ventVitesseMaximum;
  if ((myValue < 0) || (myValue > 250)) {
    myValue = METEO_VALUE_OUT_OF_BOUNDS;
  }
  sensorValuesArray[nbrEntry][SENSOR_VALUES_DIMENSION_VENT_VITESSE_MAXIMUM]   = myValue;
    
  return;
}

//Fonction qui vide le tableau sensorAverageValuesArray 
void Meteo_sensorAverageValuesArray_reset() { 
  for (int i = 0; i < SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {   
   sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE]   = METEO_VALUE_UNSET;
   sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION] = METEO_VALUE_UNSET;
  }
  
  return;
}

//Fonction qui retourne le nombre d'entree du tableau sensorAverageValuesArray qui contient des donnees.
int Meteo_sensorAverageValuesArray_getNbrEntry() {
  int nbrEntry = 0;

  //On calcule le nombre d'entree du tableau 
  for (int i = 0; i < SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {   
    if (sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE] != METEO_VALUE_UNSET) {
     nbrEntry++;
    } else {
     break;
    }    
  }     
    
  return nbrEntry;
}

//Fonction qui ajoute des donnes au tableau sensorAverageValuesArray 
void Meteo_sensorAverageValuesArray_addTo(int ventVitesse, int ventDirection) {
  int nbrEntry = Meteo_sensorAverageValuesArray_getNbrEntry();
  int myValue;

  // On ajoute à la suite du tableau
  // Si le tableau est plein et bien on est pret a perdre l'entree la plus ancienne (donc la premiere entree qui est l'indice 0).
  // Donc decale toutes les entrees de -1 (ex.: l'entree 1 va devenir l'entree 0, l'entree 2, va devenir l'entree 1, etc.
  // puis ensuite on ajoute l'entree a la dernierre position (donc SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES - 1)

  if (nbrEntry >= SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES) {
   //On decale les entrees
   for (int i = 1; i < SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {   
    sensorAverageValuesArray[i - 1][SENSOR_VALUES_DIMENSION_VENT_VITESSE]   = sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE];
    sensorAverageValuesArray[i - 1][SENSOR_VALUES_DIMENSION_VENT_DIRECTION] = sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION];
   }  

   nbrEntry = (SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES - 1);
  }

  //Voici les valeurs acceptables:
  // ventVitesse:   0 a 250 
  // ventDirection: 0 a 7
  // A titre informatif, les valeures de 251 a 255 sont reservees. Voir par exemple METEO_VALUE_UNSET et METEO_VALUE_OUT_OF_BOUNDS

  myValue = ventVitesse;
  if ((myValue < 0) || (myValue > 250)) {
    myValue = METEO_VALUE_OUT_OF_BOUNDS;
  }
  sensorAverageValuesArray[nbrEntry][SENSOR_VALUES_DIMENSION_VENT_VITESSE]   = myValue;
  
  myValue = ventDirection;
  if ((myValue < 0) || (myValue > 7)) {
    myValue = METEO_VALUE_OUT_OF_BOUNDS;
  }
  sensorAverageValuesArray[nbrEntry][SENSOR_VALUES_DIMENSION_VENT_DIRECTION] = myValue;
    
  return;
}

//Fonction qui fait la moyenne des lectures receuillies et ajoute le resultat au tableau sensorValuesArray 
void Meteo_sensorAverageValuesArray_to_sensorValuesArray() {
  uint8_t moyenneVitesseVent           = 0; 
  uint8_t maximumVitesseVent           = 0; 
  int     totalDes_moyenneVitesseVent  = 0;
  uint8_t ventDirection = 0;
  uint8_t ventDirection_max = 0; 

  uint8_t directionVentsArray[8]; 

  directionVentsArray[DIRECTION_VENT_NORD]       = 0;
  directionVentsArray[DIRECTION_VENT_NORD_EST]   = 0;
  directionVentsArray[DIRECTION_VENT_EST]        = 0;
  directionVentsArray[DIRECTION_VENT_SUD_EST]    = 0;
  directionVentsArray[DIRECTION_VENT_SUD]        = 0;
  directionVentsArray[DIRECTION_VENT_SUD_OUEST]  = 0;
  directionVentsArray[DIRECTION_VENT_OUEST]      = 0;
  directionVentsArray[DIRECTION_VENT_NORD_OUEST] = 0;
  
  int     nbrEntry = 0; //Meteo_sensorAverageValuesArray_getNbrEntry();
  
  // On calcule le total des vitesse pour pouvoir ensuite faire la moyenne. 
  // En meme temp, on calcule le nombre d'entrees dans le tableau.
  // On en profite aussi pour recuperer la vitesse maximum atteinte
  maximumVitesseVent = 0;
  for (int i = 0; i < SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {   
    if (sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE] != METEO_VALUE_UNSET) {
     nbrEntry++;
     totalDes_moyenneVitesseVent = totalDes_moyenneVitesseVent + sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE];
     directionVentsArray[sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION]]++;
     if (sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE] > maximumVitesseVent) {
      maximumVitesseVent = sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE];
     }
    } else {
     break;
    }    
  }

  //Moyenne
  //Checkey decimal....
  moyenneVitesseVent = (uint8_t) round(totalDes_moyenneVitesseVent / (nbrEntry + 0.0)); //Le + 0.0 est important sinon le round ne se fait pas bien (il arrondit toujours a l'entier le plus bas. ex: 4.6 devrait etre arrondit a 5 mais si je ne met pas le +0.0 il arrondit a 4)
  #ifdef METEO_DEBUG
   DEBUG_PRINT(F("moyenneVitesseVent = (uint8_t) round(totalDes_moyenneVitesseVent / (nbrEntry + 0.0)) => "));
   DEBUG_PRINTLN((uint8_t) round(totalDes_moyenneVitesseVent / (nbrEntry + 0.0)));
  #endif

  //On cherche la direction qui revient le plus souvent.
  ventDirection_max = 0;
  for (int i = 0; i < 8; i++) {   
    if (directionVentsArray[i] > ventDirection_max) {
      ventDirection_max = directionVentsArray[i];
    }
  }

  //Maintenant qu'on a le max, on doit verifier si il y a plus d'une direction qui reviennt le plus souvent avec un nombre egales.
  //Dans un tel cas, et bien on prend la derniere trouvee. C'est elle qui gagne. C'est simpliste mais bon, pour maintenant on va dire que c'est ca :-)
  ventDirection = METEO_VALUE_UNSET;
  for (int i = (SENSOR_AVERAGE_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES - 1); i >= 0; i--) {   
    if (sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION] != METEO_VALUE_UNSET) {
      if (directionVentsArray[sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION]] == ventDirection_max){
        ventDirection = sensorAverageValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION];
        break;
      }
    }    
  }
  
  //On ajoute le resultat au tableau. Pour ce qui est de la temperature, on prend simplement la derniere lue qui est toujours gardee dans la variable Temperature_int.
  Meteo_sensorValuesArray_addTo(moyenneVitesseVent,ventDirection,Temperature_int,maximumVitesseVent);
  
  //On reset la tableau qui sert a receuillir les lectures
  Meteo_sensorAverageValuesArray_reset();
  
  return;
}

//Fonction qui retourne une chaine qui contient toutes les entrees du tableau sensorAverageValuesArray bout a bout
char * Meteo_getString_dataToSend() {

  char * retour = (char *) malloc ((SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES * SENSOR_VALUES_RECORD_LENGTH) +  1); 
  char * pChaine = NULL;
  
  #ifdef METEO_DEBUG
   DEBUG_PRINT(F("freeMemory() 1="));
   DEBUG_PRINTLN(freeMemory());
  #endif
    
  strcpy(retour,"");

  // On lit les valeurs stockees dans l'array a deux dimensions sensorValuesArray puis on fait une chaine avec ca.
  // Vu que la memoire du arduino est tres limitee, je vais faire une string qui contient les valeurs sans aucun separateur.

  for (int i = 0; i < SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES; i++) {    
      if (sensorValuesArray[i][0] != METEO_VALUE_UNSET) {
              
        //Vitesse du vent
        pChaine = Meteo_VitesseVent_int_to_string_plusPadding(sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE]);
        strcat(retour,pChaine);
        free(pChaine);

        //Direction du vent
        pChaine = Meteo_DirectionVent_int_to_string(sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_DIRECTION]);
        strcat(retour,pChaine);
        free(pChaine);

        //Temperature
        pChaine = Meteo_Temperature_int_to_string_plusPadding(sensorValuesArray[i][SENSOR_VALUES_DIMENSION_TEMPERATURE] - 100);
        strcat(retour,pChaine);
        free(pChaine);        

        //Vitesse du vent maximum atteints
        pChaine = Meteo_VitesseVent_int_to_string_plusPadding(sensorValuesArray[i][SENSOR_VALUES_DIMENSION_VENT_VITESSE_MAXIMUM]);
        strcat(retour,pChaine);
        free(pChaine);
      }    
  }

  #ifdef METEO_DEBUG
   DEBUG_PRINT(F("freeMemory()="));
   DEBUG_PRINTLN(freeMemory());
  #endif    
    
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("() => "));
   DEBUG_PRINTLN(retour);
  #endif
  
  return retour; //La fonction appelante doit s'occuper de faire un free() de cette variable.
}

//Fonction qui retourne la vitesse du vent au format chaine mais dans un format predefini
char * Meteo_VitesseVent_int_to_string_plusPadding(int i) {  
  
  char * retour = (char *) malloc (3 + 1); //Maximum = xx0 a 999
  char   tempBuffer[7]; 
  int    vitesse = i;
  
  //Les vitesses acceptees sont entre 0 et 999
  vitesse = max(min(i,999),0); 
  itoa(vitesse,tempBuffer,10);
  strcpy(retour,"");
  
  if (vitesse < 10) {
   strcat(retour,SENSOR_VALUES_RECORD_PADDING_CHARACTER);
   strcat(retour,SENSOR_VALUES_RECORD_PADDING_CHARACTER);
  }
  
  if ((vitesse >= 10) && (vitesse <= 99)) {   
   strcat(retour,SENSOR_VALUES_RECORD_PADDING_CHARACTER);
  }

  strcat(retour,tempBuffer);
  
  return retour; //La function appelante devra s'occuper de liberer la memoire de cette variable
}

//Fonction qui retourne la direction du vent au format chaine mais dans un format predefini
char * Meteo_DirectionVent_int_to_string(int i) {  
  
  char * retour = (char *) malloc (1 + 1); //Maximum = 0 a 7
  char   tempBuffer[7]; 
  int    directionVent = i;
  
  //Les directions acceptees sont entre 0 et 7 (voir constantes DIRECTION_VENT_...
  directionVent = max(min(i,7),0);
  itoa(directionVent,tempBuffer,10);
  strcpy(retour,tempBuffer);
  
  return retour; //La function appelante devra s'occuper de liberer la memoire de cette variable
}

//Fonction qui retourne la temperature au format chaine mais dans un format predefini
char * Meteo_Temperature_int_to_string_plusPadding(int i) {  
  
  char * retour = (char *) malloc (3 + 1); //Maximum = -99 a x99
  char   tempBuffer[7]; 
  int    temperature = i;
  
  //Les temperatures acceptees sont entre -99 et 99
  temperature = max(min(i,99),-99);
  itoa(abs(temperature),tempBuffer,10);
  strcpy(retour,"");

  if (temperature >= 10) {   
   strcat(retour,SENSOR_VALUES_RECORD_PADDING_CHARACTER);
  }

  if (temperature <= -10) {   
   strcat(retour,"-");
  }
    
  if ((temperature > -10) && (temperature < 0)) {   
   strcat(retour,SENSOR_VALUES_RECORD_PADDING_CHARACTER);
   strcat(retour,"-");
  }

  if ((temperature < 10) && (temperature >= 0)) {   
   strcat(retour,SENSOR_VALUES_RECORD_PADDING_CHARACTER);
   strcat(retour,SENSOR_VALUES_RECORD_PADDING_CHARACTER);
  }   

  strcat(retour,tempBuffer);
  
  return retour; //La function appelante devra s'occuper de liberer la memoire de cette variable
}

//"Function overloading" pour autre fonction du meme nom
void Meteo_updateLCD() {
  Meteo_updateLCD(1,F(""));
  return;
}

//"Function overloading" pour autre fonction du meme nom
void Meteo_updateLCD(int lineNumber, const __FlashStringHelper * stringToDisplay) { 
  char * myStringOfChar = (char *) calloc(MAX_LCD_CHARACTERS + 1, sizeof(char)); 
  strncpy_P(myStringOfChar,(const char PROGMEM *)stringToDisplay,MAX_LCD_CHARACTERS);

  #ifdef METEO_DEBUG     
    //DEBUG_PRINT(F("__FlashStringHelper * stringToDisplay="));
    //DEBUG_PRINTLN(stringToDisplay);
    //DEBUG_PRINT(F("myStringOfChar="));
    //DEBUG_PRINTLN(myStringOfChar);
  #endif
  
  Meteo_updateLCD(lineNumber,myStringOfChar);
  free(myStringOfChar);
  
  return;
}

//Fonction "intelligente" qui met a jour le display LCD selon ou elle est rendu. //Finalement, un bon bout ne sert pas ou plus.
void Meteo_updateLCD(int lineNumber, const char * stringToDisplay ) {
  
  char myStringOfChar[MAX_LCD_CHARACTERS + 1];
  char * pChaine = NULL;
  
  int stringToDisplay_len = strlen(stringToDisplay);
  static uint8_t what_info_to_display_now_line_2 = INFO_TO_DISPLAY_TYPE_DATE; //Ligne 2
         uint8_t what_info_to_display_now_line_1 = what_info_to_display_now_line_2 -1; //Ligne 1 
  
  if (stringToDisplay_len == 0) {
   if ( ((lastTime_Meteo_updateLCD_fired + INFO_TO_DISPLAY_INTERVAL) > millis()) && (lastTime_Meteo_updateLCD_fired > 0) ) {
    #ifdef METEO_DEBUG
     //DEBUG_PRINTLN(F("Sort de Meteo_updateLCD() sans rien faire."));
    #endif 
    
    return;
   }
  }
  
  lastTime_Meteo_updateLCD_fired = millis();
  
  if (stringToDisplay_len == 0) {       
   if (what_info_to_display_now_line_1 < 1) {
    what_info_to_display_now_line_1 = INFO_TO_DISPLAY_LAST_TYPE_NUMBER; //Le dernier
   }
  }

  #ifdef METEO_DEBUG
   //DEBUG_PRINT(F("stringToDisplay_len="));
   //DEBUG_PRINTLN(stringToDisplay_len);
   DEBUG_PRINT(F("stringToDisplay="));
   DEBUG_PRINT(stringToDisplay);
   DEBUG_PRINT(F(" what_info_to_display_now_line_2="));
   DEBUG_PRINT(what_info_to_display_now_line_2);
   DEBUG_PRINT(F(" what_info_to_display_now_line_1="));
   DEBUG_PRINTLN(what_info_to_display_now_line_1);
  #endif 
  
  //Si stringToDisplay_len est 0, ca veut dire qu'on affiche simplement les informations "habituelles" 
  //sinon et bien on affiche le texte recu en parametre via stringToDisplay.
  if (stringToDisplay_len == 0) {
    //************ LIGNE 1 *****************
    //On efface puis affiche affiche la ligne 1 (qui est en fait la ligne 2 en cours) //on scroll les informations vers le haut.
    LCD_AfficherChaine(0x00, F("                ")); //On commence par effacer la ligne en affichant des espaces

    //Affiche contenue ligne 1
    pChaine = Meteo_getInfoToDisplayOnLCD(what_info_to_display_now_line_1);
    strncpy(myStringOfChar,pChaine,MAX_LCD_CHARACTERS);
    //strcat(myStringOfChar,'\0'); //Marche pas???
    myStringOfChar[MAX_LCD_CHARACTERS] = '\0'; //Au cas ou
    
    #ifdef METEO_DEBUG
     //DEBUG_PRINT(F("Ligne 1 pChaine="));
     //DEBUG_PRINT(pChaine); 
     //DEBUG_PRINT(F("<= myStringOfChar="));
     //DEBUG_PRINT(myStringOfChar); 
     //DEBUG_PRINTLN(F("<=="));
    #endif 
    
    free(pChaine);
    LCD_AfficherChaine(0x00, myStringOfChar); 

    //************ LIGNE 2 *****************
    //On efface puis affiche affiche la ligne 2
    //Affiche 16 espaces sur la igne 2
    //for (int i=0; i < MAX_LCD_CHARACTERS; i++) { myStringOfBytes[i] = ' ';  }
    //myStringOfBytes[MAX_LCD_CHARACTERS - 1] = '\0';     
    LCD_AfficherChaine(0x40, F("                ")); //On commence par effacer la ligne en affichant des espaces

    //Affiche contenue ligne 2
    pChaine = Meteo_getInfoToDisplayOnLCD(what_info_to_display_now_line_2);
    strncpy(myStringOfChar,pChaine,MAX_LCD_CHARACTERS);
    //strcat(myStringOfChar,'\0'); //Marche pas???
    myStringOfChar[MAX_LCD_CHARACTERS] = '\0';
    free(pChaine);
    LCD_AfficherChaine(0x40, myStringOfChar); 
    
  } else {
    strncpy(myStringOfChar,stringToDisplay,MAX_LCD_CHARACTERS);
    myStringOfChar[MAX_LCD_CHARACTERS] = '\0'; //Au cas que stringToDisplay copiee ci-haut etait plus longue que myStringOfChar

    byte ParamAdresse = 0x00; //Ligne 1 Position 1
    if (lineNumber == 2) {
      ParamAdresse = 0x40;    //Ligne 2 Position 1
    }  

    #ifdef METEO_DEBUG
     //DEBUG_PRINT(F("lineNumber=")); 
     //DEBUG_PRINT(lineNumber);
     //DEBUG_PRINT(F(" ParamAdresse=")); 
     //DEBUG_PRINTLN(ParamAdresse);
    #endif    

    //Affiche 16 espaces sur la igne 
    //for (int i=0; i < MAX_LCD_CHARACTERS; i++) { myStringOfBytes[i] = ' ';  }
    //myStringOfBytes[MAX_LCD_CHARACTERS - 1] = '\0';     
    LCD_AfficherChaine(ParamAdresse, F("                ")); //On commence par effacer la ligne en affichant des espaces
    LCD_AfficherChaine(ParamAdresse, myStringOfChar);
  }    
  
  #ifdef METEO_DEBUG
   //DEBUG_PRINT(__func__); 
   //DEBUG_PRINT(F("("));
   //DEBUG_PRINT(lineNumber); 
   //DEBUG_PRINT(F(","));
   //DEBUG_PRINT(myStringOfChar); 
   //DEBUG_PRINTLN(F(");"));
  #endif    

  if (stringToDisplay_len == 0) {
   what_info_to_display_now_line_2++;
   
   if (what_info_to_display_now_line_2 > INFO_TO_DISPLAY_LAST_TYPE_NUMBER) {
    what_info_to_display_now_line_2 = 1; //Le premier, normalement INFO_TO_DISPLAY_TYPE_DATE
   }
  }
  
  return;  
}

//Fonction qui retourne une chaine a afficher sur le LCD selon les informations (parametre info_to_display) demandees
char * Meteo_getInfoToDisplayOnLCD(int info_to_display) {  
  
  char * pChaine = NULL;
  char   tempBuffer[10]; 
  char * retour = (char *) calloc(MAX_LCD_CHARACTERS + 1, sizeof(char)); 

  strncpy_P(retour,(const char PROGMEM *)F("***"),MAX_LCD_CHARACTERS);
  itoa(info_to_display,tempBuffer,10);
  strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
  strncat_P(retour,(const char PROGMEM *)F("***"),MAX_LCD_CHARACTERS - strlen(retour));
  
  int    nbrEntry = Meteo_sensorValuesArray_getNbrEntry();  
  
      switch(info_to_display) {
        case INFO_TO_DISPLAY_TYPE_DATE:
          if (myResponse_server_date_time > 0) {
            
            Meteo_secs_to_tm(myResponse_server_date_time + ((uint32_t)((millis() - myResponse_server_date_time_millis) / 1000)),my_tm_ptr); //Ex.: 1461488730 = 04/24/2016 @ 9:05am (UTC)
            
            #ifdef METEO_DEBUG
             DEBUG_PRINT(F("myResponse_server_date_time + ((uint32_t)((millis() - myResponse_server_date_time_millis) / 1000)) ="));
             DEBUG_PRINT(myResponse_server_date_time + ((uint32_t)((millis() - myResponse_server_date_time_millis) / 1000))); 
             DEBUG_PRINT(F(" myResponse_server_date_time_millis="));
             DEBUG_PRINT(myResponse_server_date_time_millis); 
             
             DEBUG_PRINTLN(F(""));
            #endif    
            
            itoa(my_tm.tm_year + 1900,tempBuffer,10);
            strncpy(retour,tempBuffer,MAX_LCD_CHARACTERS);
            strncat_P(retour,(const char PROGMEM *)F("-"),MAX_LCD_CHARACTERS - strlen(retour));
            if ((my_tm.tm_mon + 1) < 10) {
              strncat_P(retour,(const char PROGMEM *)F("0"),MAX_LCD_CHARACTERS - strlen(retour));
            }
            itoa(my_tm.tm_mon + 1,tempBuffer,10);
            strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
            strncat_P(retour,(const char PROGMEM *)F("-"),MAX_LCD_CHARACTERS - strlen(retour));
            if ((my_tm.tm_mday + 1) < 10) {
              strncat_P(retour,(const char PROGMEM *)F("0"),MAX_LCD_CHARACTERS - strlen(retour));
            }
            itoa(my_tm.tm_mday,tempBuffer,10);
            strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
            strncat_P(retour,(const char PROGMEM *)F(" "),MAX_LCD_CHARACTERS - strlen(retour));
            itoa(my_tm.tm_hour,tempBuffer,10);
            strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
            strncat_P(retour,(const char PROGMEM *)F(":"),MAX_LCD_CHARACTERS - strlen(retour));
            if ((my_tm.tm_min + 1) < 10) {
              strncat_P(retour,(const char PROGMEM *)F("0"),MAX_LCD_CHARACTERS - strlen(retour));
            }
            itoa(my_tm.tm_min,tempBuffer,10);
            strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
          } else {
            strncpy_P(retour,(const char PROGMEM *)F("Date: ???"),MAX_LCD_CHARACTERS - strlen(retour));
          }
          break; 
        
        case INFO_TO_DISPLAY_TYPE_WIND:
          if (nbrEntry > 0) {
            itoa(sensorValuesArray[nbrEntry - 1][SENSOR_VALUES_DIMENSION_VENT_VITESSE],tempBuffer,10);
            strncpy(retour,tempBuffer,MAX_LCD_CHARACTERS);
            strncat_P(retour,(const char PROGMEM *)F("Km/h "),MAX_LCD_CHARACTERS - strlen(retour));
            pChaine = Meteo_getVentDirectionIntToString(sensorValuesArray[nbrEntry - 1][SENSOR_VALUES_DIMENSION_VENT_DIRECTION]);
            strncat(retour,pChaine,MAX_LCD_CHARACTERS - strlen(retour));
            free(pChaine);
            strncat_P(retour,(const char PROGMEM *)F(" "),MAX_LCD_CHARACTERS - strlen(retour));
            itoa((sensorValuesArray[nbrEntry - 1][SENSOR_VALUES_DIMENSION_TEMPERATURE] - 100),tempBuffer,10);
            strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
            strncat_P(retour,(const char PROGMEM *)F("C"),MAX_LCD_CHARACTERS - strlen(retour));
          }
          break; 

        case INFO_TO_DISPLAY_TYPE_NBR_LECTURES:
          strncpy_P(retour,(const char PROGMEM *)F("Lectures: "),MAX_LCD_CHARACTERS);
          itoa(nbrEntry,tempBuffer,10);
          strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
          strncat_P(retour,(const char PROGMEM *)F("/"),MAX_LCD_CHARACTERS - strlen(retour));
          itoa(SENSOR_VALUES_ARRAY_NBR_ENTRIES_TO_KEEP_BEFORE_SENDING_TO_SERVER,tempBuffer,10);
          strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
          break; 
          
        case INFO_TO_DISPLAY_TYPE_BATTERY:
          //myResponse_battery_voltage
          dtostrf(myResponse_battery_voltage, 4, 3, tempBuffer);  //4 is mininum width, 3 is precision
          
          strncpy_P(retour,(const char PROGMEM *)F("B:"),MAX_LCD_CHARACTERS);
          strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
          strncat_P(retour,(const char PROGMEM *)F("V M:"),MAX_LCD_CHARACTERS - strlen(retour));
          itoa(freeMemory(),tempBuffer,10);
          strncat(retour,tempBuffer,MAX_LCD_CHARACTERS - strlen(retour));
          break; 

        case INFO_TO_DISPLAY_TYPE_LAST_COMM_STATUS:
          switch(myResponse_last_comm_status) {
            case METEO_CMD_STATUS_SUCCESS:
             strncpy_P(retour,(const char PROGMEM *)F("DerComSer:Ok    "),MAX_LCD_CHARACTERS);
             break; 
            case METEO_CMD_STATUS_ERROR:
             strncpy_P(retour,(const char PROGMEM *)F("DerComSer:Erreur"),MAX_LCD_CHARACTERS);
             break;  
            default:
             strncpy_P(retour,(const char PROGMEM *)F("DerComSer:???   "),MAX_LCD_CHARACTERS);
             break;
          }
          break;

        case INFO_TO_DISPLAY_TYPE_LAST_RESPONSE_STATUS:
          switch(myResponse_status) {
            case METEO_CMD_STATUS_SUCCESS:
             strncpy_P(retour,(const char PROGMEM *)F("DerRepSer:Ok    "),MAX_LCD_CHARACTERS);
             break; 
            case METEO_CMD_STATUS_ERROR:
             strncpy_P(retour,(const char PROGMEM *)F("DerRepSer:Erreur"),MAX_LCD_CHARACTERS);
             break;  
            default:
             strncpy_P(retour,(const char PROGMEM *)F("DerRepSer:???   "),MAX_LCD_CHARACTERS);
             break;
          }
          break;

        case INFO_TO_DISPLAY_TYPE_INFO_1:
          strncpy_P(retour,(const char PROGMEM *)F("Projet r{alis{  "),MAX_LCD_CHARACTERS);
          break; 

        case INFO_TO_DISPLAY_TYPE_INFO_2:
          strncpy_P(retour,(const char PROGMEM *)F("par les fr}res  "),MAX_LCD_CHARACTERS);
          break; 

        case INFO_TO_DISPLAY_TYPE_INFO_3:
          strncpy_P(retour,(const char PROGMEM *)F("Guylain et      "),MAX_LCD_CHARACTERS);
          break; 

        case INFO_TO_DISPLAY_TYPE_INFO_4:
          strncpy_P(retour,(const char PROGMEM *)F("Paul-Andr{      "),MAX_LCD_CHARACTERS);
          break; 

        case INFO_TO_DISPLAY_TYPE_INFO_5:
          strncpy_P(retour,(const char PROGMEM *)F("Plante          "),MAX_LCD_CHARACTERS);
          break; 

        case INFO_TO_DISPLAY_TYPE_EMPTY:
          strncpy_P(retour,(const char PROGMEM *)F("                "),MAX_LCD_CHARACTERS);
          break;   

        default: 
          strncpy_P(retour,(const char PROGMEM *)F("*default*"),MAX_LCD_CHARACTERS);
          break; 
      }
  
  
  return retour; //La function appelante devra s'occuper de liberer la memoire de cette variable
}

//Fonction qui retourne la direction du vent sour forme de chaine
char * Meteo_getVentDirectionIntToString(int ventDirection) {  

  char * retour = (char *) calloc (2 + 1,sizeof(char)); 

  switch(ventDirection) {
    case DIRECTION_VENT_NORD:
      strcpy_P(retour,(const char PROGMEM *)F("N"));
      break; 
        
    case DIRECTION_VENT_NORD_EST:
      strcpy_P(retour,(const char PROGMEM *)F("NE"));
      break; 

    case DIRECTION_VENT_EST:
      strcpy_P(retour,(const char PROGMEM *)F("E"));
      break; 

    case DIRECTION_VENT_SUD_EST:
      strcpy_P(retour,(const char PROGMEM *)F("SE"));
      break; 

    case DIRECTION_VENT_SUD:
      strcpy_P(retour,(const char PROGMEM *)F("S"));
      break; 

    case DIRECTION_VENT_SUD_OUEST:
      strcpy_P(retour,(const char PROGMEM *)F("SO"));
      break; 

    case DIRECTION_VENT_OUEST:
      strcpy_P(retour,(const char PROGMEM *)F("O"));
      break; 

    case DIRECTION_VENT_NORD_OUEST:
      strcpy_P(retour,(const char PROGMEM *)F("NO"));
      break;                  
  }
  
  return retour; //La function appelante devra s'occuper de liberer la memoire de cette variable
}

//Fonction qui a partir d'un Unix Timestamp retourne une structure qui contient les informations de date et heure
int Meteo_secs_to_tm(long long t, struct tm *tm) //http://git.musl-libc.org/cgit/musl/tree/src/time/__secs_to_tm.c?h=v0.9.15
{
  long long days, secs;

  int32_t remdays, remsecs, remyears;    //int //Patch GP: J'ai du change une couple de type sinon on defoncait dans certains cas comme par exemple pour la date 1462526746 (parametre t) mais ca fonctionnait quand meme par exempe pour la date 1462348800.
  int32_t qc_cycles, c_cycles, q_cycles; //int //Patch GP
  int years, months;
  int wday, yday, leap;
  static const char days_in_month[] = {31,30,31,30,31,31,30,31,30,31,31,29};

  /* Reject time_t values whose year would overflow int */
  if (t < INT_MIN * 31622400LL || t > INT_MAX * 31622400LL)
    return -1;

  secs = t - LEAPOCH;
  days = secs / 86400;
  remsecs = secs % 86400;
  if (remsecs < 0) {
    remsecs += 86400;
    days--;
  }

  wday = (3+days)%7;
  if (wday < 0) wday += 7;

  qc_cycles = days / DAYS_PER_400Y;
  remdays = days % DAYS_PER_400Y;
  if (remdays < 0) {
    remdays += DAYS_PER_400Y;
    qc_cycles--;
  }

  c_cycles = remdays / DAYS_PER_100Y;
  if (c_cycles == 4) c_cycles--;
  remdays -= c_cycles * DAYS_PER_100Y;

  q_cycles = remdays / DAYS_PER_4Y;
  if (q_cycles == 25) q_cycles--;
  remdays -= q_cycles * DAYS_PER_4Y;

  remyears = remdays / 365;
  if (remyears == 4) remyears--;
  remdays -= remyears * 365;

  leap = !remyears && (q_cycles || !c_cycles);
  yday = remdays + 31 + 28 + leap;
  if (yday >= 365+leap) yday -= 365+leap;

  years = remyears + 4*q_cycles + 100*c_cycles + 400*qc_cycles;

  for (months=0; days_in_month[months] <= remdays; months++)
    remdays -= days_in_month[months];

  if (years+100 > INT_MAX || years+100 < INT_MIN)
    return -1;

  tm->tm_year = years + 100;
  tm->tm_mon = months + 2;
  if (tm->tm_mon >= 12) {
    tm->tm_mon -=12;
    tm->tm_year++;
  }
  tm->tm_mday = remdays + 1;
  tm->tm_wday = wday;
  tm->tm_yday = yday;

  tm->tm_hour = remsecs / 3600;
  tm->tm_min = remsecs / 60 % 60;
  tm->tm_sec = remsecs % 60;

  return 0;
}

//Fonction qui s'occupe d'envoyer les donness (lectures des senseurs) sur le site web a l'aide de la carte FONA 3G
boolean fona3g_sendDataToWebServer(char const *dataToSend) {
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINTLN(F(""));
  #endif  

  myResponse_status           = METEO_CMD_STATUS_UNDEFINED;
  myResponse_last_comm_status = METEO_CMD_STATUS_ERROR;
  
  boolean retour      = false;
  boolean canContinue = true;
  char    atCommand[96];     //Pour le moment, la plus longue commande AT que j'envoi est d'environ 46 characteres (AT+CGSOCKCONT=1,"IP","rogers-core-appl1.apn") donc je me garde un petit jeu.
  char    httpRequest[300];  //256 pour 238//Pour le moment, ma requete GET pourrait atteindre 287 caracteres donc je me garde un petit jeu. //SENSOR_VALUES_RECORD_LENGTH x SENSOR_VALUES_ARRAY_MAXIMUM_NBR_ENTRIES = 150 en partant. On ajoute les autres parametres (voltage, etc.) ainsi que la requete http elle-meme. Voir code plus bas pour plus de details.
  //char    nbrCharsToSend[7]; 
  char    *my_char_pointer_1;
  char    *my_char_pointer_2;
  char    *ptr;
  int     pointerIndex = 0;
  char    tempBuffer[12]; 

  Serial.println(F("FONA: Initialisation @ 4800 baud..."));
  Meteo_updateLCD(1,F("Envoi donn{es"));
  Meteo_updateLCD(2,F("sur serveur..."));
  //Le port devrait normalement etre a 4800 baud parceque c'est a cette vitesse que je l'ai configure
  fonaSerial->begin(4800);
  canContinue = fona3g_init(*fonaSerial); //La fonction fona3g_init() "power on" le fona egalement.
  
  /*
  if (!canContinue) {
   //Si ca n'a pas marche, on essai a 115200 baud qui est la vitesse par defaut mais il parrait que la communication entre 
   //le Arduino et le Fona ne fonctionne pas bien a cette vitesse. On essai quand meme... //to log
   Serial.println(F("Initialisation du FONA @ 115200 baud..."));
   fonaSerial->begin(115200);
   canContinue = fona3g_init(*fonaSerial);
  }

  if (!canContinue) {
   //On reessai une derniere fois a 4800 baud
   Serial.println(F("Initialisation du FONA @ 4800 baud..."));
   fonaSerial->begin(4800);
   canContinue = fona3g_init(*fonaSerial);
  }
  */

  if (canContinue) {
    Serial.println(F("FONA: On attend 5 secondes pour voir si le fona a quelque chose a dire apres avoir reinitialise..."));
    // On passe une chaine bidon pour que la routine attende le timeout recu au complet
    // J'ai determine que 5 secondes etait suffisant. Anyway, on a le temps de tout facon :-)
    fona3g_waitForDeferredResponse(F("CHAINE BIDON"),5000); 
    
  }

  if (canContinue) {
    strcpy_P(atCommand,(const char PROGMEM *)F("AT+CBC")); //Voltage de la batterie
    fona3g_sendOnly(atCommand);
    fona3g_waitForDeferredResponse(F("+CBC:"),3000);
    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);
    
    // On garde le voltage de la batterie dans myResponse_battery_voltage 
    // Donc il faut decortiquer la reponse pour aller chercher la partie voltage (donc 3.810) si on se fie e l'exemple de reponse ci-dessous.
    // Exemple de reponse recue dans fona3g_replybuffer:+CBC: 0,75,3.810V 
    
    my_char_pointer_1 = strstr_P(fona3g_replybuffer, (const char PROGMEM *)F(".")); //Recherche le point
    
    if (my_char_pointer_1) { // Si le point a ete trouve
      pointerIndex = my_char_pointer_1 - fona3g_replybuffer; //On garde la position du point dans pointerIndex
      while (pointerIndex > 0) { 
        if (strncmp_P(my_char_pointer_1, (const char PROGMEM *)F(","), strlen_P((const char PROGMEM *)F(","))) == 0) { //En partant du point, on remonte la string caractere par caractere a la recherche de la virgule positionnee avant le debut du voltage.
          //On a trouve la virgule qu'on recherchait
          break;
        } else {
          my_char_pointer_1--; //On a pas encore trouve la virgule donc on decale le pointeur vers la gaucge dans la string
          pointerIndex = my_char_pointer_1 - fona3g_replybuffer; //Lorsque pointerIndex sera 0, ca veut dire qu'on a remonter jusqu'au debut de la string sans avoir trouve la virgule recherchee.
        }

        #ifdef METEO_DEBUG
         DEBUG_PRINT(F("pointerIndex="));
         DEBUG_PRINT(pointerIndex);
         DEBUG_PRINT(F(" car trouve="));
         DEBUG_PRINTLN(my_char_pointer_1);
        #endif
      }
      
      if (pointerIndex > 0) {
        //Si on a trouve la virgule, on re-incremente le pointer parcequ'on ne la veut pas dans la reponse
        my_char_pointer_1++;

        //Maintenant on enleve le "V" a la fin de la string
        my_char_pointer_2 = strstr_P(my_char_pointer_1, (const char PROGMEM *)F("V"));
        if (strncmp_P(my_char_pointer_2, (const char PROGMEM *)F("V"), strlen_P((const char PROGMEM *)F("V"))) == 0) {
         //my_char_pointer_1[strlen(my_char_pointer_1)-1] = 0; 
         my_char_pointer_2[0] = 0;
        }

        myResponse_battery_voltage = strtod(my_char_pointer_1, &ptr); 
        
        #ifdef METEO_DEBUG
         DEBUG_PRINT(F("BAT="));
         DEBUG_PRINTLN(my_char_pointer_1);
        #endif
      
      } else {
        #ifdef METEO_DEBUG
         DEBUG_PRINTLN(F("Virgule non trouvee"));
        #endif
      }
    } 
    // ***** Fin Garde voltage batterie
    
    #ifdef METEO_DEBUG
     DEBUG_PRINT(F("REPONSE VOLTAGE BATTERIE: "));
     DEBUG_PRINTLN(fona3g_replybuffer);
    #endif  
    fona3g_waitForDeferredResponse(F("OK"),3000); 

  } 
  
  if (canContinue) {
    strcpy_P(atCommand, (const char PROGMEM *)F("AT+CCHSET=0,0")); //AT+CCHSET Set the parameter of common channel service
    canContinue = fona3g_sendCheckReply(atCommand, F("OK"));
    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);
  }
  
  if (canContinue) {
    strcpy_P(atCommand, (const char PROGMEM *)F("AT+CGSOCKCONT=1,\"IP\",\"")); //AT+CGSOCKCONT Define socket PDP context....
    strcat_P(atCommand, (char*)mySettings_cell_apn); //Variable PROGMEM donc manipulation spéciale
    strcat_P(atCommand, (const char PROGMEM *)F("\""));
    canContinue = fona3g_sendCheckReply(atCommand, F("OK"));
    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);
  }

  if (canContinue) {
    strcpy_P(atCommand, (const char PROGMEM *)F("AT+CSOCKSETPN=1")); //AT+CSOCKSETPN Set active PDP context’s profile number
    canContinue = fona3g_sendCheckReply(atCommand, F("OK"));
    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);
  }

  if (canContinue) {
    strcpy_P(atCommand, (const char PROGMEM *)F("AT+CCHSTART")); //AT+CCHSTART Acquire  common channel service
    canContinue = fona3g_sendCheckReply(atCommand, F("OK"));
    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);    
  }

  if (canContinue) {
    canContinue = fona3g_waitForDeferredResponse(F("+CCHSTART: 0"),15000);
  }

  if (canContinue) {
    strcpy_P(atCommand, (const char PROGMEM *)F("AT+CCHSTATE")); //AT+CCHSTATE Get the state of common channel //Juste pour etre certain qu'on est bien connecte sur le reseau (4 = Opened network)
    fona3g_sendOnly(atCommand);
    canContinue = fona3g_waitForDeferredResponse(F("+CCHSTATE: 4,1,1"),15000);
    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);
  }

  if (canContinue) {      
    canContinue = fona3g_waitForDeferredResponse(F("OK"),15000);
  }

  #ifndef METEO_DEBUG_DO_NOT_SEND_DATA_TO_SERVER //Si je defini cette constante plus haut dans le programme, les donnees ne s'enverront pas sur le serveur web histoire de ne pas payer de bande passant durant les tests
  
  if (canContinue) {
    strcpy_P(atCommand, (const char PROGMEM *)F("AT+CCHOPEN=0,\"")); //AT+CCHOPEN Open a channel
    strcat_P(atCommand, (char*)mySettings_upload_web_server_host_name); //Variable PROGMEM donc manipulation spéciale
    strcat_P(atCommand, (const char PROGMEM *)F("\","));
    strcat_P(atCommand, (char*)mySettings_upload_web_server_port); //Variable PROGMEM donc manipulation spéciale
    strcat_P(atCommand, (const char PROGMEM *)F(",2"));
    canContinue = fona3g_sendCheckReply(atCommand, F("OK")); 
    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);   
  }

  if (canContinue) {
    canContinue = fona3g_waitForDeferredResponse(F("+CCHOPEN: 0,0"),15000);
  }  
  
  if (canContinue) { 
    //strcpy(httpRequest, "GET /page1.php?data=mes-donnees&pwd=mot-de-passe HTTP/1.1");
    strcpy_P(httpRequest, (const char PROGMEM *)F("GET /"));
    strcat_P(httpRequest, (char*)mySettings_upload_web_server_page_name); //Variable PROGMEM donc manipulation spéciale
    strcat_P(httpRequest, (const char PROGMEM *)F("?data="));
    strcat(httpRequest, dataToSend);

    //SENSOR_VALUES_ARRAY_WAIT_DELAY_IN_SECONDS
    //Directement a la suite du parametre data, on ajoute le voltage,memoire libre et SENSOR_VALUES_ARRAY_WAIT_DELAY_IN_SECONDS. Voici un exemple:V99.999M9999D120T87
    dtostrf(myResponse_battery_voltage, 4, 3, tempBuffer);  //4 is mininum width, 3 is precision
    strcat_P(httpRequest,(const char PROGMEM *)F("V"));
    strcat(httpRequest,tempBuffer);
    
    strcat_P(httpRequest,(const char PROGMEM *)F("M"));
    itoa(freeMemory(),tempBuffer,10);
    strcat(httpRequest,tempBuffer);
    
    strcat_P(httpRequest,(const char PROGMEM *)F("D"));
    itoa(SENSOR_VALUES_ARRAY_WAIT_DELAY_IN_SECONDS,tempBuffer,10);
    strcat(httpRequest,tempBuffer);

    strcat_P(httpRequest,(const char PROGMEM *)F("T"));
    //itoa(0,tempBuffer,10);
    ltoa(myResponse_server_date_time,tempBuffer,10);
    strcat(httpRequest,tempBuffer);
    
    strcat_P(httpRequest, (const char PROGMEM *)F("&pwd="));
    strcat_P(httpRequest, (char*)mySettings_upload_web_server_page_password); //Variable PROGMEM donc manipulation spéciale
    strcat_P(httpRequest, (const char PROGMEM *)F(" HTTP/1.1"));
    strcat_P(httpRequest, (const char PROGMEM *)F("\r\n"));
    strcat_P(httpRequest, (const char PROGMEM *)F("HOST: ")); 
    strcat_P(httpRequest, (char*)mySettings_upload_web_server_host_name); //Variable PROGMEM donc manipulation spéciale
    strcat_P(httpRequest, (const char PROGMEM *)F("\r\n"));
    strcat_P(httpRequest, (const char PROGMEM *)F("Content-Length: 0"));
    strcat_P(httpRequest, (const char PROGMEM *)F("\r\n"));
    strcat_P(httpRequest, (const char PROGMEM *)F("\r\n"));
    
    // J'ajoute +2 ci-dessous parceque c'est le CR + LF qui est automatiquement envoyé par la commande fonaSerial->println().
    // Si jamais ca ne fonctionnerait plus, on pourrait mettre +0 et ca fonctionnerait quand même. 
    // En effet, c'est pas grave si le serveur recoit 2 caracteres de plus (le CR et LF dans ce cas-ci) qu'indique dans la commande AT+CCHSEND
    // L'important c'est d'en envoyer au moins autant qu'indique dans la commande AT+CCHSEND pour que le serveur web voit que le requete est terminee d'envoyee et 
    // qu'il nous repondre.
    itoa(strlen(httpRequest) + 0,tempBuffer,10); //+2 fonctionne mais je vais laisser +0. Voir commentaires ci-haut
    strcpy_P(atCommand, (const char PROGMEM *)F("AT+CCHSEND=0,")); //AT+CCHSEND Send data to peer
    strcat(atCommand, tempBuffer);    

    //Meteo_updateLCD(1,atCommand);
    //Meteo_updateLCD(2,fona3g_replybuffer);
    fona3g_sendOnly(atCommand);
    canContinue = fona3g_sendCheckReply(httpRequest, F("OK")); //On envoi la requete au serveur web.
    
    if (canContinue) {
      //Si le serveur a recu le requete et a repondu OK, je considere que la communication avec le serveur a bien fonctionnee
      myResponse_last_comm_status = METEO_CMD_STATUS_SUCCESS;
      canContinue = fona3g_waitForDeferredResponse(F("+CCHRECV:"),15000);  // Debut Page Header
      //canContinue = fona3g_waitForDeferredResponse(F("+CCHRECV:"),15000);  // Debut Page Body //J'ai change de serveur et je ne recoit plus deux fois +CCHRECV comme au par avant... A verifier...
      fona3g_getAndProcessPageBody(20000);   // A mesure qu'on recoit le body, on décortique la reponse et on met a jour certaines variables globales.
      
      // L'appel de fonction fona3g_getAndProcessPageBody() ci-haut a mis a jour la variable globale myResponse_status
      retour = (myResponse_status == METEO_CMD_STATUS_SUCCESS);
      fona3g_waitForDeferredResponse(F("BIDON BIDON"),3000); //Sert pas a grand chose
    }
  } 

  #endif 
  
  if (canContinue) {
    fona3g_power_down();
  }

  #ifdef METEO_DEBUG
   DEBUG_PRINT(F("freeMemory() apres fona3g_power_down()="));
   DEBUG_PRINTLN(freeMemory());
  #endif
  
  #ifdef METEO_DEBUG_DO_NOT_SEND_DATA_TO_SERVER 
   return true; //On simule que tout s'est bien envoye
  #endif 
  
  return retour;
}

// Fonction qui "power on" le FONA 3G et l'initialise. 
// En effet, puisque j'envoi les donnees au serveur seulement 1 fois au 10 minutes,
// je "power off" le FONA a chaque fois pour economiser la batterie.
boolean fona3g_init(Stream &port) {
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINTLN(F(""));
  #endif
  
  boolean retour = false;
  mySerial_Fona = &port;
  
  pinMode(FONA_RST, OUTPUT);
  pinMode(FONA_KEY, OUTPUT);

  // Power on du fona 
  // Ca ne "power on" pas le fona en fait. C'est le reset ci-dessous qui le "power on" mais pour que ca fonctionne, 
  // cette partie est necessaire sinon il ne se passera rien dans le reset ci-dessous.
  #ifdef METEO_DEBUG
   DEBUG_PRINTLN(F("Power on du fona..."));
  #endif  
  digitalWrite(FONA_KEY, LOW);
  delay(5100); 
  
  //Reset du fona
  #ifdef METEO_DEBUG
   DEBUG_PRINTLN(F("Reset du fona..."));
  #endif
  digitalWrite(FONA_RST, HIGH);
  delay(10); 
  digitalWrite(FONA_RST, LOW);
  delay(100); 
  digitalWrite(FONA_RST, HIGH);  
  #ifdef METEO_DEBUG
   DEBUG_PRINTLN(F("Fin du reset du fona..."));
  #endif
  
  Serial.println(F("FONA: Essai d'ouvrir le port en envoyant la commande AT plusieurs fois."));
  Serial.println(F("FONA: On attend un maximum de 7 secondes pour laisser le temps au fona de redemarrer..."));

  // give 7 seconds to reboot
  int16_t timeout = 7000;

  while (timeout > 0) {
    while (mySerial_Fona->available()) mySerial_Fona->read();
    if (fona3g_sendCheckReply(F("AT"), F("OK")))
      break;
    while (mySerial_Fona->available()) mySerial_Fona->read();
    if (fona3g_sendCheckReply(F("AT"), F("AT"))) 
      break;

    delay(500);
    timeout-=500;
  }
  
  if (timeout <= 0) {
    Serial.println(F("FONA: Oups! Pas de reponse. On tente un ultime essai..."));
    fona3g_sendCheckReply(F("AT"), F("OK"));
    delay(100);
    fona3g_sendCheckReply(F("AT"), F("OK"));
    delay(100);
    fona3g_sendCheckReply(F("AT"), F("OK"));
    delay(100);
  }

  // turn off Echo!
  fona3g_sendCheckReply(F("ATE0"), F("OK"));
  delay(100);

  if (! fona3g_sendCheckReply(F("ATE0"), F("OK"))) {
    retour = false;
  } else {
    retour = true;
  }
  
  if (retour) {
    // turn on hangupitude
    fona3g_sendCheckReply(F("AT+CVHU=0"), F("OK"));
    delay(100);
    fona3g_flushInput();
    
    #ifdef METEO_DEBUG
     DEBUG_PRINT(F("\t---> ")); DEBUG_PRINTLN(F("ATI"));
    #endif
    
    mySerial_Fona->println(F("ATI"));
    fona3g_readline(500, true);
    
    #ifdef METEO_DEBUG
     DEBUG_PRINT (F("\t<--- ")); DEBUG_PRINTLN(fona3g_replybuffer);
    #endif 
  }

  Serial.print(F("FONA: "));  
  Serial.print(__func__);  
  Serial.print(F("()=>"));  
  
  if (retour) {  
   Serial.println(F("Succes"));   
  } else {
   Serial.println(F("Erreur"));
  }  
 
  return retour;
}

// Fonction qui "power off" le FONA 3G
void fona3g_power_down() {
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINTLN(F(" debut..."));
  #endif
  
  digitalWrite(FONA_KEY, HIGH); //Important sinon le fona ne ferme pas meme en envoyant la commande ci-dessous  
  mySerial_Fona->println(F("AT+CPOF")); //Power down the fona  
  fona3g_sendCheckReply(F("AT+CPOF"), F("OK"));    
    
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINTLN(F(" fin!"));
  #endif
  
  return;
}

// Fonction qui traite et decortique le corps de la page web recue a la recherche de la reponse qui 
// indique si les donnees envoyees ont bien ete enregistrees sur le serveur. Le serveur retourne egalement d'autres informations utiles
// comme la date et l'heure du serveur. Le serveur retourne aussi un message mais pour le moment, on ne s'en sert pas.
boolean fona3g_getAndProcessPageBody(uint16_t timeout) {
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINTLN(F(""));
  #endif
  
  Serial.print(F("FONA: "));
  Serial.print(__func__);
  Serial.println(F("()..."));

  boolean  retour = false;
  uint16_t myStep = 5000; //100; //ms
  uint16_t longeur;  
  char    *myPointer;
  char    *ptr;
  uint32_t millisStartedAt = millis();
  
  while ((millisStartedAt + timeout) > millis()) {
   longeur = fona3g_readline(myStep, false); 
   if (longeur > 0) {
    #ifdef METEO_DEBUG
     DEBUG_PRINT (F("FONA: ***b nbr car recu=")); 
     DEBUG_PRINT(longeur);
     DEBUG_PRINT (F("\t<--- ")); 
     DEBUG_PRINTLN(fona3g_replybuffer);
    #endif
   }

   //Si on recoit +CCH_PEER_CLOSED: 0 ca veut dire que la connection avec le serveur s'est ferme et que la page est terminee de telechargee.
   if (strncmp_P(fona3g_replybuffer, (const char PROGMEM *)F("+CCH_PEE"), strlen_P((const char PROGMEM *)F("+CCH_PEE"))) == 0) { //On a termine de recevoir la page //+CCH_PEER_CLOSED: //Je ne sait pas pourquoi mais des fois on recoit seulement +CCH_PEE au lieu de +CCH_PEER_CLOSED...
    retour = true;
    break;    
   }

   // Patch GP: Selon la documentation, on devrait recevoir +CCH_PEER_CLOSED: 0 mais on ne recoit jamais le "+" au debut...
   if (strncmp_P(fona3g_replybuffer, (const char PROGMEM *)F("CCH_PEE"), strlen_P((const char PROGMEM *)F("CCH_PEE"))) == 0) { //On a termine de recevoir la page //+CCH_PEER_CLOSED: //Je ne sait pas pourquoi mais des fois on recoit seulement +CCH_PEE au lieu de +CCH_PEER_CLOSED...
    retour = true;
    break;    
   }

   // Reponse pour savoir si les donnees que nous avons envoyees au serveur ont bien ete enregistrees.
   if (strncmp_P(fona3g_replybuffer, (const char PROGMEM *)F("R="), strlen_P((const char PROGMEM *)F("R="))) == 0) { 
    //R=0 -> Erreur
    //R=1 -> Succes (Les donnees envoyees ont bien etees enregistrees dans la base de donnees sur le serveur)
    //myResponse_status = (strcmp_P(fona3g_replybuffer,(const char PROGMEM *)F("R=1")) == 0);
    if (strcmp_P(fona3g_replybuffer,(const char PROGMEM *)F("R=1")) == 0) {
      myResponse_status = METEO_CMD_STATUS_SUCCESS;
    } else {
      myResponse_status = METEO_CMD_STATUS_ERROR;
    }
    Serial.print(F("\tReponse du serveur=>"));
    Serial.print(fona3g_replybuffer);
    Serial.print(F(" =>Donc "));
    
    if (myResponse_status == METEO_CMD_STATUS_SUCCESS) {
     Serial.println(F("Success"));  
    } else {
     Serial.println(F("Erreur"));  
    }
    
    #ifdef METEO_DEBUG
     DEBUG_PRINT(F("FONA: ***myResponse_status="));
     DEBUG_PRINTLN(myResponse_status);
    #endif
    
   }

   // Date recue du serveur
   if (strncmp_P(fona3g_replybuffer, (const char PROGMEM *)F("DT="), strlen_P((const char PROGMEM *)F("DT="))) == 0) { 
    myPointer = strstr_P(fona3g_replybuffer, (const char PROGMEM *)F("="));
    if (myPointer) {
      myPointer++; //Pour ne pas avoir le caractère =
      myResponse_server_date_time = strtol(myPointer, &ptr, 10); //atoi(myPointer); //Le serveur nous envoi le Unix Timestamp selon le fuseau horaire de Montreal, Canada et non selon le fuseau UTC.
      myResponse_server_date_time_millis = millis();
      
      #ifdef METEO_DEBUG
       DEBUG_PRINT(F("***myResponse_server_date_time="));
       DEBUG_PRINTLN(myResponse_server_date_time);
      #endif
    }
   }

   // Message du jour recu du serveur a afficher sur le LCD. Peut-etre mettre en fonction plus tard
   /*
   if (strncmp_P(fona3g_replybuffer, (const char PROGMEM *)F("MSG="), strlen_P((const char PROGMEM *)F("MSG="))) == 0) {
    myPointer = strstr_P(fona3g_replybuffer, (const char PROGMEM *)F("="));
    if (myPointer) {
      myPointer++; //Pour ne pas avoir le caractère =
      strcpy(myResponse_message_to_display,myPointer);
      #ifdef METEO_DEBUG
       DEBUG_PRINT(F("***myResponse_message_to_display="));
       DEBUG_PRINTLN(myResponse_message_to_display);
      #endif
    }
   }
   */
   
  }

  Serial.print(F("FONA: "));
  Serial.print(__func__);
  Serial.print(F("()=>"));
  
  if (retour) {
   Serial.println(F("Ok, body recu au complet."));
  } else {
   Serial.println(F("*Humm, pas recu +CCH_PEER_CLOSED:..."));
  }

  return retour;
}

// "Function overloading" pour fonction du meme nom plus bas
boolean fona3g_waitForDeferredResponse(const __FlashStringHelper * expectedResponse, uint16_t timeout ) { 
  boolean retour;
  //char * myStringOfChar = (char *) calloc(MAX_LCD_CHARACTERS + 1, sizeof(char)); 
  char * myStringOfChar = (char *) calloc(strlen_P( (const char PROGMEM *)expectedResponse) + 1, sizeof(char)); 
  strncpy_P(myStringOfChar,(const char PROGMEM *)expectedResponse,MAX_LCD_CHARACTERS);

  #ifdef METEO_DEBUG     
    //DEBUG_PRINT(F("__FlashStringHelper * expectedResponse="));
    //DEBUG_PRINTLN(expectedResponse);
    //DEBUG_PRINT(F("myStringOfChar="));
    //DEBUG_PRINTLN(myStringOfChar);
  #endif
  
  retour = fona3g_waitForDeferredResponse(myStringOfChar,timeout);
  free(myStringOfChar);
  
  return retour;
}

// Fonction qui attend une reponse (suite a une commande AT) du FONA en differee
boolean fona3g_waitForDeferredResponse(char const *expectedResponse, uint16_t timeout) {
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINT(F(" ***Attente de la reponse "));
   DEBUG_PRINT(expectedResponse);
   DEBUG_PRINTLN(F(" en differe..."));
  #endif

  uint16_t myStep = 2000; //100; //ms
  uint16_t longeur;
  boolean  retour = false;
  uint32_t millisStartedAt = millis();
  
  while ((millisStartedAt + timeout) > millis()) {  
  //while (timeout > 0) {
   //timeout = timeout - myStep;
   longeur = fona3g_readline(myStep, false); 
   if (longeur > 0) {
    #ifdef METEO_DEBUG
     DEBUG_PRINT(F("***d nbr car recu=")); 
     DEBUG_PRINT(longeur);
     DEBUG_PRINT(F("***\t<--- ")); 
     DEBUG_PRINTLN(fona3g_replybuffer);
    #endif    
   }
   
   if (strncmp(fona3g_replybuffer, expectedResponse, strlen(expectedResponse)) == 0) { //On a recu la reponse qu'on attendait
    retour = true;
    break;    
   }

  }

  #ifdef METEO_DEBUG
   DEBUG_PRINT(F("FONA: "));
   DEBUG_PRINT(__func__);
   DEBUG_PRINT(F("()=>"));
  
   if (retour) {
    DEBUG_PRINTLN(F("Bonne reponse recu!"));
   } else {
    DEBUG_PRINTLN(F("Rien ou autre chose recu."));
   }
  #endif

  return retour;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
uint8_t fona3g_sendOnly(char const *atCommand) {
  fona3g_flushInput();
  
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINT(F("\t---> ")); DEBUG_PRINTLN(atCommand);
  #endif
  
  fonaSerial->println(atCommand);

  return true;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
boolean fona3g_sendCheckReply(char *send, char *reply, uint16_t timeout) {
  boolean retour = false;

  if (fona3g_getReply(send, timeout) > 0) {
    retour = (strcmp(fona3g_replybuffer, reply) == 0);
  }
  
  #ifdef METEO_DEBUG
   //DEBUG_PRINT(__func__);
   //DEBUG_PRINT(F("()=>"));  
   //DEBUG_PRINTLN(retour);     
  #endif   
    
  return retour;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
boolean fona3g_sendCheckReply(FONAFlashStringPtr send, FONAFlashStringPtr reply, uint16_t timeout) {
  boolean retour = false;
  
  if (fona3g_getReply(send, timeout) > 0) {
    retour = (strcmp_P(fona3g_replybuffer, (char PROGMEM*)reply) == 0);
  }

  #ifdef METEO_DEBUG
   //DEBUG_PRINT(__func__);  
   //DEBUG_PRINT(F("()=>"));  
   //DEBUG_PRINTLN(retour);     
  #endif   
  
  return retour;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
boolean fona3g_sendCheckReply(char* send, FONAFlashStringPtr reply, uint16_t timeout) {
  boolean retour = false;
  
  if (fona3g_getReply(send, timeout) > 0) {
    retour = (strcmp_P(fona3g_replybuffer, (char PROGMEM*)reply) == 0);
  }
  
  #ifdef METEO_DEBUG
   //DEBUG_PRINT(__func__);  
   //DEBUG_PRINT(F("()=>"));  
   //DEBUG_PRINTLN(retour);     
  #endif   
    
  return retour;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
boolean fona3g_sendCheckReply(FONAFlashStringPtr prefix, char *suffix, FONAFlashStringPtr reply, uint16_t timeout) {
  boolean retour = false;
  
  if (fona3g_getReply(prefix, suffix, timeout) > 0) {
    retour = (strcmp_P(fona3g_replybuffer, (char PROGMEM*)reply) == 0);
  }

  #ifdef METEO_DEBUG
   //DEBUG_PRINT(__func__);  
   //DEBUG_PRINT(F("()=>"));  
   //DEBUG_PRINTLN(retour);     
  #endif   
  
  return retour;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
boolean fona3g_sendCheckReply(FONAFlashStringPtr prefix, int32_t suffix, FONAFlashStringPtr reply, uint16_t timeout) {
  boolean retour = false;
  
  if (fona3g_getReply(prefix, suffix, timeout) > 0) {
    retour = (strcmp_P(fona3g_replybuffer, (char PROGMEM*)reply) == 0);
  }

  #ifdef METEO_DEBUG
   //DEBUG_PRINT(__func__);  
   //DEBUG_PRINT(F("()=>"));  
   //DEBUG_PRINTLN(retour);     
  #endif   
  
  return retour;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
boolean fona3g_sendCheckReply(FONAFlashStringPtr prefix, int32_t suffix1, int32_t suffix2, FONAFlashStringPtr reply, uint16_t timeout) {
  boolean retour = false;
  
  if (fona3g_getReply(prefix, suffix1, suffix2, timeout) > 0) {
    retour = (strcmp_P(fona3g_replybuffer, (char PROGMEM*)reply) == 0);
  }

  #ifdef METEO_DEBUG
   //DEBUG_PRINT(__func__);  
   //DEBUG_PRINT(F("()=>"));  
   //DEBUG_PRINTLN(retour);     
  #endif   
  
  return retour;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
uint8_t fona3g_getReply(char *send, uint16_t timeout) {
  fona3g_flushInput();
  
  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINT(F("\t---> ")); DEBUG_PRINTLN(send);
  #endif
  
  fonaSerial->println(send);
  uint8_t l = fona3g_readline(timeout);

  #ifdef METEO_DEBUG
   DEBUG_PRINT (F("\t<--- ")); DEBUG_PRINTLN(fona3g_replybuffer);
  #endif
  
  return l;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
uint8_t fona3g_getReply(FONAFlashStringPtr send, uint16_t timeout) {
  fona3g_flushInput();

  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINT(F("\t---> ")); DEBUG_PRINTLN(send);
  #endif
  
  fonaSerial->println(send);
  uint8_t l = fona3g_readline(timeout);

  #ifdef METEO_DEBUG
   DEBUG_PRINT (F("\t<--- ")); DEBUG_PRINTLN(fona3g_replybuffer);
  #endif
  
  return l;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
uint8_t fona3g_getReply(FONAFlashStringPtr prefix, char *suffix, uint16_t timeout) {
  fona3g_flushInput();

  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINT(F("\t---> ")); DEBUG_PRINT(prefix); DEBUG_PRINTLN(suffix);
  #endif
  
  fonaSerial->print(prefix);
  fonaSerial->println(suffix);
  uint8_t l = fona3g_readline(timeout);

  #ifdef METEO_DEBUG
   DEBUG_PRINT (F("\t<--- ")); DEBUG_PRINTLN(fona3g_replybuffer);
  #endif
  
  return l;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
uint8_t fona3g_getReply(FONAFlashStringPtr prefix, int32_t suffix, uint16_t timeout) {
  fona3g_flushInput();

  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINT(F("\t---> ")); DEBUG_PRINT(prefix); DEBUG_PRINTLN(suffix, DEC);
  #endif 
  
  fonaSerial->print(prefix);
  fonaSerial->println(suffix, DEC);
  uint8_t l = fona3g_readline(timeout);

  #ifdef METEO_DEBUG
   DEBUG_PRINT (F("\t<--- ")); DEBUG_PRINTLN(fona3g_replybuffer);
  #endif
  
  return l;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
uint8_t fona3g_getReply(FONAFlashStringPtr prefix, int32_t suffix1, int32_t suffix2, uint16_t timeout) {
  fona3g_flushInput();

  #ifdef METEO_DEBUG
   DEBUG_PRINT(__func__); DEBUG_PRINT(F("()"));
   DEBUG_PRINT(F("\t---> ")); DEBUG_PRINT(prefix);
   DEBUG_PRINT(suffix1, DEC); DEBUG_PRINT(','); DEBUG_PRINTLN(suffix2, DEC);
  #endif
  
  fonaSerial->print(prefix);
  fonaSerial->print(suffix1);
  fonaSerial->print(',');
  fonaSerial->println(suffix2, DEC);
  uint8_t l = fona3g_readline(timeout);

  #ifdef METEO_DEBUG
   DEBUG_PRINT (F("\t<--- ")); DEBUG_PRINTLN(fona3g_replybuffer);
  #endif
  
  return l;
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
void fona3g_flushInput() {
    // Read all available serial input to flush pending data.
    uint16_t timeoutloop = 0;
    while (timeoutloop++ < 40) {
        while(fonaSerial->available()) {
            fonaSerial->read();
            timeoutloop = 0;  // If char was received reset the timer
        }
        delay(1);
    }
}

// Fonction copiee de la librairie du FONA 3G (par Adafruit Industries) mais parfois adaptee a nos besoins
uint8_t fona3g_readline(uint16_t timeout, boolean multiline) {
  uint16_t replyidx = 0;

  while (timeout--) {
    if (replyidx >= 254) {
      #ifdef METEO_DEBUG
       //DEBUG_PRINTLN(F("FONA: MAXIMUM DE fona3g_replybuffer ATTEINT!"));
      #endif
      
      break;
    }

    while(fonaSerial->available()) {
      char c =  fonaSerial->read();
      if (c == '\r') continue;
      if (c == 0xA) {
        if (replyidx == 0)   // the first 0x0A is ignored
          continue;

        if (!multiline) {
          timeout = 0;       // the second 0x0A is the end of the line
          break;
        }
      }
      fona3g_replybuffer[replyidx] = c;
      //DEBUG_PRINT(c, HEX); DEBUG_PRINT(F("#")); DEBUG_PRINTLN(c);
      replyidx++;
    }

    if (timeout == 0) {
      //DEBUG_PRINTLN(F("TIMEOUT"));
      break;
    }
    delay(1);
  }
  
  fona3g_replybuffer[replyidx] = 0;  // null term
  
  return replyidx;
}

